/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: extern.c,v 1.2 2001/01/02 20:56:19 lefessan Exp $ */

/* Structured output */

#include <stdio.h>
#include <string.h>
#include "alloc.h"
#include "fail.h"
#include "gc.h"
#include "intext.h"
#include "io.h"
#include "memory.h"
#include "misc.h"
#include "mlvalues.h"
#include "reverse.h"
#include "str.h"
#include "instruct.h"
#include "join.h"
#include "roots.h"
/* To keep track of sharing in externed objects */

#define HEADER_SIZE 5

typedef unsigned long byteoffset_t;

struct extern_obj {
  byteoffset_t ofs;
  value obj;
};

extern void finalize_channel(value vchan);

static byteoffset_t initial_ofs = 1; /* Initial value of object offsets */
static struct extern_obj * extern_table = NULL;
static unsigned long extern_table_size;
static byteoffset_t obj_counter;    /* Number of objects emitted so far */

/* join */
value saved_pointers = (value)NULL;
static unsigned long saved_pointers_size;
static unsigned long nbr_saved_pointers;
static int inside_join = 0;
static int register_saved_roots = 1;
/* join */

#ifdef ARCH_SIXTYFOUR
#define Hash(v) (((unsigned long) ((v) >> 3)) % extern_table_size)
#else
#define Hash(v) (((unsigned long) ((v) >> 2)) % extern_table_size)
#endif

/* Allocate a new extern table */
static void alloc_extern_table(void)
{
  asize_t i;
  extern_table = (struct extern_obj *)
                 stat_alloc(extern_table_size * sizeof(struct extern_obj));
  for (i = 0; i < extern_table_size; i++) extern_table[i].ofs = 0;
}

/* Grow the extern table */
static void resize_extern_table(void)
{
  asize_t oldsize;
  struct extern_obj * oldtable;
  value obj;
  byteoffset_t ofs;
  asize_t i, h;

  oldsize = extern_table_size;
  oldtable = extern_table;
  extern_table_size = 2 * extern_table_size;
  alloc_extern_table();
  for (i = 0; i < oldsize; i++) {
    ofs = oldtable[i].ofs;
    if (ofs >= initial_ofs) {
      obj = oldtable[i].obj;
      h = Hash(obj);
      while (extern_table[h].ofs >= initial_ofs) {
        h++;
        if (h >= extern_table_size) h = 0;
      }
      extern_table[h].ofs = ofs;
      extern_table[h].obj = obj;
    }
  }
  stat_free((char *) oldtable);
}

/* Free the extern table. We keep it around for next call if
   it's still small (we did not grow it) and the initial offset
   does not risk running over next time. */
static void free_extern_table(void)
{
  if (extern_table_size > INITIAL_EXTERN_TABLE_SIZE ||
      initial_ofs >= INITIAL_OFFSET_MAX) {
    stat_free((char *) extern_table);
    extern_table = NULL;
  }
}

/* To buffer the output */

static char * extern_block, * extern_ptr, * extern_limit;
static int extern_block_malloced;

static void alloc_extern_block(void)
{
  extern_block = stat_alloc(INITIAL_EXTERN_BLOCK_SIZE);
  extern_limit = extern_block + INITIAL_EXTERN_BLOCK_SIZE;
  extern_ptr = extern_block;
  extern_block_malloced = 1;
}

static void resize_extern_block(int required)
{
  long curr_pos, size, reqd_size;

  if (! extern_block_malloced) {
    initial_ofs += obj_counter;
    free_extern_table();
    failwith("Marshal.to_buffer: buffer overflow");
  }
  curr_pos = extern_ptr - extern_block;
  size = extern_limit - extern_block;
  reqd_size = curr_pos + required;
  while (size <= reqd_size) size *= 2;
  extern_block = stat_resize(extern_block, size);
  extern_limit = extern_block + size;
  extern_ptr = extern_block + curr_pos;
}

/* join */
static void alloc_saved_pointers(void)
{
  asize_t i;

  if(register_saved_roots) /* Only once */
    {
      register_global_root(&saved_pointers);
      register_saved_roots = 0;
    }
  saved_pointers_size = INITIAL_SAVED_POINTERS_SIZE;
  saved_pointers = (value)stat_alloc(saved_pointers_size*sizeof(value));
  for (i = 0; i < saved_pointers_size; i++) 
    Field(saved_pointers,i) = Val_int(0);
}

static void free_saved_pointers(void)
{
  if (saved_pointers != (value)NULL){
    if (saved_pointers_size > INITIAL_SAVED_POINTERS_SIZE){
      stat_free((char *) saved_pointers);
      saved_pointers = (value)NULL;
    } else {
      while(nbr_saved_pointers){
	nbr_saved_pointers--;
	Field(saved_pointers,nbr_saved_pointers) = Val_int(0);
      }
    }
  }
}


static void resize_saved_pointers(void)
{
  asize_t i;

  saved_pointers_size *= 2;
  saved_pointers = (value)stat_resize((char*)saved_pointers,
			       saved_pointers_size*sizeof(value));
  for(i=nbr_saved_pointers;i<saved_pointers_size;++i)
    Field(saved_pointers,i) = Val_int(0);
}
/* join */

/* Write characters, integers, and blocks in the output buffer */

#define Write(c) \
  if (extern_ptr >= extern_limit) resize_extern_block(1); \
  *extern_ptr++ = (c)

static void writeblock(char *data, long int len)
{
  if (extern_ptr + len > extern_limit) resize_extern_block(len);
  bcopy(data, extern_ptr, len);
  extern_ptr += len;
}

static void writecode8(int code, long int val)
{
  if (extern_ptr + 2 > extern_limit) resize_extern_block(2);
  extern_ptr[0] = code;
  extern_ptr[1] = val;
  extern_ptr += 2;
}

static void writecode16(int code, long int val)
{
  if (extern_ptr + 3 > extern_limit) resize_extern_block(3);
  extern_ptr[0] = code;
  extern_ptr[1] = val >> 8;
  extern_ptr[2] = val;
  extern_ptr += 3;
}

int rang;

static void write32(long int val)
{
  if (extern_ptr + 4 > extern_limit) resize_extern_block(4);
  extern_ptr[0] = val >> 24;
  extern_ptr[1] = val >> 16;
  extern_ptr[2] = val >> 8;
  extern_ptr[3] = val;
  extern_ptr += 4;
}

static void writecode32(int code, long int val)
{
  if (extern_ptr + 5 > extern_limit) resize_extern_block(5);
  extern_ptr[0] = code;
  extern_ptr[1] = val >> 24;
  extern_ptr[2] = val >> 16;
  extern_ptr[3] = val >> 8;
  extern_ptr[4] = val;
  extern_ptr += 5;
}

#ifdef ARCH_SIXTYFOUR
static void writecode64(int code, long val)
{
  int i;
  if (extern_ptr + 9 > extern_limit) resize_extern_block(9);
  *extern_ptr ++ = code;
  for (i = 64 - 8; i >= 0; i -= 8) *extern_ptr++ = val >> i;
}
#endif

/* Marshal the given value in the output buffer */

static unsigned long size_32;  /* Size in words of 32-bit block for struct. */
static unsigned long size_64;  /* Size in words of 64-bit block for struct. */

static int extern_ignore_sharing; /* Flag to ignore sharing */
static int extern_closures;     /* Flag to allow externing code pointers */

static void extern_invalid_argument(char *msg)
{
  if (extern_block_malloced) stat_free(extern_block);
  initial_ofs += obj_counter;
  free_extern_table();
  invalid_argument(msg);
}

static int predef_exception(value v)
{
  if( v == Field(global_data,OUT_OF_MEMORY_EXN)) return OUT_OF_MEMORY_EXN;
  else if(v == Field(global_data,SYS_ERROR_EXN)) return SYS_ERROR_EXN;
  else if(v == Field(global_data,FAILURE_EXN)) return FAILURE_EXN;
  else if(v == Field(global_data,INVALID_EXN)) return INVALID_EXN;
  else if(v == Field(global_data,END_OF_FILE_EXN)) return END_OF_FILE_EXN;
  else if(v == Field(global_data,ZERO_DIVIDE_EXN)) return ZERO_DIVIDE_EXN;
  else if(v == Field(global_data,NOT_FOUND_EXN)) return NOT_FOUND_EXN;
  else if(v == Field(global_data,MATCH_FAILURE_EXN)) return MATCH_FAILURE_EXN;
  else if(v == Field(global_data,STACK_OVERFLOW_EXN))return STACK_OVERFLOW_EXN;
  return -1;
}


/*

On rajoute deux codes CODE_CACHE(n) et CODE_SETCACHE(n,v).


 le cache:

0 : numero prochain emplacement vide
1 : l'identificateur de cache
...: donnees

Il ne faut pas d'allocation: la position dans le cache doit etre placee
sur l'objet sans allocation: on utilise donc un entier pour identifier a la
fois le cache et la position --> limitation a 31 bits
   nombre de stubs dans le cache: 1 << CACHE_SIZE  (default: 512)
   nombre maximal de caches: 1 << (31 - CACHE_SIZE)(default: 4194304)
   CACHE_SIZE maximum: 7
*/
static value extern_cache = Val_unit;
static int next_cache_pos = 0;
static int cache_left = 0;
static int cache_size = 0;
static int cache_number = 0;

#define CACHE_SIZE 10
#define CACHE_OFFSET 2
value next_cache_number = 1;
value cache_create(value size)
{
  int sz = Int_val(size);
  value res = alloc(CACHE_OFFSET +  sz, 0);
  int i;

  if(sz> 1 << CACHE_SIZE) invalid_argument("Cache_create");
  Field(res,0) = Val_int(CACHE_OFFSET);
  Field(res,1) = Val_int(next_cache_number << CACHE_SIZE);
  next_cache_number++;
  for(i=CACHE_OFFSET;i<CACHE_OFFSET + sz;i++){Field(res,i) = Val_int(0);}

  return res;
}

/* retourne 0 si dans le cache */
int cache_put(value stub)
{
  value stub_cache = Int_val(Field(stub,STUB_CACHE));
  int pos;

  if(stub_cache >> CACHE_SIZE == cache_number &&
     Field(extern_cache,
	   (pos = (stub_cache & ((1 << CACHE_SIZE)-1))+CACHE_OFFSET)) == stub){
    /* target already in cache */
    /* fprintf(stderr,"extern: in cache %d \n", pos); */
    writecode16(CODE_CACHE,pos);
    return 1;
  }

  if(cache_left > 0){
    if(next_cache_pos == cache_size) next_cache_pos = CACHE_OFFSET;
    Field(stub,STUB_CACHE) = 
      Val_int((cache_number << CACHE_SIZE)+(next_cache_pos-CACHE_OFFSET));
    Assign(Field(extern_cache,next_cache_pos),stub);
    writecode16(CODE_SETCACHE,next_cache_pos);
    /* fprintf(stderr,"extern: to cache  %d \n", next_cache_pos); */
    cache_left--;
    next_cache_pos++;
  }
  return 0;
}

value cache_clean(value cache)
{
  int i;
  int cache_size = Wosize_val(cache);

  Field(cache,0)= Val_int(CACHE_OFFSET);
  for(i = CACHE_OFFSET;i< cache_size;i++){ Assign(Field(cache,i), Val_unit);}
  
  return Val_unit;
}

/* fin du hack cache ... */

static int check_seen(value vv)
{
  int h;

  if (! extern_ignore_sharing) {
    if (2 * obj_counter >= extern_table_size) resize_extern_table();
    h = Hash(vv);
    while (extern_table[h].ofs >= initial_ofs) {
      if (extern_table[h].obj == vv) {
        byteoffset_t d = obj_counter - (extern_table[h].ofs - initial_ofs);
        if (d < 0x100) {
          writecode8(CODE_SHARED8, d);
        } else if (d < 0x10000) {
            writecode16(CODE_SHARED16, d);
          } else {
            writecode32(CODE_SHARED32, d);
          }
          return 1;
      }
      h++;
      if (h >= extern_table_size) h = 0;
    }    
    
    /* Not seen yet. Record the object */
    extern_table[h].ofs = initial_ofs + obj_counter;
    extern_table[h].obj = vv;
    obj_counter++;
  }
  return 0;
}

static void reloc_pointer(value v)
{
  if(saved_pointers==(value)NULL) 
    alloc_saved_pointers();
  else
    if(nbr_saved_pointers==saved_pointers_size)
      resize_saved_pointers();
    Field(saved_pointers,nbr_saved_pointers) = v;
  nbr_saved_pointers++;
}

static void extern_rec(value v)
{
  tailcall:
  if (Is_long(v)) {
    long n = Long_val(v);
    if (n >= 0 && n < 0x40) {
      Write(PREFIX_SMALL_INT + n);
    } else if (n >= -(1 << 7) && n < (1 << 7)) {
        writecode8(CODE_INT8, n);
      } else if (n >= -(1 << 15) && n < (1 << 15)) {
          writecode16(CODE_INT16, n);
          #ifdef ARCH_SIXTYFOUR
        } else if (n < -(1L << 31) || n >= (1L << 31)) {
            writecode64(CODE_INT64, n);
            #endif
          } else
            writecode32(CODE_INT32, n);
          return;
  }
  if (Is_young(v) || Is_in_heap(v) || Is_atom(v)) {
    header_t hd = Hd_val(v);
    tag_t tag = Tag_hd(hd);
    mlsize_t sz = Wosize_hd(hd);
    asize_t h;
    /* Atoms are treated specially for two reasons: they are not allocated
       in the externed block, and they are automatically shared. */
    if (sz == 0) {
      if (tag < 16) {
        Write(PREFIX_SMALL_BLOCK + tag);
      } else {
        writecode32(CODE_BLOCK32, hd);
      }
      return;
    }
    /* Check if already seen */
    if(check_seen(v)) return;

    /* Output the contents of the object */
    switch(tag) {
      case String_tag: {      
        mlsize_t len = string_length(v);
        if (len < 0x20) {
          Write(PREFIX_SMALL_STRING + len);
        } else if (len < 0x100) {
            writecode8(CODE_STRING8, len);
          } else {
            writecode32(CODE_STRING32, len);
          }
          writeblock(String_val(v), len);
        size_32 += 1 + (len + 4) / 4;
        size_64 += 1 + (len + 8) / 8;
        break;
      }
      case Double_tag: {
        if (sizeof(double) != 8)
          extern_invalid_argument("output_value: non-standard floats");
        Write(CODE_DOUBLE_NATIVE);
        writeblock((char *) v, 8);
        size_32 += 1 + 2;
        size_64 += 1 + 1;
        break;
      }
      case Double_array_tag: {
        mlsize_t nfloats;
        if (sizeof(double) != 8)
          extern_invalid_argument("output_value: non-standard floats");
        nfloats = Wosize_val(v) / Double_wosize;
        if (nfloats < 0x100) {
          writecode8(CODE_DOUBLE_ARRAY8_NATIVE, nfloats);
        } else {
          writecode32(CODE_DOUBLE_ARRAY32_NATIVE, nfloats);
        }
        writeblock((char *) v, Bosize_val(v));
        size_32 += 1 + nfloats * 2;
        size_64 += 1 + nfloats;
        break;
      }
      case Abstract_tag:
      fprintf(stderr,"Abstract tag\n");
      goto final;
      
      case Final_tag:
      /* Instead of extern_invalid_argument, we attempt to replace the
      non-marshallable object by Val_unit, hoping it will not be access
      anymore ... (a better approach would be to add special tags for 
      channels) */
      fprintf(stderr,"Final tag\n");
      if(Field(v,0) == ((value)finalize_channel))
        fprintf(stderr,"OK, a channel ...\n");
      
      final:
      fprintf(stderr,"Size %ld\n",sz);
      Write(PREFIX_SMALL_INT);
      break;
      case Infix_tag:
      writecode32(CODE_INFIXPOINTER, Infix_offset_hd(hd));
      extern_rec(v - Infix_offset_hd(hd));
      break;
      
      case Classinfo_tag:
      reloc_pointer(v);
      writecode32(CODE_CLASSINFO, hd & ~Black);
      {
        int i;

        for (i = 0; i < sz; i++){
          extern_rec(Field(v, i));
        }
      }
      return;      

      case Object_tag:
      case Location_tag:
      case Automata_tag:
      case Thread_tag:
      Good_stub("extern_rec",v);
      
      if(inside_join && cache_put(v)) return;

/* The Marshal tag corresponds to special objects which are encapsulated
with marshaling information. */
      
      case Marshal_tag: 
      
      reloc_data:
      if(!inside_join){
        extern_invalid_argument("output_value: Join value");
        break;
      }
      reloc_pointer(v);
      Write(CODE_JOINPOINTER);
      return;
      
      case Code_tag:{
      
      /* Un pointeur de code encapsule :
          - Une instruction 
          - Un numero de module
          - Un offset
      */
        code_t vv = (code_t)v;
        
        if(vv[0] == THREADED(CODEPTR))
          writecode32(CODE_JOINRETURN,vv[2]);
        else
          writecode32(CODE_JOINCLOSURE,vv[2]);
        
        size_32 += 1 + 4;
        size_64 += 1 + 2;
        v = Field(module_table,vv[1]);
        
        goto tailcall;
      }

/* When mixing native code and bytecode, bytecode closures are preceeded by
a header, corresponding to a native closure calling the bytecode interpreter.
*/
      
      
      case Closure_byte_tag:
      {
        mlsize_t i;
        
        writecode32(CODE_BYTECLOSURE, sz - NATIVE_CLOSURE_SIZE);
        size_32 += 1 + sz - NATIVE_CLOSURE_SIZE + NATIVE_HEADER;
        size_64 += 1 + sz - NATIVE_CLOSURE_SIZE + NATIVE_HEADER;
        if (inside_join){
          int32 offset;
          value module;
          
          module = find_code_ptr(Code_val(v),&offset);
          if(module != Val_unit){
            value vv = Code_val(v);
            
            if(check_seen(vv)) goto suite_closure;
            
            writecode32(CODE_JOINCLOSURE,offset);
            size_32 += 1 + 4;
            size_64 += 1 + 2;
            extern_rec(module);
          
          } else {
            fprintf(stderr,"Closure tag\n");
            extern_invalid_argument("Closure_tag: can not locate code pointer");
          }
        } else { 
          extern_rec(Field(v,0));
        }
        suite_closure:
        if(sz == NATIVE_CLOSURE_SIZE + 1) return;
        for (i = NATIVE_CLOSURE_SIZE + 1; i < sz - 1; i++) 
          extern_rec(Field(v, i));
        v = Field(v, i);
        goto tailcall;
      }
      
      case Weakset_tag:
      extern_invalid_argument("Marshaller: attempt to marshal weakset pointer");    case Weak_tag:
      extern_invalid_argument("Marshaller: attempt to marshal weak pointer");
      
      case Predef_exn_tag : /* une exception predefinie */
      if(inside_join) { 
        int code = predef_exception(v);
        if (code != -1){
          writecode8(CODE_JOINPREDEFEXN, code);
          return;
        }
      }
      
      case Reloc_tag:
      case Exn_tag :
      /* en dehors du join, les modules de relocation sont sauves 
      normalement par le cas par defaut (on passe Exn_tag et Closure_asm_tag)
      */      

      if(inside_join)        goto reloc_data;
      
      case Closure_asm_tag:
      default: {
          mlsize_t i;
          if (tag < 16 && sz < 8) {
            Write(PREFIX_SMALL_BLOCK + tag + (sz << 4));
          } else {
            writecode32(CODE_BLOCK32, hd & ~Black);
          }
          size_32 += 1 + sz;
          size_64 += 1 + sz;
          for (i = 0; i < sz -1; i++){
            extern_rec(Field(v, i));
          }
          v = Field(v, sz - 1);
          goto tailcall;          
        }
      }
      return;
    }
    
    if (inside_join){
      int32 offset;
      value module;
      
      module = find_code_ptr((code_t)v,&offset);
      if(module!=Val_unit){
        value vv = v;
        asize_t h;

        if(check_seen(vv)) return;

        writecode32(CODE_JOINRETURN,offset);
      size_32 += 1 + 4;
      size_64 += 1 + 2;
      extern_rec(module);
      /************************************/
      return;
    } else {
      fprintf(stderr,"Pointer tag\n");
      extern_invalid_argument("abstract value: can not locate code pointer");
    }
  }
  
  if(with_native){
    /* un pointeur vers du code ntaif ?? */
    if ((char *) v >= code_area_start && (char *) v < code_area_end) {
      if (!extern_closures)
        extern_invalid_argument("output_value: functional value");
      writecode32(CODE_CODEASMPOINTER, (char *) v - code_area_start);
      writeblock((char *) asm_code_checksum(), 16);
      return;
    }
  }
  if(with_bytecode){
    if ((char *) v >= ((char*)start_code) && 
        (char *) v < ((char *) start_code + code_size)) {
      if (!extern_closures)
        extern_invalid_argument("output_value: functional value");
      writecode32(CODE_CODEBYTEPOINTER, (char *) v - ((char*)start_code));
      writeblock((char *) byte_code_checksum(), 16);
      return;
    }
  }
  fprintf(stderr,"v = %lx\n",v);
  extern_invalid_argument("output_value: abstract value");
}

enum { NO_SHARING = 1, CLOSURES = 2 };
static int extern_flags[] = { NO_SHARING, CLOSURES };

static long extern_value(value v, value flags)
{
  long res_len;
  int fl;
  /* Parse flag list */
  fl = convert_flag_list(flags, extern_flags);
  extern_ignore_sharing = fl & NO_SHARING;
  extern_closures = fl & CLOSURES;
  /* Allocate hashtable of objects already seen, if needed */
  extern_table_size = INITIAL_EXTERN_TABLE_SIZE;
  if (extern_table == NULL) {
    alloc_extern_table();
    initial_ofs = 1;
  }
  obj_counter = 0;
  size_32 = 0;
  size_64 = 0;
  /* Write magic number */
  write32(Intext_magic_number);
  /* Set aside space for the sizes */
  extern_ptr += 4*4;
   
  /* Marshal the object */
  extern_rec(v);
  /* Update initial offset for next call to extern_value(),
     if we decide to keep the table of shared objects. */
  initial_ofs += obj_counter;
  /* Free the table of shared objects (if needed) */
  free_extern_table();
  /* Write the sizes */
#ifdef ARCH_SIXTYFOUR
  if (size_32 >= (1L << 32) || size_64 >= (1L << 32)) {
    /* The object is so big its size cannot be written in the header.
       Besides, some of the block sizes or string lengths or shared offsets
       it contains may have overflowed the 32 bits used to write them. */
    failwith("output_value: object too big");
  }
#endif
  res_len = extern_ptr - extern_block;
  extern_ptr = extern_block + 4;
  write32(res_len - 5*4); 
  write32(obj_counter);
  write32(size_32);
  write32(size_64);
  
  /* Result is res_len bytes starting at extern_block */
  return res_len;
}

void output_val(struct channel *chan, value v, value flags)
{
  long len;
  alloc_extern_block();
  len = extern_value(v, flags);
  really_putblock(chan, extern_block, len);
  stat_free(extern_block);
}

value output_value(value vchan, value v, value flags) /* ML */
{
  struct channel * channel = Channel(vchan);
  inside_join=0;
  
  
  rang = 0;
  Begin_roots2(v,flags);
  Lock(channel);
  output_val(channel, v, flags);
  Unlock(channel);
  End_roots();
  
  return Val_unit;
}

value output_value_to_string(value v, value flags) /* ML */
{
  long len;
  value res;
  inside_join = 0;
  
  alloc_extern_block();
  len = extern_value(v, flags);
  res = alloc_string(len);
  bcopy(extern_block, String_val(res), len);
  stat_free(extern_block);

  return res;
}

value output_value_to_buffer(value buf, value ofs, value len, value v, value flags) /* ML */
{
  long len_res;

  inside_join = 0;
  extern_block = &Byte(buf, Long_val(ofs));
  extern_limit = extern_block + Long_val(len);
  extern_ptr = extern_block;
  extern_block_malloced = 0;
  len_res = extern_value(v, flags);
  
  return Val_long(len_res);
}

/* join

On complique un peu: 2 niveaux de caches
niveau 1: donnees a ne jamais envoyer (space_id local et remote par exemple)
niveau 2: stubs deja envoyes avec timeout

 */
int join_marshal_init = 1;
static void (*prev_scan_roots_hook) (scanning_action);

static void marshal_scan_roots(scanning_action action)
{
  int i;
  value *sp = (value*)saved_pointers;
  for(i=0;i<nbr_saved_pointers;i++){
    action(*sp,sp);
    sp++;
  }
  /* Hook */
  if (prev_scan_roots_hook != NULL) (*prev_scan_roots_hook)(action);
}

value join_marshal_message(value v,value cache) /* ML */
{
  long len,i;
  value res,array,str;
  
  if(join_marshal_init){    /* Initialize GC */
    prev_scan_roots_hook = scan_roots_hook;
    scan_roots_hook = marshal_scan_roots;
    join_marshal_init = 0;
  }
  nbr_saved_pointers = 0;
  alloc_extern_block();

  inside_join = 1;
  /* le cache */
  extern_cache = cache;
  cache_size = Wosize_val(cache);
  cache_number = Int_val(Field(cache,1)) >> CACHE_SIZE;
  next_cache_pos = Int_val(Field(cache,0));
  cache_left = cache_size - CACHE_OFFSET;

  len = extern_value(v,Val_int(0));
  
  Field(cache,0) = Val_int(next_cache_pos);
  
  inside_join = 0;
  if(nbr_saved_pointers)
    {
      i= nbr_saved_pointers;
       /* si un GC mineur ou une compaction arrive lors de cette allocation,
       on est perdu car les saved_pointers ne seront pas scannes.
       BUG A CORRIGER
       */
      array = alloc_shr(nbr_saved_pointers,0);
      while(i>0){
        i--;
        Assert(!Is_blue_hd(Hd_val(Field(saved_pointers,i))));
        Init(Field(array,i),Field(saved_pointers,i));
      }
      free_saved_pointers();
    }
  else
    array = Atom(0);
  
  Begin_root(array);
  str = alloc_string(len);

  bcopy(extern_block, String_val(str), len);
  stat_free(extern_block);
  
  Begin_roots2(str,array);
  res = alloc(2,0);
  End_roots();

  Field(res,0)=str;
  Field(res,1)=array;

  End_roots();
  return res;
}
 

