/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: interp.c,v 1.1 1999/09/30 13:05:06 lefessan Exp $ */

/* The bytecode interpreter */

#include <stdio.h>
#include "alloc.h"
#include "callback.h"
#include "debugger.h"
#include "fail.h"
#include "fix_code.h"
#include "instrtrace.h"
#include "instruct.h"
#include "interp.h"
#include "major_gc.h"
#include "memory.h"
#include "misc.h"
#include "mlvalues.h"
#include "prims.h"
#include "signals.h"
#include "stacks.h"
#include "str.h"
#include "join.h"
#include "common.h"


static opcode_t callback1_code[] = { ACC1, APPLY1, POP, 1, STOP };
static opcode_t callback2_code[] = { ACC2, APPLY2, POP, 1, STOP };
static opcode_t callback3_code[] = { ACC3, APPLY3, POP, 1, STOP };

#ifdef THREADED_CODE

static int callback_code_threaded = 0;

static void thread_callback(void)
{
  thread_code(callback1_code, sizeof(callback1_code));
  thread_code(callback2_code, sizeof(callback2_code));
  thread_code(callback3_code, sizeof(callback3_code));
  callback_code_threaded = 1;
}

#define Init_callback() if (!callback_code_threaded) thread_callback()

#else

#define Init_callback()

#endif


#ifdef NATIVE_CODE
#define Getenv(n) Field(env,NATIVE_CLOSURE_SIZE+(n))
#else
#define Getenv(n) Field(env,n)
#endif

#if macintosh
#include "rotatecursor.h"
extern int volatile have_to_interact;
#endif
/* Registers for the abstract machine:
        pc         the code pointer
        sp         the stack pointer (grows downward)
        accu       the accumulator
        env        heap-allocated environment
        trapsp     pointer to the current trap frame
        extra_args number of extra arguments provided by the caller

sp is a local copy of the global variable extern_sp. */

/* Instruction decoding */

/* #  define Instruct(name) lbl_##name: fprintf(stderr,#name " at %x\n",pc); */

#ifdef THREADED_CODE
#  define Instruct(name) lbl_##name:
#  if defined(ARCH_SIXTYFOUR) && !defined(ARCH_CODE32)
#    define Jumptbl_base ((char *) &&lbl_ACC0)
#  else
#    define Jumptbl_base ((char *) 0)
#    define jumptbl_base ((char *) 0)
#  endif
#    define Next goto *(void *)(jumptbl_base + *pc++)
#else
#  define Instruct(name) case name: 
#  define Next break
#endif

/* GC interface */
#define Setup_for_gc { sp -= 2; sp[0] = accu; sp[1] = env; extern_sp = sp; }
#define Restore_after_gc { accu = sp[0]; env = sp[1];  sp += 2; }
#define Setup_for_c_call { *--sp = env; extern_sp = sp; }
#define Restore_after_c_call { sp = extern_sp; env = *sp++;  }
#define Setup_for_asm_call { *--sp = env; *--sp = exception_traps;\
  extern_sp = sp;exception_traps = Val_int(0);}
#define Restore_after_asm_call { sp = extern_sp; exception_traps = *sp++;\
  env = *sp++;  sp++; }
/* Debugger interface */

#define Setup_for_debugger \
   { sp -= 4; \
     sp[0] = accu; sp[1] = (value)(pc - 1); \
     sp[2] = env; sp[3] = Val_long(extra_args); \
     extern_sp = sp; }
#define Restore_after_debugger { sp += 4; }

#ifdef THREADED_CODE
#define Restart_curr_instr \
  goto *(jumptable[saved_code[pc - 1 - start_code]])
#else
#define Restart_curr_instr \
  curr_instr = saved_code[pc - 1 - start_code]; \
  goto dispatch_instr
#endif

/* Register optimization.
   Some compilers underestimate the use of the local variables representing
   the abstract machine registers, and don't put them in hardware registers,
   which slows down the interpreter considerably.
   For GCC, I have hand-assigned hardware registers for several architectures.
*/

#if defined(__GNUC__)
#ifdef __mips__
#define PC_REG asm("$16")
#define SP_REG asm("$17")
#define ACCU_REG asm("$18")
#endif
#ifdef __sparc__
#define PC_REG asm("%l0")
#define SP_REG asm("%l1")
#define ACCU_REG asm("%l2")
#endif
#ifdef __alpha__
#define PC_REG asm("$9")
#define SP_REG asm("$10")
#define ACCU_REG asm("$11")
#define JUMPTBL_BASE_REG asm("$12")
#endif
#ifdef __i386__
#define PC_REG asm("%esi")
#define SP_REG asm("%edi")
#define ACCU_REG asm("%ebx")
#endif
#if defined(PPC) || defined(_POWER) || defined(_IBMR2)
#define PC_REG asm("26")
#define SP_REG asm("27")
#define ACCU_REG asm("28")
#endif
#ifdef __hppa__
#define PC_REG asm("%r18")
#define SP_REG asm("%r17")
#define ACCU_REG asm("%r16")
#endif
#ifdef __mc68000__
#define PC_REG asm("a5")
#define SP_REG asm("a4")
#define ACCU_REG asm("d7")
#endif
#endif

/* The interpreter itself */

value exception_traps = Val_int(0);
extern value native_caller1,native_caller2,native_caller3;
extern value native_closure;
int yield_thread = 0;
value preemption_handler = Val_unit;
int callback_depth = -1;

value interprete(code_t prog, asize_t prog_size)
{
#ifdef PC_REG
  register code_t pc PC_REG;
  register value * sp SP_REG;
  register value accu ACCU_REG;
#else
  register code_t pc;
  register value * sp;
  register value accu;
#endif
#if defined(THREADED_CODE) && defined(ARCH_SIXTYFOUR) && !defined(ARCH_CODE32)
#ifdef JUMPTBL_BASE_REG
  register char * jumptbl_base JUMPTBL_BASE_REG;
#else
  register char * jumptbl_base;
#endif
#endif
  value env;
  long extra_args;
  struct longjmp_buffer * initial_external_raise;
  int initial_sp_offset;
  struct caml__roots_block *initial_local_roots;
  int initial_callback_depth;
  struct longjmp_buffer raise_buf;
  value * modify_dest, modify_newval;
#ifndef THREADED_CODE
  opcode_t curr_instr;
#endif

#ifdef THREADED_CODE
  static void * jumptable[] = {
#    include "jumptbl.h"
  };
#endif

  if (prog == NULL) {           /* Interpreter is initializing */
#ifdef THREADED_CODE
    instr_table = (char **) jumptable; 
    instr_base = Jumptbl_base;
#endif
    Init_callback();
    

    return Val_unit;
  }

#if defined(THREADED_CODE) && defined(ARCH_SIXTYFOUR) && !defined(ARCH_CODE32)
  jumptbl_base = Jumptbl_base;
#endif
  initial_local_roots = local_roots;
  initial_sp_offset = (char *) stack_high - (char *) extern_sp;
  initial_external_raise = external_raise;
  initial_callback_depth = ++callback_depth;

  if (sigsetjmp(raise_buf.buf, 1)) {
    local_roots = initial_local_roots;
    callback_depth = initial_callback_depth;
    accu = exn_bucket;
    goto raise_exception;
  }
  external_raise = &raise_buf;

  sp = extern_sp;
  pc = prog;
  extra_args = 0;
  env = Atom(0);
  accu = Val_int(0);

  /*  DPRINT1("Programme %x\n",pc); */

#ifdef THREADED_CODE
  goto *(void *)(jumptbl_base + *pc++); /* Jump to the first instruction */
#else
  while(1) {
#ifdef DEBUG
    if (icount-- == 0) stop_here ();
    if (trace_flag) disasm_instr(pc);
    Assert(sp >= stack_low);
    Assert(sp <= stack_high);
#endif
    curr_instr = *pc++;
  dispatch_instr:
    switch(curr_instr) {
#endif

/* Basic stack operations */

    Instruct(ACC0)
      accu = sp[0]; Next;
    Instruct(ACC1)
      accu = sp[1]; Next;
    Instruct(ACC2)
      accu = sp[2]; Next;
    Instruct(ACC3)
      accu = sp[3]; Next;
    Instruct(ACC4)
      accu = sp[4]; Next;
    Instruct(ACC5)
      accu = sp[5]; Next;
    Instruct(ACC6)
      accu = sp[6]; Next;
    Instruct(ACC7)
      accu = sp[7]; Next;

    Instruct(PUSH) Instruct(PUSHACC0)
      *--sp = accu; Next;
    Instruct(PUSHACC1)
      *--sp = accu; accu = sp[1]; Next;
    Instruct(PUSHACC2)
      *--sp = accu; accu = sp[2]; Next;
    Instruct(PUSHACC3)
      *--sp = accu; accu = sp[3]; Next;
    Instruct(PUSHACC4)
      *--sp = accu; accu = sp[4]; Next;
    Instruct(PUSHACC5)
      *--sp = accu; accu = sp[5]; Next;
    Instruct(PUSHACC6)
      *--sp = accu; accu = sp[6]; Next;
    Instruct(PUSHACC7)
      *--sp = accu; accu = sp[7]; Next;

    Instruct(PUSHACC)
      *--sp = accu;
      /* Fallthrough */
    Instruct(ACC)
      accu = sp[*pc++];
      Next;

    Instruct(POP)
      sp += *pc++;
    /*      DPRINT1("POP to %x\n",sp); */
      Next;
    Instruct(ASSIGN)
      sp[*pc++] = accu;
      accu = Val_unit;
      Next;

/* Access in heap-allocated environment */

    Instruct(ENVACC1)
      accu = Getenv(1); Next; 
    Instruct(ENVACC2)
      accu = Getenv(2); Next; 
    Instruct(ENVACC3)
      accu = Getenv(3); Next; 
    Instruct(ENVACC4)
      accu = Getenv(4); Next; 

    Instruct(PUSHENVACC1)
      *--sp = accu; accu = Getenv(1); Next; 
    Instruct(PUSHENVACC2)
      *--sp = accu; accu = Getenv(2); Next; 
    Instruct(PUSHENVACC3)
      *--sp = accu; accu = Getenv(3); Next; 
    Instruct(PUSHENVACC4)
      *--sp = accu; accu = Getenv(4); Next; 

    Instruct(PUSHENVACC)
      *--sp = accu;
      /* Fallthrough */
    Instruct(ENVACC)
      accu = Getenv(*pc++); 
      Next;

/* Function application */

    Instruct(PUSH_RETADDR) {
      sp -= 3;
      sp[0] = (value) (pc + *pc);
      sp[1] = env;
      sp[2] = Val_long(extra_args);
      pc++;
      Next;
    }
    Instruct(PUSHGETGLOBALFIELDAPPLY_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELDAPPLY,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELDAPPLY)
      /*DPRINT("PUSHGETGLOBALFIELDAPPLY\n");*/
      *--sp = accu;
      accu = Field(global_data, *pc);
      pc++;
      accu = Field(accu, *pc);
      pc++;
      /* Fallthrough */
      
    Instruct(APPLY)
      /*DPRINT4("%d %d %d - %d\n",pc[-3],pc[-2],pc[-1],pc[1]);*/
      /*DPRINT2("extra args : %d at %x\n",*pc - 1,pc);*/
      extra_args = *pc - 1;
#ifdef NATIVE_CODE
    next_apply:
    if(Tag_val(accu) == Closure_byte_tag){
#endif        
      pc = Code_val(accu);
      env = accu;
      /* fall-through check-stacks */
#ifdef NATIVE_CODE
    } else {
      Setup_for_asm_call;
      accu = callback3(native_caller1,accu,sp[2],sp[0]);
      Restore_after_asm_call;
      
      if(extra_args-->0) goto next_apply; /* encore un argument */
      pc = (code_t)(sp[0]); /* sinon on retire l'adresse de retour */
      env = sp[1];
      
      extra_args = Long_val(sp[2]);
      sp+=3;
      Next;
    }
#endif
    check_stacks: /* Stack checks */
    if (sp < stack_threshold) {
      /* May be Something to do ...
      if(something_to_do){
        stack_threshold = something_to_do;
        if (sp < stack_threshold) {
        extern_sp = sp;
        realloc_stack();
        sp = extern_sp;
        }
        goto process_signal;
      } else {
        */
      extern_sp = sp;
      realloc_stack();
      sp = extern_sp;
    }
    /* Fall through CHECK_SIGNALS */
/* Signal handling:
  Signals are handled in the Garbage collector. Here, we only test for
  the something_to_do used by the scheduler to preempt bytecode threads. 
 */

    Instruct(CHECK_SIGNALS)    /* accu not preserved */
      if (something_to_do) goto process_signal;
      Next;

    process_signal:
      something_to_do = 0;
      if (force_major_slice){
        Setup_for_gc;
        minor_collection ();
        Restore_after_gc;
      }
      if(yield_thread && callback_depth==0){
        yield_thread = 0;
        /* Push a return frame to the current code location */
        sp -= 4;
        sp[0] = Val_unit;
        sp[1] = (value) pc;
        sp[2] = env;
        sp[3] = Val_long(extra_args);
        /* Branch to the signal handler */
        env = preemption_handler;
        pc = Code_val(env);
        extra_args = 0;
      }
      Next;
    
    Instruct(PUSHGETGLOBALFIELDAPPLY1_UNRESOLVED)
      /*      DPRINT("PUSHGETGLOBALFIELDAPPLY1_UNRESOLVED");*/
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELDAPPLY1,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELDAPPLY1)
      /*DPRINT("PUSHGETGLOBALFIELDAPPLY\n");*/
      *--sp = accu;
    /*
      DPRINT1("PUSHGETGLOBALFIELDAPPLY1 : PUSH GETGLOBAL(%d) FIELD(%d)\n",pc[0],pc[1]);
      */
      accu = Field(global_data, *pc);
      pc++;
      accu = Field(accu, *pc);
      pc++;
      /* Fallthrough */
      Instruct(APPLY1)
#ifdef NATIVE_CODE
        next_apply1:
#endif
      {
        value arg1 = sp[0];
          
#ifdef NATIVE_CODE
        if(Tag_val(accu) == Closure_byte_tag){
#endif
          sp -= 3;
          sp[0] = arg1;
          sp[1] = (value)pc;
          sp[2] = env;
          sp[3] = Val_long(extra_args);
          pc = Code_val(accu);
          env = accu;
          extra_args = 0;
          goto check_stacks;
#ifdef NATIVE_CODE
        } else {
          Setup_for_asm_call;
          accu = callback3(native_caller1,accu,arg1,sp[0]);
          Restore_after_asm_call;
        }
#endif
      }
      Next;
      
    Instruct(PUSHGETGLOBALFIELDAPPLY2_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELDAPPLY2,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELDAPPLY2)
      /*DPRINT("PUSHGETGLOBALFIELDAPPLY2\n");*/
      *--sp = accu;
      accu = Field(global_data, *pc++);
      accu = Field(accu, *pc++);
      /* Fallthrough */
      /* APPLY2 = PUSH_RETADDR + CALL 2args
        On doit le faire comme 
        PUSH_RETADDR

      */
      Instruct(APPLY2)
#ifdef NATIVE_CODE
      next_apply2:
#endif
    {
      value arg1 = sp[0];
#ifdef NATIVE_CODE
      if(Tag_val(accu) == Closure_byte_tag){
#endif
        {
          value arg2 = sp[1];
          sp -= 3;
          sp[0] = arg1;
          sp[1] = arg2;
          sp[2] = (value)pc;
          sp[3] = env;
          sp[4] = Val_long(extra_args);
          pc = Code_val(accu);
          env = accu;
          extra_args = 1;
          goto check_stacks;
        }
#ifdef NATIVE_CODE
      } else {
        /* on applique une premiere fois */
        Setup_for_asm_call;
        accu = callback3(native_caller1,accu,arg1,sp[0]);
        Restore_after_asm_call;
        
        /* on applique une seconde fois */
        goto next_apply1;
      }
#endif
    }
    
    Instruct(PUSHGETGLOBALFIELDAPPLY3_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELDAPPLY3,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELDAPPLY3)
      /*DPRINT("PUSHGETGLOBALFIELDAPPLY3\n");*/
      *--sp = accu;
      accu = Field(global_data, *pc++);
      accu = Field(accu, *pc++);
      /* Fallthrough */
    Instruct(APPLY3) 
#ifdef NATIVE_CODE
      next_apply3:
#endif
      {
        value arg1 = sp[0];
#ifdef NATIVE_CODE
        if(Tag_val(accu) == Closure_byte_tag){
#endif
          {
            value arg2 = sp[1];
            value arg3 = sp[2];
            sp -= 3;
            sp[0] = arg1;
            sp[1] = arg2;
            sp[2] = arg3;
            sp[3] = (value)pc;
            sp[4] = env;
            sp[5] = Val_long(extra_args);
            pc = Code_val(accu);
            env = accu;
            extra_args = 2;
            goto check_stacks;
          }
#ifdef NATIVE_CODE
        } else {

          Setup_for_asm_call;
          accu = callback3(native_caller1,accu,arg1,sp[0]);
          Restore_after_asm_call;

          goto next_apply2;
        }
#endif
      }

    Instruct(PUSHGETGLOBALFIELDAPPTERM_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELDAPPTERM,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELDAPPTERM)
      /*DPRINT("PUSHGETGLOBALFIELDAPPTERM\n");*/
      *--sp = accu;
      accu = Field(global_data, *pc++);
      accu = Field(accu, *pc++);
      /* Fallthrough */
    Instruct(APPTERM) 
      {
        int nargs = *pc++;
        int slotsize = *pc;
        value * newsp;
        int i;
        /* Slide the nargs bottom words of the current frame to the top
        of the frame, and discard the remainder of the frame */
        newsp = sp + slotsize - nargs;
        for (i = nargs - 1; i >= 0; i--) newsp[i] = sp[i];
        sp = newsp;
        extra_args += nargs - 1;
#ifdef NATIVE_CODE
        goto next_apply;
#else
	pc = Code_val(accu);
	env = accu;
	goto check_stacks;
#endif
      }
    
    Instruct(PUSHGETGLOBALFIELDAPPTERM1_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELDAPPTERM1,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELDAPPTERM1)
      /*DPRINT("PUSHGETGLOBALFIELDAPPTERM1\n");*/
      *--sp = accu;
      accu = Field(global_data, *pc++);
      accu = Field(accu, *pc++);
      /* Fallthrough */
    Instruct(APPTERM1) 
      {
        value arg1 = sp[0];
        sp = sp + *pc - 1;
        sp[0] = arg1;
#ifdef NATIVE_CODE
        goto next_apply;
#else
	pc = Code_val(accu);
	env = accu;
	goto check_stacks;
#endif
      }
    
    Instruct(PUSHGETGLOBALFIELDAPPTERM2_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELDAPPTERM2,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELDAPPTERM2)
      *--sp = accu;
      accu = Field(global_data, *pc++);
      accu = Field(accu, *pc++);
      /* Fallthrough */
    Instruct(APPTERM2)
      {
        value arg1 = sp[0];
        value arg2 = sp[1];
        sp = sp + *pc - 2;
        sp[0] = arg1;
        sp[1] = arg2;
#ifdef NATIVE_CODE
        extra_args += 1;
        goto next_apply;
#else        
      pc = Code_val(accu);
      env = accu;
      extra_args += 1;
      goto check_stacks;
#endif
      }

    Instruct(PUSHGETGLOBALFIELDAPPTERM3_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELDAPPTERM3,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELDAPPTERM3)
      *--sp = accu;
      accu = Field(global_data, *pc++);
      accu = Field(accu, *pc++);
      /* Fallthrough */
    Instruct(APPTERM3)
      {
        value arg1 = sp[0];
        value arg2 = sp[1];
        value arg3 = sp[2];
        sp = sp + *pc - 3;
        sp[0] = arg1;
        sp[1] = arg2;
        sp[2] = arg3;
#ifdef NATIVE_CODE        
        extra_args += 2;
        goto next_apply;
#else
	pc = Code_val(accu);
	env = accu;
	extra_args += 2;
	goto check_stacks;
#endif
      }
    
    Instruct(RETURN) {
      sp += *pc++;
      if (extra_args > 0) { /* on doit appliquer a nouveau */
        extra_args--;
#ifdef NATIVE_CODE        
        goto next_apply;
#else
        pc = Code_val(accu);
        env = accu;
#endif
      } else {
        pc = (code_t)(sp[0]);
        env = sp[1];
        extra_args = Long_val(sp[2]);
        sp += 3;
      }
      Next;
    }

    Instruct(RESTART) {
      int num_args,i;
#ifdef NATIVE_CODE
      num_args = Wosize_val(env) - 2 - NATIVE_CLOSURE_SIZE; 
#else
      num_args = Wosize_val(env) - 2; 
#endif
      sp -= num_args;
      for (i = 0; i < num_args; i++) sp[i] = Getenv(i+2); 
      env = Getenv(1); 
      extra_args += num_args;
      Next;
    }

    Instruct(GRAB) {
      int required = *pc++;
      if (extra_args >= required) {
        extra_args -= required;
      } else {
        mlsize_t num_args, i;
        num_args = 1 + extra_args; /* arg1 + extra args */
#ifdef NATIVE_CODE
        Alloc_small(accu,num_args+2+NATIVE_CLOSURE_SIZE, Closure_byte_tag);
        Field(accu, 1+NATIVE_CLOSURE_SIZE) = env;
        for (i = 0; i < num_args; i++) 
          Field(accu,i+2+NATIVE_CLOSURE_SIZE) = sp[i];
        for (i = 0; i < NATIVE_CLOSURE_SIZE; i++) 
          Field(accu, i) =  Field(native_closure,i);
        Field(accu,BYTE_CLOSURE_POS) = accu;
#else
        Alloc_small(accu,num_args+2, Closure_byte_tag);
        Field(accu, 1) = env;
        for (i = 0; i < num_args; i++) 
          Field(accu,i+2) = sp[i];
#endif
        Code_val(accu) = pc - 3; /* Point to the preceding RESTART instr. */
        sp += num_args;
        pc = (code_t)(sp[0]);
        env = sp[1];
        
        extra_args = Long_val(sp[2]);
        sp += 3;
      }
      Next;
    }

    Instruct(CLOSURE) {
      int nvars = *pc++;
      int i;
      if (nvars > 0) *--sp = accu;
#ifdef NATIVE_CODE
      Alloc_small(accu, 1 + nvars + NATIVE_CLOSURE_SIZE, Closure_byte_tag);
      Code_val(accu) = pc + *pc;
      for (i = 0; i < nvars; i++) 
	Field(accu, i + 1 + NATIVE_CLOSURE_SIZE) = sp[i];
      for (i = 0; i < NATIVE_CLOSURE_SIZE; i++) 
        Field(accu, i) =  Field(native_closure,i);
      Field(accu,BYTE_CLOSURE_POS) = accu;
#else
      Alloc_small(accu, 1 + nvars, Closure_byte_tag);
      Code_val(accu) = pc + *pc;
      for (i = 0; i < nvars; i++) Field(accu, i + 1) = sp[i];
#endif
      sp += nvars;
      pc++;
      Next;
    }

    Instruct(CLOSUREREC) {
      int nvars = *pc++;
      int i;
      if (nvars > 0) *--sp = accu;
#ifdef NATIVE_CODE
      Alloc_small(accu, 2 + nvars + NATIVE_CLOSURE_SIZE, Closure_byte_tag);
      Code_val(accu) = pc + *pc;
      Field(accu, 1 + NATIVE_CLOSURE_SIZE) = Val_int(0);
      for (i = 0; i < nvars; i++) 
	Field(accu, i + 2 + NATIVE_CLOSURE_SIZE) = sp[i];
      /* boxed closure for native call */
      for (i = 0; i < NATIVE_CLOSURE_SIZE; i++) 
        Field(accu, i ) = Field(native_closure,i);
      Field(accu,BYTE_CLOSURE_POS) =  accu;
      modify(&Field(accu, 1 + NATIVE_CLOSURE_SIZE), accu);
#else
      Alloc_small(accu, 2 + nvars, Closure_byte_tag);
      Code_val(accu) = pc + *pc;
      Field(accu, 1) = Val_int(0);
      for (i = 0; i < nvars; i++) 
	Field(accu, i + 2) = sp[i];
      modify(&Field(accu, 1), accu);
#endif
      sp += nvars;
      pc++;

      Next;
    }

    Instruct(PUSHGETGLOBAL_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBAL,1);
      Restore_after_c_call;
      /* Fallthrough */      
    Instruct(PUSHGETGLOBAL)
      *--sp = accu;
      /* Fallthrough */
    Instruct(GETGLOBAL)
      accu = Field(global_data, *pc);
      pc++;
      Next;

    Instruct(PUSHGETGLOBALFIELD_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc++,PUSHGETGLOBALFIELD,1);
      Restore_after_c_call;
      /* Fallthrough */
    Instruct(PUSHGETGLOBALFIELD)
      *--sp = accu;
      /* Fallthrough */
    Instruct(GETGLOBALFIELD) {
      accu = Field(global_data, *pc);
      pc++;
      accu = Field(accu, *pc);
      pc++;
      Next;
    }

    Instruct(SETMODULE)
      /*      DPRINT1("SET local %s\n",
	      String_val(Field(Field(module_table,*pc),MODULE_NAME)));*/
      modify(&Field(Field(module_table,pc[0]),INTERFACE),Val_int(pc[1]));
      pc++;
      /*      DPRINT1("in GLOBAL %d\n",*pc);*/
      /* Fallthrough */
    Instruct(SETGLOBAL)
      modify(&Field(global_data, *pc), accu);
      accu = Val_unit;
      pc++;
      Next;

/* Allocation of blocks */

    Instruct(PUSHATOM0)
      *--sp = accu;
      /* Fallthrough */
    Instruct(ATOM0)
      accu = Atom(0); Next;

    Instruct(PUSHATOM)
      *--sp = accu;
      /* Fallthrough */
    Instruct(ATOM)
      accu = Atom(*pc++); Next;

    Instruct(MAKEBLOCK) {
      mlsize_t wosize = *pc++;
      tag_t tag = *pc++;
      mlsize_t i;
      value block;
      Alloc_small(block, wosize, tag);
      Field(block, 0) = accu;
      for (i = 1; i < wosize; i++) Field(block, i) = *sp++;
      accu = block;
      Next;
    }
    Instruct(MAKEBLOCK1) {
      tag_t tag = *pc++;
      value block;
      Alloc_small(block, 1, tag);
      Field(block, 0) = accu;
      accu = block;
      Next;
    }
    Instruct(MAKEBLOCK2) {
      tag_t tag = *pc++;
      value block;
      Alloc_small(block, 2, tag);
      Field(block, 0) = accu;
      Field(block, 1) = sp[0];
      sp += 1;
      accu = block;
      Next;
    }
    Instruct(MAKEBLOCK3) {
      tag_t tag = *pc++;
      value block;
      Alloc_small(block, 3, tag);
      Field(block, 0) = accu;
      Field(block, 1) = sp[0];
      Field(block, 2) = sp[1];
      sp += 2;
      accu = block;
      Next;
    }
    Instruct(MAKEFLOATBLOCK) {
      mlsize_t size = *pc++;
      mlsize_t i;
      value block;
      Alloc_small(block, size * Double_wosize, Double_array_tag);
      Store_double_field(block, 0, Double_val(accu));
      for (i = 1; i < size; i++){
        Store_double_field(block, i, Double_val(*sp));
        ++ sp;
      }
      accu = block;
      Next;
    }

/* Access to components of blocks */

    Instruct(GETFIELD0)
      accu = Field(accu, 0); Next;
    Instruct(GETFIELD1)
      accu = Field(accu, 1); Next;
    Instruct(GETFIELD2)
      accu = Field(accu, 2); Next;
    Instruct(GETFIELD3)
      accu = Field(accu, 3); Next;
    Instruct(GETFIELD)
      accu = Field(accu, *pc); pc++; Next;
    Instruct(GETFLOATFIELD) {
      double d = Double_field(accu, *pc);
      Alloc_small(accu, Double_wosize, Double_tag);
      Store_double_val(accu, d);
      pc++;
      Next;
    }

    Instruct(SETFIELD0)
      modify_dest = &Field(accu, 0);
      modify_newval = *sp++;
    modify:
      Modify(modify_dest, modify_newval);
      accu = Val_unit;
      Next;
    Instruct(SETFIELD1)
      modify_dest = &Field(accu, 1);
      modify_newval = *sp++;
      goto modify;
    Instruct(SETFIELD2)
      modify_dest = &Field(accu, 2);
      modify_newval = *sp++;
      goto modify;
    Instruct(SETFIELD3)
      modify_dest = &Field(accu, 3);
      modify_newval = *sp++;
      goto modify;
    Instruct(SETFIELD)
      modify_dest = &Field(accu, *pc);
      pc++;
      modify_newval = *sp++;
      goto modify;
    Instruct(SETFLOATFIELD)
      Store_double_field(accu, *pc, Double_val(*sp));
      accu = Val_unit;
      sp++;
      pc++;
      Next;

/* For recursive definitions */

    Instruct(DUMMY) {
#ifdef NATIVE_CODE
      int size = NATIVE_CLOSURE_SIZE + *pc++;
#else
      int size = *pc++;
#endif
      Alloc_small(accu, size, 0);
      while (size--) Field(accu, size) = Val_long(0);
      Next;
    }
    Instruct(UPDATE) {
      value newval = *sp++;
      mlsize_t size, n;
      size = Wosize_val(newval);
      Assert(size == Wosize_val(accu));
      Tag_val(accu) = Tag_val(newval);
      for (n = 0; n < size; n++) {
        modify(&Field(accu, n), Field(newval, n));
      }
      accu = Val_unit;
      Next;
    }

/* Array operations */

    Instruct(VECTLENGTH) {
      mlsize_t size = Wosize_val(accu);
      if (Tag_val(accu) == Double_array_tag) size = size / Double_wosize;
      accu = Val_long(size);
      Next;
    }
    Instruct(GETVECTITEM)
      accu = Field(accu, Long_val(sp[0]));
      sp += 1;
      Next;
    Instruct(SETVECTITEM)
      modify_dest = &Field(accu, Long_val(sp[0]));
      modify_newval = sp[1];
      sp += 2;
      goto modify;

/* String operations */

    Instruct(GETSTRINGCHAR)
      accu = Val_int(Byte_u(accu, Long_val(sp[0])));
      sp += 1;
      Next;
    Instruct(SETSTRINGCHAR)
      Byte_u(accu, Long_val(sp[0])) = Int_val(sp[1]);
      sp += 2;
      accu = Val_unit;
      Next;

/* Branches and conditional branches */

    Instruct(BRANCH)
      pc += *pc;
      Next;
    Instruct(BRANCHIF)
      if (accu != Val_false) pc += *pc; else pc++;
      Next;
    Instruct(BRANCHIFNOT)
      if (accu == Val_false) pc += *pc; else pc++;
      Next;
    Instruct(SWITCH) {
      uint32 sizes = *pc++;
      if (Is_block(accu)) {
        long index = Tag_val(accu);
        Assert(index >= 0 && index < (sizes >> 16));
        pc += pc[(sizes & 0xFFFF) + index];
      } else {
        long index = Long_val(accu);
        if ((unsigned long) index < (sizes & 0xFFFF))
          pc += pc[index];
        else
          pc += (sizes & 0xFFFF) + (sizes >> 16);
      }
      Next;
    }
    Instruct(BOOLNOT)
      accu = Val_not(accu);
      Next;

/* Exceptions */

    Instruct(PUSHTRAP)
      sp -= 4;
      Trap_pc(sp) = pc + *pc;
      Trap_link(sp) = trapsp;
      exception_traps+=2;
      sp[2] = env;
      sp[3] = Val_long(extra_args);
      trapsp = sp;
      pc++;
      Next;

    Instruct(POPTRAP)
      /* We should check here if a signal is pending, to preserve the
         semantics of the program w.r.t. exceptions. Unfortunately,
         process_signal destroys the accumulator, and there is no
         convenient way to preserve it... */
      trapsp = Trap_link(sp);
      exception_traps-=2;
      sp += 4;
      Next;

    Instruct(RAISE)
    raise_exception:
      if (trapsp >= trap_barrier) debugger(TRAP_BARRIER);
      sp = trapsp;
      if ((char *) sp >= (char *) stack_high - initial_sp_offset) {
        exn_bucket = accu;
        external_raise = initial_external_raise;
        siglongjmp(external_raise->buf, 1);
      }
      pc = Trap_pc(sp);
      trapsp = Trap_link(sp);
      exception_traps-=2;
      env = sp[2];
      
      extra_args = Long_val(sp[3]);
      sp += 4;
      Next;

/* Calling C functions */

    Instruct(C_CALL1_UNRESOLVED) 
      Setup_for_c_call;
      resolv_prim(pc++,C_CALL1,1);
      Restore_after_c_call;
      /*      DPRINT1("Resolved by %d \n",pc[0]);*/
    /* Fallthrough */
    Instruct(C_CALL1)
      Setup_for_c_call;
      DPRINT1("Call %s\n",names_of_cprim[*pc]);
      accu = cprim[*pc](accu);
      Restore_after_c_call;
      pc++;
      Next;

    Instruct(C_CALL2_UNRESOLVED)
      Setup_for_c_call;
      resolv_prim(pc++,C_CALL2,1);
      Restore_after_c_call;
    /* Fallthrough */
    Instruct(C_CALL2)
      Setup_for_c_call;
    DPRINT1("Call %s\n",names_of_cprim[*pc]);
      accu = cprim[*pc](accu, sp[1]);
      Restore_after_c_call;
      sp += 1;
      pc++;
      Next;

    Instruct(C_CALL3_UNRESOLVED) 
      Setup_for_c_call;
      resolv_prim(pc++,C_CALL3,1);
      Restore_after_c_call;
    /* Fallthrough */
    Instruct(C_CALL3)
      Setup_for_c_call;
    DPRINT1("Call %s\n",names_of_cprim[*pc]);
      accu = cprim[*pc](accu, sp[1], sp[2]);
      Restore_after_c_call;
      sp += 2;
      pc++;
      Next;

    Instruct(C_CALL4_UNRESOLVED) 
      Setup_for_c_call;
      resolv_prim(pc++,C_CALL4,1);
      Restore_after_c_call;
    /* Fallthrough */
    Instruct(C_CALL4)
      Setup_for_c_call;
    DPRINT1("Call %s\n",names_of_cprim[*pc]);
      accu = cprim[*pc](accu, sp[1], sp[2], sp[3]);
      Restore_after_c_call;
      sp += 3;
      pc++;
      Next;

    Instruct(C_CALL5_UNRESOLVED) 
      Setup_for_c_call;
      resolv_prim(pc++,C_CALL5,1);
      Restore_after_c_call;
    /* Fallthrough */
    Instruct(C_CALL5)
      Setup_for_c_call;
    DPRINT1("Call %s\n",names_of_cprim[*pc]);
      accu = cprim[*pc](accu, sp[1], sp[2], sp[3], sp[4]);
      Restore_after_c_call;
      sp += 4;
      pc++;
      Next;
 
    Instruct(C_CALLN_UNRESOLVED) 
      Setup_for_c_call;
      resolv_prim(pc++,C_CALLN,2);
      Restore_after_c_call;
    /* Fallthrough */
    Instruct(C_CALLN) {
      int nargs = *pc++;
      *--sp = accu;
      Setup_for_c_call;
      DPRINT1("Call %s\n",names_of_cprim[*pc]);
      accu = cprim[*pc](sp + 1, nargs);
      Restore_after_c_call;
      sp += nargs;
      pc++;
      Next;
    }

/* Integer constants */

    Instruct(CONST0)
      accu = Val_int(0); Next;
    Instruct(CONST1)
      accu = Val_int(1); Next;
    Instruct(CONST2)
      accu = Val_int(2); Next;
    Instruct(CONST3)
      accu = Val_int(3); Next;

    Instruct(PUSHCONST0)
      *--sp = accu; accu = Val_int(0); Next;
    Instruct(PUSHCONST1)
      *--sp = accu; accu = Val_int(1); Next;
    Instruct(PUSHCONST2)
      *--sp = accu; accu = Val_int(2); Next;
    Instruct(PUSHCONST3)
      *--sp = accu; accu = Val_int(3); Next;

    Instruct(PUSHCONSTINT)
      *--sp = accu;
      /* Fallthrough */
    Instruct(CONSTINT)
      accu = Val_int(*pc);
      pc++;
      Next;

/* Integer arithmetic */

    Instruct(NEGINT)
      accu = (value)(2 - (long)accu); Next;
    Instruct(ADDINT)
      accu = (value)((long) accu + (long) *sp++ - 1); Next;
    Instruct(SUBINT)
      accu = (value)((long) accu - (long) *sp++ + 1); Next;
    Instruct(MULINT)
      accu = Val_long(Long_val(accu) * Long_val(*sp++)); Next;

    Instruct(DIVINT) {
      long divisor = Long_val(*sp++);
      if (divisor == 0) { Setup_for_c_call; raise_zero_divide(); }
      accu = Val_long(Long_val(accu) / divisor);
      Next;
    }
    Instruct(MODINT) {
      long divisor = Long_val(*sp++);
      if (divisor == 0) { Setup_for_c_call; raise_zero_divide(); }
      accu = Val_long(Long_val(accu) % divisor);
      Next;
    }
    Instruct(ANDINT)
      accu = (value)((long) accu & (long) *sp++); Next;
    Instruct(ORINT)
      accu = (value)((long) accu | (long) *sp++); Next;
    Instruct(XORINT)
      accu = (value)(((long) accu ^ (long) *sp++) | 1); Next;
    Instruct(LSLINT)
      accu = (value)((((long) accu - 1) << Long_val(*sp++)) + 1); Next;
    Instruct(LSRINT)
      accu = (value)((((unsigned long) accu - 1) >> Long_val(*sp++)) | 1);
      Next;
    Instruct(ASRINT)
      accu = (value)((((long) accu - 1) >> Long_val(*sp++)) | 1); Next;

#define Integer_comparison(opname,tst) \
    Instruct(opname) \
      accu = Val_int((long) accu tst (long) *sp++); Next;

    Integer_comparison(EQ, ==)
    Integer_comparison(NEQ, !=)
    Integer_comparison(LTINT, <)
    Integer_comparison(LEINT, <=)
    Integer_comparison(GTINT, >)
    Integer_comparison(GEINT, >=)

    Instruct(OFFSETINT)
      accu += *pc << 1;
      pc++;
      Next;
    Instruct(OFFSETREF)
      Field(accu, 0) += *pc << 1;
      accu = Val_unit;
      pc++;
      Next;

/* Object-oriented operations */

#define Lookup(obj, lab, tmv) \
  Field (Field (Field (obj, tmv), ((lab) >> 16) / sizeof (value)), \
         ((lab) / sizeof (value)) & 0xFF)

    Instruct(GETMETHOD_UNRESOLVED)
      *--sp = accu;
      Setup_for_c_call;
      accu = resolv_method(pc);
      Restore_after_c_call;
      pc++; /* passe arg de  CONSTINT */
      pc++; /* passe GETMETHOD */
      /* Fall-through */
    Instruct(GETMETHOD)
      accu = Lookup(sp[0], accu,5);
      Next;

    Instruct(GETANCMETHOD)
      accu = Lookup(sp[0], accu,6);
      Next;

/* Join extension */
    Instruct(NOP)
      Next;
	
    Instruct(GETGLOBAL_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc,GETGLOBAL,1);
      Restore_after_c_call;
      Next;
      
    Instruct(GETGLOBALFIELD_UNRESOLVED)
      Setup_for_c_call;
      resolv_interface(pc,GETGLOBALFIELD,1);
      Restore_after_c_call;
      Next;

    Instruct(BEGINMODULE)
      Setup_for_c_call;
      begin_module(pc,Field(global_data,*pc));
      Restore_after_c_call;
      pc++;
      Next;

    
    Instruct(PUSHGETLOCAL) 
      *--sp = accu;
      /* Fallthrough */
    Instruct(GETLOCAL)
      /*      DPRINT2("GETLOCAL %d %d \n",pc[0],pc[1]);*/
      accu = Field(Field(module_table,pc[0]),pc[1]+MODULE_DATA_SIZE);
      pc+=2;
      Next;

    Instruct(REPLACECODEPTR)
     /* Setup_for_c_call; */
      pc = resolv_codeptr(pc);
    /* Restore_after_c_call; */
      Getenv(0) = (value)pc;
      Next;

    Instruct(CODEPTR)
    /*Setup_for_c_call; */
      pc = resolv_codeptr(pc);
    /*Restore_after_c_call;*/
      Next;

    Instruct(MODULE_UNRESOLVED)
      Setup_for_c_call;
      resolv_module(pc++,MODULE,1);
      Restore_after_c_call;
      /* Fall-through */
    Instruct(MODULE)
      accu = Field(module_table,*pc++);
      Next;

    Instruct(SETLOCAL)
      /*      DPRINT1("SET local %s\n",
	      String_val(Field(Field(module_table,pc[0]),MODULE_NAME)));*/
      Field(Field(module_table,pc[0]),pc[1])=accu;
      pc+=2;
      Next;
      /* accu not modified since SETGLOBAL may follow */

/* Debugging and machine control */

    Instruct(STOP)
      external_raise = initial_external_raise;
      extern_sp = sp;
      callback_depth = initial_callback_depth - 1;
      /*      DPRINT1("Fin de Programme: %x\n",accu);*/
      return accu;

    Instruct(EVENT)
      if (--event_count == 0) {
        Setup_for_debugger;
        debugger(EVENT_COUNT);
        Restore_after_debugger;
      }
      Restart_curr_instr;

    Instruct(BREAK)
      Setup_for_debugger;
      debugger(BREAKPOINT);
      Restore_after_debugger;
      Restart_curr_instr;

#ifndef THREADED_CODE
    default:
      fatal_error_arg("Fatal error: bad opcode (%lx)\n",
                      (char *)(long)(*(pc-1)));
    }
  }
#endif
}

#ifdef NATIVE_CODE
value byte_callback(value closure, value arg)
#else
value callback(value closure, value arg)
#endif
{
  extern_sp -= 2;
  extern_sp[0] = arg;
  extern_sp[1] = closure;
  return interprete(callback1_code, sizeof(callback1_code));
}

#ifdef NATIVE_CODE
value byte_callback2(value closure, value arg1, value arg2)
#else
value callback2(value closure, value arg1, value arg2)
#endif
{
  extern_sp -= 3;
  extern_sp[0] = arg1;
  extern_sp[1] = arg2;
  extern_sp[2] = closure;
  return interprete(callback2_code, sizeof(callback2_code));
}

#ifdef NATIVE_CODE
value byte_callback3(value closure, value arg1, value arg2,value arg3)
#else
value callback3(value closure, value arg1, value arg2,value arg3)
#endif
{
  extern_sp -= 4;
  extern_sp[0] = arg1;
  extern_sp[1] = arg2;
  extern_sp[2] = arg3;
  extern_sp[3] = closure;
  return interprete(callback3_code, sizeof(callback3_code));
}

#ifdef NATIVE_CODE
/*
If we are in native code, all is fine. If we are in bytecode, we could need
to warn the native code callback that we have installed an exception handler.
This can only be done by calling an official bytecode-to-native handler.
*/
value callback(value closure, value arg) /* ML */
{
  if(exception_traps == Val_int(0))
    return asm_callback(closure,arg);
  {
    value traps = exception_traps;
    value res;
    
    exception_traps = Val_int(0);
    res = asm_callback3(native_caller1,closure,arg,traps);
    exception_traps = traps;
    return res;
  }
}

value callback2(value closure, value arg1,value arg2) /* ML */
{
  if(exception_traps == Val_int(0))
    return asm_callback2(closure,arg1,arg2);
  {
    value traps = exception_traps;
    value res;

    Begin_roots3(closure,arg1,arg2);
    res = alloc(2,0);
    Field(res,0) = arg1;
    Field(res,1) = arg2;
    End_roots();
    
    exception_traps = Val_int(0);
    res = asm_callback3(native_caller2,closure,res,traps);
    exception_traps = traps;
    return res;
  }
}

value callback3(value closure, value arg1,value arg2,value arg3) /* ML */
{
  if(exception_traps == Val_int(0))
    return  asm_callback3(closure,arg1,arg2,arg3);
  {
    value traps = exception_traps;
    value res;

    Begin_roots3(closure,arg1,arg2);
    res = alloc(3,0);
    Field(res,0) = arg1;
    Field(res,1) = arg2;
    Field(res,2) = arg3;
    End_roots();
    
    exception_traps = Val_int(0);
    res = asm_callback3(native_caller3,closure,res,traps);
    exception_traps = traps;
    return res;
  }  
}

value asm_callbyte(value byte_closure,value arg)
{
  return byte_callback(byte_closure,arg);
}

#endif
