#ifndef JOIN_H
#define JOIN_H

#include "mlvalues.h"


/* The thread descriptors */
typedef struct caml_thread    *caml_thread_t;

struct caml_thread {
  value ident;                  /* Unique id (for equality comparisons) */
  struct caml_thread * next;  /* Double linking of threads */
  struct caml_thread * prev;
  value * stack_low;            /* The execution stack for this thread */
  value * stack_high;
  value * stack_threshold;
  value * sp;
  value * trapsp;
  value exception_traps;
  value status;                 /* RUNNABLE, KILLED. etc (see below) */
  value readfds, writefds, exceptfds;
                /* Lists of file descriptors on which we're doing select() */
  value delay;                  /* Time until which this thread is blocked */
  value joining;                /* Thread we're trying to join */
  value waitpid;                /* PID of process we're waiting for */
  value retval;                 /* Value to return when thread resumes */

  value location;           /* Location to which this thread belongs */
  value stub;                   /* A stub for this thread */
  value reused;                 /* offset from the tail of the stack
				   to find the next frame */
  value sync;                   /* sync > 0 means that this thread is used 
				   for a synchronous call in the same 
				   location */
  value passe;              /* to minimize scheduler work */
};



/* Initial size of stack when a thread is created (4 Ko) */
#define Thread_stack_size (Stack_size / 4)

/* Max computation time before rescheduling, in microseconds (50ms) */
#define Thread_timeout 50000

#define RUNNABLE Val_int(0)
#define KILLED Val_int(1)
#define SUSPENDED Val_int(2)
#define BLOCKED_IO Val_int(4)
#define BLOCKED_DELAY Val_int(8)
#define BLOCKED_JOIN Val_int(16)
#define BLOCKED_WAIT Val_int(32)

#define BLOCKED_GC Val_int(64)
#define BLOCKED_END Val_int(128)
#define BLOCKED_REPLY Val_int(256)
#define BLOCKED_FAIL Val_int(512)
#define RAISE_EXCEPTION Val_int(1024)

#define BLOCKED_ANY ((BLOCKED_IO-1) + (BLOCKED_DELAY - 1)\
   + (BLOCKED_JOIN - 1) + (BLOCKED_WAIT - 1)\
   + (BLOCKED_GC - 1) + (BLOCKED_FAIL-1))

#define RESUMED_WAKEUP Val_int(0)
#define RESUMED_DELAY Val_int(1)
#define RESUMED_JOIN Val_int(2)

#define TAG_RESUMED_IO 0
#define TAG_RESUMED_WAIT 1

#define NO_FDS Val_unit
#define NO_DELAY Val_unit
#define NO_JOINING Val_unit
#define NO_WAITPID Val_int(0)

#define DELAY_INFTY 1E30        /* +infty, for this purpose */

#ifdef DEBUG
#  define check_runqueue()\
{\
  thread_t th = curr_thread;\
  \
  Assert(Is_in_heap(th));\
  th = th->next;\
  while(th != curr_thread){\
    Assert(Is_in_heap(th));\
    th = th->next;\
    }\
}
#else
#  define check_runqueue()
#endif

/* Since context switching is not allowed in callbacks, a thread that
   blocks during a callback is a deadlock. */
    /*    fatal_error("Thread: deadlock during callback");*/           

#define check_callback(msg) \
  if (callback_depth > 0){  \
     fprintf(stderr, "scheduler exited: callback depth : %d in %s \n",callback_depth,msg);\
	fatal_error("Thread: deadlock during callback");\
  }


extern value next_ident;

#define STACKS_BUFFER_SIZE 90
extern void free_stack(caml_thread_t);
extern value *get_stack();
extern int nbre_threads;
extern int terminaison; /* we detect terminaison */

typedef struct location_stub    *location_t;
typedef struct local_loc_struct *local_location_t;
/* Types.local_location 
#define LOCATION_FATHER  0
#define LOCATION_DEFS    1
#define LOCATION_LOCS    2
#define LOCATION_MODULE  3
#define LOCATION_ATTACH  4
#define LOCATION_FREE_STUBS      5
#define LOCATION_FREE_STUBS_SIZE 6
#define LOCATION_NEXT_FREE_STUB  7
#define LOCATION_SIZE    8
*/

#define LLOC_FATHER     0
#define LLOC_DEFWEAKSET 1
#define LLOC_LOCWEAKSET 2
#define LLOC_MODULE     3
#define LLOC_ATTACH     4
/*
#define LLOC_FREE_STUBS 5
#define LLOC_FREE_STUBS_SIZE 6
#define LLOC_NEXT_FREE_STUB  7
*/
#define LLOC_SIZE    8

#define MAX_STUBS_REUSED 100

struct local_loc_struct {
  location_t father;            /* the location in which is is */
  value      def_weakset;       /* set of contained automatas */
  value      loc_weakset;       /* set of contained locations */
  value      module;            /* inutile a priori */
  value      attach;            /* reloc attach */
  /*
  value      free_stubs;
  value      free_stubs_size;
  value      next_free_stub;
  */
};



/* byterun/obj.c */
value obj_block(value , value );

/* scheduler.c */
extern caml_thread_t curr_thread;
extern value local_thread_reuse(value location,value clos,value args
			 ,caml_thread_t thread);

extern value extract_location_threads (value location_v);
extern value gen_thread_new(value location,value clos,value param,value clos_arg);
extern value schedule_go_thread(value);
extern value set_critical_section(value);
extern value schedule_thread();
extern value thread_kill(value th);
extern void remove_thread(caml_thread_t th);

/* threads/scheduler.c */
extern int threads_reused, threads_allocated, stacks_reused, stacks_allocated;
extern int dirty_threads,scheduler_miss, scheduler_success;

/* local.c */
extern value get_thread_stub(caml_thread_t thread);
extern value trigger_def (value args);



/* Jtypes.locator */
#define LOCATOR_AGE      0
#define LOCATOR_WEAKSPACE 1
#define LOCATOR_WEAKNUM       2
#define LOCATOR_UID    3
#define LOCATOR_TYPE     4

/* Marshalled thread */ 	
#define THREAD_LOCATION       0
#define THREAD_TRAP_OFFSET    1 
#define THREAD_STACKTHRESHOLD 2
#define THREAD_STACKSIZE      3
#define THREAD_STATUS         4
#define THREAD_SYNC           5
#define THREAD_JOINING        6
#define THREAD_TRAPS          7
#define THREAD_MARSH_SIZE     8

typedef struct reloc_module     *reloc_module_t;
typedef struct def_stub         *def_t;
typedef struct local_def_struct *local_def_t;

struct reloc_module {
  value    module_name;
  value    module_slot;
  value    module_code_offset;
  value    module_code_len;
  value    module_code_begin;
  value    module_code_end;
  value    module_reloc;
  value    module_bytecode;
  value    module_interface;
  value    module_exceptions;
  value    module_migrate;
  value    module_data[1];     /* some local data */ 
};

#define LDEF_STATE     0
#define LDEF_STRUCTURE 1
#define LDEF_PROCESSES 2
#define LDEF_LOCATION  3
#define LDEF_OFFSET    4

#define STRUCT_NMULT  0
#define STRUCT_NCHANS 1
#define STRUCT_OFFSET 2

#define STATE_SIZE  3
/*
#define LDEF_OFFSET 2
#define LDEF_NTESTS Int_val(Field(ldef,1))
#define LDEF_STATE  Field(ldef,0)
#define LDEF_PATTERN(ldef,i)   Field(ldef,LDEF_OFFSET+STATE_SIZE*i)
#define LDEF_PROCESS(ldef,i)   Field(ldef,LDEF_OFFSET+1+STATE_SIZE*i)
#define LDEF_THREAD(ldef,i)    Field(ldef,LDEF_OFFSET+2+STATE_SIZE*i)
#define LDEF_ARGS(ldef,i)      Field(def,LDEF_OFFSET+ \
				     STATE_SIZE*LDEF_NTESTS(ldef)+i)
                                     
*/
struct local_def_struct {
  value state;
  value ndefs;
  value patterns[1];
};

/* Types.stub */
#define STUB_LOCATOR      0
#define STUB_POSITION     1
#define STUB_LOCAL        2
#define STUB_CACHE        3
#define STUB_GCSTUB         4
#define STUB_SIZE         5

#define STUB_EXT_TMV      5
#define STUB_IN_TMV       6

struct location_stub {
  value*           locator;
  value            space;
  local_location_t lloc;
  value            cache;
  value            gcstub;
};

struct def_stub {
  value*           locator;
  value            space;
  local_def_t      ldef;
  value*           cache;
  value            gcstub;
};

struct thread_stub {
  value*           locator;
  value            space;
  caml_thread_t    lthread;
  value            cache;
  value            gcstub;
};


#define DEF_WEAKSET_SIZE 2000
#define LOC_WEAKSET_SIZE 20

extern int threadstub_reused;
extern int threadstub_allocated;

extern value reuse_thread_loc(value location,value clos,
                              value args,value thread_stub);

#define RemoteObject_val    Val_int(0)
#define LocalLocation_tag   0
#define LocalThread_tag     1
#define LocalAutomata_tag   2
#define FailedObject_tag    3

#define SPACE_ID    0
#define SPACE_STATE 1

#define SPACE_Normal Val_int(0)
#define SPACE_MaybeDown Val_int(1)
#define SPACE_Down   Val_int(2)


/*

Chaque module commence par executer :
  BEGINMODULE initial_data

Initial data:
0 --> name:string
1 --> module_slot:int
2 --> offset: int
3 --> longueur: int
4 --> migrations: int
5 --> end_code: Obj.t
6 --> reloc: int array
7 --> bytecode: string

*/

/* Also defined in join.ml */
/* Slots INTERFACE(8) and EXCEPTIONS(9) are also defined in 
   bytecomp/translmod.ml */
#define MODULE_NAME 0
#define MODULE_SLOT 1
#define OFFSET      2
#define LONGUEUR    3
#define MIGRATIONS  4
#define _END_CODE   5
#define RELOC       6
#define BYTECODE    7
#define INTERFACE   8
#define CLASSES     9
#define MIGRATE     10
#define EXCEPTIONS  11
#define MODULE_DATA_SIZE 12

extern value module_table;

extern void init_modules(void);
extern void begin_module(code_t begin_code, value initial_data);
extern void resolv_module(code_t pc,int32 opcode,int offset);
extern void resolv_prim(code_t pc,int32 opcode,int offset);
extern value resolv_method(code_t pc);
extern code_t resolv_codeptr(code_t pc);
extern value find_code_ptr (code_t f,int32 *offset);
extern value install_module(value initial_data,value module);
#ifdef THREADED_CODE
#define THREADED(opcode)  ((opcode_t)(instr_table[opcode] - instr_base))
#else
#define  THREADED(opcode) opcode
#endif


extern void good_stub(value,char*,char*,int);
extern void failed_message(char *msg,char * expr,char *file, int line);


#ifdef DEBUG
#define Good_stub(msg,stub) good_stub(stub,msg, __FILE__, __LINE__)
#else
#define Good_stub(msg,stub)
#endif

#define GCSTUB_MIN       0
#define GCSTUB_MAX       1
#define GCSTUB_BCK_MARK  2
#define GCSTUB_BCK_TRACE 3

#define STAMP_MARK       0
#define STAMP_COLOR      1
#define STAMP_DISTANCE   2

#define COLOR_WHITE      Val_int(0)
#define COLOR_GRAY       Val_int(1)
#define COLOR_BLACK      Val_int(2)

#define MARK_RANGE   0
#define MARK_ID      1
#define MARK_SUBGEN  2

#define GCSCION_STAMP      0
#define GCSCION_NEXT_STAMP 1
#define GCSCION_BCK_MARK   2
#define GCSCION_BCK_STUB   3
#define GCSCION_GENERATOR  4
#define GCSCION_DELETED    5
#define GCSCION_FWD_STAMP  6
#define GCSCION_NEXT       7
#define GCSCION_TARGET     8
#define GCSCION_GCSTUB     9

#define REMOTE_OBJECT_TAG 0
#define LOCAL_LOCATION_TAG 1
#define LOCAL_THREAD_TAG 2
#define LOCAL_AUTOMATA_TAG 3
#define FAILED_OBJECT_TAG 4

#endif
