(***********************************************************************)
(*                                                                     *)
(*                           Dynlink                                   *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Interp
  
(* from typing/ident.ml *)
module Ident = struct
    type t = { stamp: int; name: string; mutable global: bool }
    let name i = i.name
  end

(* from parsing/asttypes.mli *)
module Asttype = struct
    type constant =
      Const_int of int
    | Const_char of char
    | Const_string of string
    | Const_float of string
  end

(* from bytecomp/lambda.ml *)
module Lambda = struct
    open Asttype
    
    type structured_constant =
      Const_base of constant
    | Const_pointer of int
    | Const_block of int * structured_constant list
    | Const_float_array of string list
  end


module Dumpobj = struct
    let inputu ic =
      let b1 = input_byte ic in
      let b2 = input_byte ic in
      let b3 = input_byte ic in
      let b4 = input_byte ic in
      (b4 lsl 24) + (b3 lsl 16) + (b2 lsl 8) + b1
  end

(* Translate structured constants *)

module Emitcode = struct 
(* from bytecomp/emitcode.ml *)
    open Lambda
    open Asttype
    
    type reloc_info =
      Reloc_literal of structured_constant    (* structured constant *)
    | Reloc_getglobal of Ident.t             (* reference to a global *)
    | Reloc_setglobal of Ident.t             (* definition of a global *)
    | Reloc_primitive of string               (* C primitive number *)
    type compilation_unit =
      { cu_name: string;                    (* Name of compilation unit *)
        mutable cu_pos: int;                (* Absolute position in file *)
        cu_codesize: int;                   (* Size of code block *)
        cu_reloc: (reloc_info * int) list;  (* Relocation information *)
        cu_imports: (string * Digest.t) list;
                 (* Names and CRC of intfs imported *)
        cu_primitives: string list;    (* Primitives declared inside *)
        mutable cu_force_link: bool;   (* Must be linked even if unref'ed *)
        mutable cu_debug: int;         (* Position of debugging info, or 0 *)
        cu_debugsize: int }            (* Length of debugging info *)
    
    
    let rec transl_const = function
        Const_base(Const_int i) -> Obj.repr i
      | Const_base(Const_char c) -> Obj.repr c
      | Const_base(Const_string s) -> Obj.repr s
      | Const_base(Const_float f) -> Obj.repr(float_of_string f)
      | Const_pointer i -> Obj.repr i
      | Const_block(tag, fields) ->
          let block = Obj.new_block tag (List.length fields) in
          let pos = ref 0 in
          List.iter
            (fun c -> Obj.set_field block !pos (transl_const c); incr pos)
          fields;
          block
      | Const_float_array fields ->
          Obj.repr(Array.of_list(List.map (fun f -> float_of_string f) fields))
  end



(* from utils/config.ml *)
module Config = struct
    let exec_magic_number = "Caml1999X004"
    and cmi_magic_number = "Caml1999I004"
    and cmo_magic_number = "Caml1999O004"
    and cma_magic_number = "Caml1999A004"
    and cmx_magic_number = "Caml1999Y006"
    and cmxa_magic_number = "Caml1999Z006"
    and ast_impl_magic_number = "Caml1999M005"
    and ast_intf_magic_number = "Caml1999N005"
  end

(* from tools/objinfo.ml *)

module Objinfo = struct
    open Emitcode
    
    let print_digest d =
      for i = 0 to String.length d - 1 do
        Printf.printf "%02x" (Char.code d.[i])
      done
    
    let print_info cu =
      print_string "  Unit name: "; print_string cu.cu_name; print_newline();
      print_string "  Interfaces imported:"; print_newline();
      List.iter
        (fun (name, digest) ->
          print_string "\t"; print_digest digest; print_string "\t";
          print_string name; print_newline())
      cu.cu_imports;
      print_string "  Uses unsafe features: ";
      begin match cu.cu_primitives with
          [] -> print_string "no"; print_newline()
        | l  -> print_string "YES"; print_newline();
            print_string "  Primitives declared in this module:";
            print_newline();
            List.iter
              (fun name -> print_string "\t"; print_string name; print_newline())
            l
      end
  end

(* from utils/misc.ml *)
module Misc = struct
    let find_in_path path name =
      if not (Filename.is_implicit name) then
        if Sys.file_exists name then name else raise Not_found
      else begin
          let rec try_dir = function
              [] -> raise Not_found
            | dir::rem ->
                let fullname = Filename.concat dir name in
                if Sys.file_exists fullname then fullname else try_dir rem
          in try_dir path
        end
  end

(* Own Symtable *)
module Symtable = struct
    open Emitcode
    open Interp
    
    type error =
      Undefined_global of string
    | Unavailable_primitive of string
    
    exception Error of error
    
    let slot_for_literal sc = 
      assert(!next_free_global < Array.length caml_globals);
      caml_globals.(!next_free_global) <- Emitcode.transl_const sc;
      incr next_free_global;
      !next_free_global - 1
    
    let slot_for_getglobal id = 
      let name = Ident.name id in
      try
        let i = Hashtbl.find globals name in
        Interp.install_module i;
        i
      with
        Not_found ->
          raise (Error (Undefined_global name))
    
    let slot_for_setglobal id = 
      let mod_name = Ident.name id in
      assert(!next_free_global < Array.length caml_globals);
      Hashtbl.add globals mod_name !next_free_global;
      incr next_free_global;
      !next_free_global - 1
    
    let num_of_prim name = 
      try
        Hashtbl.find prims name
      with
        Not_found ->
          raise (Error (Unavailable_primitive name))
    
    
    let patch_int buff pos n =
      String.unsafe_set buff pos (Char.unsafe_chr n);
      String.unsafe_set buff (pos + 1) (Char.unsafe_chr (n asr 8));
      String.unsafe_set buff (pos + 2) (Char.unsafe_chr (n asr 16));
      String.unsafe_set buff (pos + 3) (Char.unsafe_chr (n asr 24))
    
    
    let patch_object buff patchlist = 
      List.iter
        (function
          (Reloc_literal sc, pos) ->
            patch_int buff pos (slot_for_literal sc)
        | (Reloc_getglobal id, pos) ->
            patch_int buff pos (slot_for_getglobal id)
        | (Reloc_setglobal id, pos) ->
            patch_int buff pos (slot_for_setglobal id)
        | (Reloc_primitive name, pos) ->
            patch_int buff pos (num_of_prim name))
      patchlist
    
    let current_state () = ()
    let update_global_table () = ()
    let restore_state old_state = ()
    let hide_additions () = ()
    
    let predef_exn (exn : exn) = Obj.field (Obj.repr exn) 0
    let _ =
      caml_globals.(0) <- predef_exn Out_of_memory;
      caml_globals.(1) <- predef_exn (Sys_error "");
      caml_globals.(2) <- predef_exn (Failure "");
      caml_globals.(3) <- predef_exn (Invalid_argument "");
      caml_globals.(4) <- predef_exn End_of_file;
      caml_globals.(5) <- predef_exn Division_by_zero;
      caml_globals.(6) <- predef_exn Not_found;
      caml_globals.(7) <- predef_exn (Match_failure ("",0,0));
      caml_globals.(8) <- predef_exn Stack_overflow;
      Hashtbl.add globals "Out_of_memory" 0;
      Hashtbl.add globals "Sys_error" 1;
      Hashtbl.add globals "Failure" 2;
      Hashtbl.add globals "Invalid_argument" 3;
      Hashtbl.add globals "End_of_file" 4;
      Hashtbl.add globals "Division_by_zero" 5;
      Hashtbl.add globals "Not_found" 6;
      Hashtbl.add globals "Match_failure" 7;
      Hashtbl.add globals "Stack_overflow" 8
    
    let init_toplevel () = ()
  end

module Meta = struct
    
    open Obj
    open Instruct
    
    
    let unit = repr 0
    
    let inputu str i =
      let pos = i * 4 in
      let b1 = Char.code str.[pos] in
      let b2 = Char.code str.[pos+1] in
      let b3 = Char.code str.[pos+2] in
      let b4 = Char.code str.[pos+3] in
      (b4 lsl 24) + (b3 lsl 16) + (b2 lsl 8) + b1
    
    let inputs str i =
      let pos = i * 4 in
      let b1 = Char.code str.[pos] in
      let b2 = Char.code str.[pos+1] in
      let b3 = Char.code str.[pos+2] in
      let b4 = Char.code str.[pos+3] in
      let b4' = if b4 >= 128 then b4-256 else b4 in
      (b4' lsl 24) + (b3 lsl 16) + (b2 lsl 8) + b1
    
    let d = 0x3fffffff
    let signed v =  (* Only for 64 bits architectures *)
      let sign = if v > d then v - (1 lsl 32) else v in
      sign
    
    let fix_code str str_size =
      let code_size = str_size / 4 in
      let code = Array.create code_size 0 in
      let rec fix i =
        if i < code_size then 
          let op = inputu str i in
          code.(i) <- op;
          match magic op with
(* one arg *)
          | OpPUSHACC | OpACC | OpPOP | OpASSIGN | OpPUSHENVACC | OpENVACC
          | OpAPPLY | OpAPPTERM1 | OpAPPTERM2 | OpAPPTERM3 | OpRETURN | OpGRAB
          | OpPUSHGETGLOBAL | OpGETGLOBAL | OpSETGLOBAL | OpPUSHATOM | OpATOM
          | OpMAKEBLOCK1 | OpMAKEBLOCK2 | OpMAKEBLOCK3 | OpMAKEFLOATBLOCK
          | OpGETFIELD
          | OpGETFLOATFIELD | OpSETFIELD | OpSETFLOATFIELD | OpC_CALL1 | OpC_CALL2 | OpC_CALL3
          | OpC_CALL4
          | OpC_CALL5
            -> 
              let arg = inputu str (i+1) in
              code.(i+1) <- arg;
              fix (i+2)

(* signed arg *)
          | OpPUSH_RETADDR | OpOFFSETCLOSURE | OpPUSHOFFSETCLOSURE | OpBRANCH
          | OpBRANCHIF | OpPUSHTRAP | OpCONSTINT | OpPUSHCONSTINT
          | OpBRANCHIFNOT | OpOFFSETINT | OpOFFSETREF
            ->
              let arg = inputu str (i+1) in
              code.(i+1) <- (signed arg);
              fix (i+2)



(* two args *) 
          | OpAPPTERM | OpPUSHGETGLOBALFIELD | OpGETGLOBALFIELD
          | OpMAKEBLOCK
          | OpC_CALLN
            -> 
              let arg1 = inputu str (i+1) in              
              let arg2 = inputu str (i+2) in
              code.(i+1) <- arg1;
              code.(i+2) <- arg2;
              fix (i+3)

(* two args, second signed *)
          | OpCLOSURE
            -> 
              let arg1 = inputu str (i+1) in              
              let arg2 = inputu str (i+2) in
              code.(i+1) <- arg1;
              code.(i+2) <- (signed arg2);
              fix (i+3)
          
          | OpSWITCH ->
              let sizes = inputu str (i+1) in
              code.(i+1) <- sizes;
              let const_size = sizes land 0xFFFF in
              let block_size = sizes lsr 16 in
              let offset = 2 + const_size + block_size in
              for j = 2 to offset - 1 do
                code.(i + j) <- (inputu str (i + j))
              done;
              fix (i + offset)
          
          |  OpCLOSUREREC ->
              let nfuncs = inputu str (i+1) in
              let nvars = inputu str (i+2) in
              code.(i+1) <- nfuncs;
              code.(i+2) <- nvars;
              let offset = 3 + nfuncs in
              for j = 3 to offset - 1 do
                code.(i + j) <- (signed (inputu str (i + j)))
              done;
              fix (i + offset)
             

(* no args *) 
          | _ -> 
              fix (i+1) 
      
      in 
      fix 0;
      code
    
    open Interp
      
    let reify_bytecode buff code_size name =
      let code = fix_code buff code_size in
      let pc = 0 in
      let accu = repr 0 in
      let env = [||] in
      (*
      let clos = { stack = stack;code = code; env=env; recenv = (0,env) } in 
      let v = interp clos pc accu in
      *)
      let v = interp code pc accu env (0,env) in
  
(*      for i = 0 to opBREAK-1 do
        Printf.printf "%s : %d \n" opnames.(i) instructs.(i)
      done;
      print_newline ();
  *)
      v
  end

module Prims = struct
    open Interp
    
    let insert_one name (f: 'a -> 'b) =
      caml_prims.(!next_free_prim) <- (Obj.magic f);
      prims_call.(!next_free_prim) <- name;
      Hashtbl.add prims name !next_free_prim;
      incr next_free_prim
    
    open Obj
    
    external alloc_dummy : int -> 'a = "alloc_dummy"
    external update_dummy : 'a -> 'b -> 'c = "update_dummy"
    external make_vect: int -> 'a -> 'a array = "make_vect"
    external make_vect: int -> 'a -> 'a array = "make_vect"
    external md5_chan: in_channel -> int -> t = "md5_chan"
    external gc_stat: unit -> t = "gc_stat"
    external gc_get: unit -> t = "gc_get"
    external gc_set: t -> unit = "gc_set"
    external gc_minor: unit -> unit = "gc_minor"
    external gc_major: unit -> unit = "gc_major"
    external gc_full_major: unit -> unit = "gc_full_major"
    external gc_compaction: unit -> unit = "gc_compaction";;
    external hash_univ_param: int -> int -> 'a -> int = "hash_univ_param" "noalloc"
    external lex_engine: t -> int -> t -> int = "lex_engine"
    external input_value: in_channel -> 'a = "input_value"
    external obj_is_block: t -> bool = "obj_is_block"
    external obj_tag: t -> int = "obj_tag"
    external obj_block: int -> int -> t = "obj_block"
    external obj_dup: t -> t = "obj_dup"
    external obj_truncate: t -> int -> unit = "obj_truncate"
    external compare: 'a -> 'a -> int = "compare" "noalloc"
    external power_float: float -> float -> float = "power_float" "pow" "float"
    external sqrt_float: float -> float = "sqrt_float" "sqrt" "float"
    external exp_float: float -> float = "exp_float" "exp" "float"
    external log_float: float -> float = "log_float" "log" "float"
    external log10_float: float -> float = "log10_float" "log10" "float"
    external cos_float: float -> float = "cos_float" "cos" "float"
    external sin_float: float -> float = "sin_float" "sin" "float"
    external tan_float: float -> float = "tan_float" "tan" "float"
    external acos_float: float -> float = "acos_float" "acos" "float"
    external asin_float: float -> float = "asin_float" "asin" "float"
    external atan_float: float -> float = "atan_float" "atan" "float"
    external atan2_float: float -> float -> float = "atan2_float" "atan2" "float"
    external cosh_float: float -> float = "cosh_float" "cosh" "float"
    external sinh_float: float -> float = "sinh_float" "sinh" "float"
    external tanh_float: float -> float = "tanh_float" "tanh" "float"
    external ceil_float: float -> float = "ceil_float" "ceil" "float"
    external floor_float: float -> float = "floor_float" "floor" "float"
    external fmod_float: float -> float -> float = "fmod_float" "fmod" "float"
    external frexp_float: float -> float * int = "frexp_float"
    external ldexp_float: float -> int -> float = "ldexp_float"
    external modf_float: float -> float * float = "modf_float" "modf"
    external int_of_string: string -> int = "int_of_string"
    external float_of_string: string -> float = "float_of_string"
    external create_string: int -> string = "create_string"
    external sys_file_exists: string -> bool = "sys_file_exists"
    external sys_remove: string -> unit = "sys_remove"
    external sys_rename: string -> string -> unit = "sys_rename"
    external sys_getenv: string -> string = "sys_getenv"
    external sys_system_command: string -> int = "sys_system_command"
    external sys_time: unit -> float = "sys_time"
    external sys_chdir: string -> unit = "sys_chdir"
    external sys_getcwd: unit -> string = "sys_getcwd"
    external add_float : float -> float -> float = "add_float"
    
    external setup : out_channel -> t = "terminfo_setup";;
    external backup : int -> unit = "terminfo_backup";;
    external standout : bool -> unit = "terminfo_standout";;
    external resume : int -> unit = "terminfo_resume";;
    
    external array_get_float : t -> float -> t = "array_get_float"
    external array_set_float : t -> float -> t -> unit = "array_set_float"
    external array_get : t -> int -> t = "array_get"
    external array_set : t -> int -> t -> unit = "array_set"
    external array_unsafe_get : t -> int -> t = "array_unsafe_get"
    external array_unsafe_set : t -> int -> t -> unit = "array_unsafe_set"
    external lessthan : t -> t -> bool = "lessthan"
    external lessequal : t -> t -> bool = "lessequal"
    external string_equal: string -> string -> bool = "string_equal"
    external equal: string -> string -> bool = "equal"
    external notequal: string -> string -> bool = "notequal"
    external bitvect_test: string -> string -> bool = "bitvect_test"
    external string_notequal: string -> string -> bool = "string_notequal"
    external array_get_addr: 'a array -> int -> 'a = "array_get_addr"
    external array_set_addr: 'a array -> int -> 'a -> unit = "array_set_addr"
    external blit_string : string -> int -> string -> int -> int -> unit
    = "blit_string" "noalloc"
    external output_value : t -> 'a -> t -> unit = "output_value"
    external output_value_to_string: 'a -> t -> string = "output_value_to_string"
    external input_value : t -> 'a = "input_value"
    external greaterthan : t -> t -> bool = "greaterthan"
    external greaterequal : t -> t -> bool = "greaterequal"
    
    
    let not_implemented prim_name =
      fun _ ->
        failwith ("Primitive not implemented "^prim_name)
    
    let insert_all () =
      insert_one "ml_string_length" String.length;
      insert_one "string_set" String.set;
      insert_one "string_get" String.get;
      insert_one "make_vect" make_vect;
      insert_one "make_vect" make_vect;
      insert_one "md5_chan" md5_chan;
      insert_one "gc_stat" gc_stat;
      insert_one "gc_get" gc_get;
      insert_one "gc_set" gc_set;
      insert_one "gc_minor" gc_minor;
      insert_one "gc_major" gc_major;
      insert_one "gc_full_major" gc_full_major;
      insert_one "gc_compaction" gc_compaction;
      insert_one "hash_univ_param" hash_univ_param;
      insert_one "lex_engine" lex_engine;
      insert_one "input_value" input_value;
      insert_one "obj_is_block" obj_is_block;
      insert_one "obj_tag" obj_tag;
      insert_one "obj_block" obj_block;
      insert_one "obj_dup" obj_dup;
      insert_one "obj_truncate" obj_truncate;
      insert_one "compare" compare;
      insert_one "power_float" power_float;
      insert_one "sqrt_float" sqrt_float;
      insert_one "exp_float" exp_float;
      insert_one "log_float" log_float;
      insert_one "log10_float" log10_float;
      insert_one "cos_float" cos_float;
      insert_one "sin_float" sin_float;
      insert_one "tan_float" tan_float;
      insert_one "acos_float" acos_float;
      insert_one "asin_float" asin_float;
      insert_one "atan_float" atan_float;
      insert_one "atan2_float" atan2_float;
      insert_one "cosh_float" cosh_float;
      insert_one "sinh_float" sinh_float;
      insert_one "tanh_float" tanh_float;
      insert_one "ceil_float" ceil_float;
      insert_one "floor_float" floor_float;
      insert_one "fmod_float" fmod_float;
      insert_one "frexp_float" frexp_float;
      insert_one "ldexp_float" ldexp_float;
      insert_one "modf_float" modf_float;
      insert_one "int_of_string" int_of_string;
      insert_one "float_of_string" float_of_string;
      insert_one "create_string" create_string;
      insert_one "sys_file_exists" sys_file_exists;
      insert_one "sys_remove" sys_remove;
      insert_one "sys_rename" sys_rename;
      insert_one "sys_getenv" sys_getenv;
      insert_one "sys_system_command" sys_system_command;
      insert_one "sys_time" sys_time;
      insert_one "sys_chdir" sys_chdir;
      insert_one "sys_getcwd" sys_getcwd;
      insert_one "string_equal" string_equal;
      insert_one "equal" equal;
      insert_one "notequal" notequal;
      insert_one "lessthan" lessthan;
      insert_one "lessequal" lessequal;
      insert_one "array_set" array_set;
      insert_one "array_get" array_get;
      insert_one "array_set_float" array_set_float;
      insert_one "array_get_float" array_get_float;
      insert_one "add_float" add_float;
      insert_one "array_unsafe_set" array_unsafe_set;
      insert_one "array_unsafe_get" array_unsafe_get;
      insert_one "blit_string" blit_string;
      
      insert_one "bitvect_test" bitvect_test;
      insert_one "string_notequal" string_notequal;
      insert_one "array_get_addr" array_get_addr;
      insert_one "array_set_addr" array_set_addr;
      insert_one "terminfo_setup" setup ;
      insert_one "terminfo_backup" backup ;
      insert_one "terminfo_standout" standout ;
      insert_one "terminfo_resume" resume ;
      
      insert_one "get_global_data" (not_implemented "get_global_data");
      insert_one "realloc_global" (not_implemented "realloc_global" );
      insert_one "static_alloc" (not_implemented "static_alloc");
      insert_one "static_free" (not_implemented "static_free");
      insert_one "static_resize" (not_implemented "static_resize" );
      insert_one "reify_bytecode" (not_implemented "reify_bytecode");
      insert_one "available_primitives" (not_implemented "available_primitives" );
      insert_one "invoke_traced_function" (not_implemented "invoke_traced_function");
      
      
      insert_one "output_value_to_string" output_value_to_string;
      insert_one "output_value" output_value;
      insert_one "input_value" input_value;
      
      insert_one "alloc_dummy" alloc_dummy;
      insert_one "update_dummy" update_dummy;
      insert_one "greaterthan" greaterthan;
      insert_one "greaterequal" greaterequal;
  end
  