(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Efuns
open Text
open Frame
open Simple
open Select
open Interactive

let rec save_buffers_and_action frame buffers action =
  match buffers with
    [] -> action frame
  | (_,buf) :: buffers ->
      let text = buf.buf_text in
      if buf.buf_last_saved = version text  ||
        buf.buf_name.[0] = '*'
      then
        save_buffers_and_action frame buffers action
      else
      let map = Keymap.create () in
      let request = Printf.sprintf "Save buffer %s ? (y,n,!,a)" buf.buf_name
      in
      let yes mini_frame =
        Minibuffer.kill mini_frame frame;
        Buffer.save buf;
        save_buffers_and_action frame buffers action
      in
      let no mini_frame =
        Minibuffer.kill mini_frame frame;
        save_buffers_and_action frame buffers action
      in
      let action_immediately mini_frame _ = 
        Minibuffer.kill mini_frame frame;
        action mini_frame
      in
      let abort mini_frame =
        Minibuffer.kill mini_frame frame
      in
      Keymap.add_binding map [NormalMap, Char.code 'y'] yes;
      Keymap.add_binding map [NormalMap, Char.code 'Y'] yes;
      Keymap.add_binding map [NormalMap, Char.code 'n'] no;
      Keymap.add_binding map [NormalMap, Char.code 'N'] no;
      Keymap.add_binding map [NormalMap, Char.code '!'] action_immediately;
      Keymap.add_binding map [NormalMap, Char.code 'a'] abort;
      Keymap.add_binding map [NormalMap, Char.code 'A'] abort;
      Keymap.add_binding map [ControlMap, Char.code 'g'] abort;
      Minibuffer.create frame map request; ()

let exit_efuns frame =
  let top_window = Window.top frame.frm_window in
  let location = top_window.top_location in
  let buffers = Misc.list_of_hash location.loc_buffers in
  save_buffers_and_action frame buffers (fun _ -> exit 0)

let save_some_buffers frame =
  let top_window = Window.top frame.frm_window in
  let location = top_window.top_location in
  let buffers = Misc.list_of_hash location.loc_buffers in
  save_buffers_and_action frame buffers (fun _ -> ())

let load_buffer frame = 
  set_previous_frame frame;
  select_filename frame "Find file: " 
    (fun str -> 
      Frame.load_file frame.frm_window str)

let insert_file frame =
  select_filename frame "Insert file: "
    (fun str ->
      let inc = open_in str in
      insert_string frame (Misc.read_string inc);
      close_in inc
  )

let write_buffer frame = 
  let buf = frame.frm_buffer in
  select_filename frame "Save file as: " 
    (fun str -> 
      let top_window = Window.top frame.frm_window in
      Buffer.change_name top_window.top_location buf str;
      Buffer.save buf)

let save_buffer frame =
  let buf = frame.frm_buffer in
  match buf.buf_filename with
    Some _ -> Buffer.save buf
  | None -> write_buffer frame

let window_load_buffer frame = 
  select_filename frame "Find file: " 
    (fun str -> 
      let top_window = Window.top frame.frm_window in
      let top_window = Top_window.create top_window.top_location
          (Window.display top_window)
      in
      Frame.load_file top_window.top_windows str)


let change_buffer frame =
  let default = get_previous_frame () in
  set_previous_frame frame;
  select_buffer frame " Switch to buffer: " default (fun str ->
      let window = frame.frm_window in
      change_buffer window str)

let window_change_buffer frame =
  select_buffer frame "Switch to buffer in new frame: " 
    (get_previous_frame ())
  (fun name ->
      let top_window = Window.top frame.frm_window in
      let top_window = Top_window.create 
          top_window.top_location (Window.display top_window) in
      Frame.change_buffer top_window.top_windows name
  )

let change_font frame =
  Minibuffer.create_return frame (Keymap.create ()) "Find font: " "fixed"
    (fun old_frame name ->
      let window = frame.frm_window in
      let top_window = Window.top window in
      let xterm = Window.xterm top_window in
      Xterm.change_font xterm name
  ); ()

let color buf regexp strict attr =
  let text = buf.buf_text in
  let point = Text.add_point text in
  try
    while true do
      let len = Text.search_forward text regexp point in
      let before =
        if Text.bmove text point 1 = 1 then
          (let c = Text.get_char text point in
            Text.fmove text point (len+1);c)
        else
          (let c = Text.get_char text point in
            Text.fmove text point (len+1); c)
      in
      let after = Text.get_char text point in
      if not (strict && (buf.buf_syntax_table.(Char.code before) ||
            buf.buf_syntax_table.(Char.code after))) then
        begin
          bmove text point len;
          Text.set_attr text point len attr;
          fmove text point len;
          ()
        end
    done
  with
    Not_found -> 
      Text.remove_point text point;
      buf.buf_modified <- buf.buf_modified + 1

let display_hist = ref []
let open_display frame =
  select frame "open_display :" display_hist ""
    (fun _ -> [])
  (fun s -> s)
  (fun name -> 
      let top_window = Window.top frame.frm_window in
      let location = top_window.top_location in
      let display = Xterm.create_display name 
          location.loc_colors_names location.loc_fonts_names
          "Efuns" "efuns" in
      Top_window.create location display;
      ())

let goto_line frame =
  let buf = frame.frm_buffer in
  let text = buf.buf_text in
  let point = frame.frm_point in
  simple_select frame "goto-line:" 
    (fun name ->
      let line = int_of_string name in
      Text.goto_line text point (line - 1))

let goto_char frame =
  let buf = frame.frm_buffer in
  let text = buf.buf_text in
  let point = frame.frm_point in
  simple_select frame "goto-char:" 
    (fun name ->
      let char = int_of_string name in
      Text.set_position text point char)


let get_pos frame =
  let buf = frame.frm_buffer in
  let point = frame.frm_point in
  let text = buf.buf_text in
  let top_window = Window.top frame.frm_window in
  Top_window.message top_window 
    (Printf.sprintf "Char position %d" 
      (get_position text point))

let mark_at_point frame =
  let top_window = Window.top frame.frm_window in
  let buf = frame.frm_buffer in
  let point = frame.frm_point in
  Buffer.set_mark buf point;
  Top_window.message top_window "Mark set";
  ()

let mkdir frame =
  select_filename frame "Make directory: "
    (fun str -> Misc.wrap_unix (fun () -> Unix.mkdir str))

let _ =
  Efuns.add_start_hook (fun location ->
      Keymap.add_interactive (Location location) "make-directory" mkdir
  
  )
  