(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)


exception UnboundKey
module Vars :
  sig
    type key = string
    and 'a t
    val empty : 'a t
    val add : key -> 'a -> 'a t -> 'a t
    val find : key -> 'a t -> 'a
    val remove : key -> 'a t -> 'a t
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val map : ('a -> 'b) -> 'a t -> 'b t
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  end
type local_vars = Obj.t ref Vars.t
and map =
  { char_map: binding array;
    mutable complex_bindings: (key * binding) list;
    mutable partial_map: bool }
and key = mod_ident * Xtypes.keySym
and action = frame -> Obj.t
and mod_ident = | NormalMap | ControlMap | MetaMap
and prefix = key list
and binding = | Function of action | Prefix of map | Unbound
and buffer =
  { mutable buf_text: Text.t;
    mutable buf_modified: int;
    mutable buf_name: string;
    mutable buf_filename: string option;
    mutable buf_last_saved: int;
    mutable buf_history: (int * Text.action) list;
    mutable buf_charreprs: string array;
    mutable buf_syntax_table: bool array;
    buf_local_map: map;
    mutable buf_sync: bool;
    mutable buf_mark: Text.point option;
    mutable buf_point: Text.point;
    mutable buf_start: Text.point;
    mutable buf_shared: int;
    mutable buf_readonly: bool;
    mutable buf_modes: string list;
    mutable buf_finalizers: (unit -> unit) list;
    mutable buf_local_vars: local_vars;
    buf_location: location }
and frame =
  { mutable frm_buffer: buffer;
    mutable frm_location: location;
    mutable frm_window: window;
    mutable frm_last_text_updated: int;
    mutable frm_last_buf_updated: int;
    mutable frm_map: map;
    mutable frm_prefix: key list;
    mutable frm_repeat_action: int;
    mutable frm_last_action: action;
    mutable frm_start: Text.point;
    mutable frm_end: Text.point;
    mutable frm_y_offset: int;
    mutable frm_point: Text.point;
    mutable frm_cursor_x: int;
    mutable frm_cursor_y: int;
    mutable frm_cursor: string;
    mutable frm_cursor_attr: Text.attribute;
    mutable frm_force_point: bool;
    mutable frm_force_start: bool;
    mutable frm_force_cursor: bool;
    mutable frm_x_offset: int;
    mutable frm_cutline: int;
    mutable frm_has_scrollbar: int;
    mutable frm_has_status_line: int;
    mutable frm_status: status;
    mutable frm_xpos: int;
    mutable frm_ypos: int;
    mutable frm_width: int;
    mutable frm_height: int;
    mutable frm_table: line_repr array;
    mutable frm_killed: bool;
    mutable frm_mini_buffer: string option;
    mutable frm_redraw: bool;
    mutable frm_scrollbar: Xterm.xterm_scrollbar option;
    mutable frm_local_vars: local_vars }
and status_info =
  | StatModified
  | StatName
  | StatLine
  | StatCol
  | StatFile
  | StatMode
and status =
  { mutable status_string: string;
    mutable status_modified: bool;
    mutable status_format: (status_info * (int * int)) list;
    mutable stat_col: int;
    mutable stat_name: string;
    mutable stat_file: string;
    mutable stat_line: int;
    mutable stat_modified: bool;
    mutable stat_modes: string list }
and line_repr =
  { mutable repr_line: Text.line;
    mutable repr_y: int;
    mutable repr_prev_offset: int;
    mutable repr_prev_reprs: Text.repr list;
    mutable repr_offset: int;
    mutable repr_reprs: Text.repr list }
and top_window =
  { mutable top_location: location;
    mutable top_display: Xterm.xterm_display option;
    mutable top_xterm: Xterm.xterm_window option;
    mutable top_windows: window;
    mutable top_mini_buffers: frame list;
    mutable top_width: int;
    mutable top_height: int;
    mutable top_name: string;
    mutable top_active_frame: frame;
    mutable top_second_cursor: frame option }
and window =
  { mutable win_xpos: int;
    mutable win_ypos: int;
    mutable win_width: int;
    mutable win_height: int;
    mutable win_down: window_down;
    mutable win_up: window_up;
    mutable win_mini: bool }
and window_up = | Window of window | TopWindow of top_window
and window_down =
  | HComb of window * window
  | VComb of window * window
  | NoFrame of unit
  | WFrame of frame
and location =
  { loc_global_map: map;
    mutable loc_windows: top_window list;
    mutable loc_buffers: (string, buffer) Hashtbl.t;
    mutable loc_files: (string, buffer) Hashtbl.t;
    mutable loc_dirname: string;
    mutable loc_width: int;
    mutable loc_height: int;
    mutable loc_fg: string;
    mutable loc_bg: string;
    mutable loc_font: string;
    mutable loc_local_vars: local_vars;
    mutable loc_counter: int;

    loc_vars_table : (string, (Obj.t -> string) * (string -> Obj.t)) Hashtbl.t;
    loc_fonts : (string,int) Hashtbl.t;
    loc_fonts_names : string array;
    mutable loc_fonts_n : int;
    loc_colors : (string,int) Hashtbl.t;
    loc_colors_names : string array;
    mutable loc_colors_n : int;
    loc_mutex: Concur.Mutex.t }
and sens = | Backward | Forward
and to_regexp = | Regexp | RegexpString

  
val efunspath : string list
val camldir : string
val path : string list
val load_path : string list ref
val init_files : string list ref
val displayname : string ref
val width : int ref
val height : int ref
val font : string ref
val fg : string ref
val bg : string ref
val ns_name : string ref
val ns_host : string ref
val ns_port : int ref

val start_hooks : (location -> unit) list ref
val add_start_hook : (location -> unit) -> unit
val init : location -> unit

type 'a local_var
type env = | Buffer of buffer | Frame of frame | Location of location

val vars_table : (string, (Obj.t -> string) * (string -> Obj.t)) Hashtbl.t
val new_local_var : string -> ('a -> string) -> (string -> 'a) -> 'a local_var
val new_local_var_abstr : string -> 'a local_var
val set_local_var : env -> 'a local_var -> 'a -> unit
val get_local_var : env -> 'a local_var -> 'a
val no_set : string -> 'a
val no_get : 'a -> string

val add_hook : env -> 'a list local_var -> 'a -> unit
val get_hooks : env -> 'a local_var -> 'a list
val exec_hooks : env -> ('a -> unit) -> 'a list local_var -> unit
