(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(*
La gestion des keymaps est inadapte'e a` la modification dynamique.
En effet, on aurait envie de pouvoir modifier les bindings selon divers
  crite`res, tels que le buffer (possible), le mode(pas possible).
*)


open Efuns

let dummy_action frame = Obj.repr 0
let unbound_key buffer _ = raise UnboundKey

let create () =
  { char_map = Array.create 256 Unbound;
    complex_bindings = [];
    partial_map = true;
  } 

let print_key (map,keysym) =
  let prefix =
    match map with
      NormalMap -> ""
    | ControlMap -> "C-"
    | MetaMap -> "M-"
  in
  let kname =
    try
      List.assoc keysym XK.keysym_to_name
    with
      Not_found -> "?"
  in
  prefix^kname
            
let rec print_key_list key_list =
  match key_list with
    key :: tail ->
      Printf.sprintf "%s %s" (print_key_list tail) (print_key key)
  | _ -> ""

let print map =
  Printf.printf "MAP:\n";
  List.iter (fun (key, binding) ->
    (match binding with
      Prefix _ -> print_string "Prefix "; 
    | Function _ -> print_string "Function"
    | Unbound -> print_string "Unbound"
          );
    print_string (print_key key);
    print_newline ()
      ) map.complex_bindings
    
let get_binding map key =
  match key with
    (NormalMap,key) when key >= 0 && key < 256 -> 
      map.char_map.(key)
  | _ -> 
      try
        List.assoc key map.complex_bindings
      with
        Not_found -> raise UnboundKey

let set_binding map key binding =
  match key with
    (NormalMap,key) when key >= 0 && key < 256 -> 
      map.char_map.(key) <- binding
  | _ -> 
      map.complex_bindings <- (key,binding) :: map.complex_bindings

let rec add_complex_binding map key_list binding =
  match key_list with
    [] -> failwith "ERROR add_complex_binding: empty key list"
  | [key] ->
      set_binding map key binding
  | key :: tail ->
      try
        match get_binding map key with
          Prefix map ->
            add_complex_binding map tail binding
        | _ -> failwith "ERROR add_complex_binding: unbound prefix"
      with
        UnboundKey -> failwith "ERROR add_complex_binding: unbound prefix"

let define_char_binding map char (f : frame -> 'a) =
  add_complex_binding map [NormalMap,Char.code char] 
  (Function (Obj.magic f : action))

let add_prefix map key_list =
  let prefix_map = create () in
  add_complex_binding  map key_list (Prefix prefix_map)

let add_binding map key_list (f : frame -> 'a) =
  add_complex_binding map key_list (Function (Obj.magic f: action))

    
let c_h = (ControlMap, Char.code 'h')
let c_x = (ControlMap, Char.code 'x')
let c_c = (ControlMap, Char.code 'c')
let n_5 = (NormalMap, Char.code '5')
  

let interactives = new_local_var "interactives" no_get no_set
  

  
let create_bindings location =
  let actives = get_local_var (Location location) interactives in
  let s = ref "Default bindings:" in
  List.iter (fun (name,(_,binding)) ->
    match !binding with
      None -> ()
    | Some key_list ->
        s := Printf.sprintf "%s\n%s : %s" !s 
             (print_key_list (List.rev key_list)) name
  ) actives;
  !s
  
let add_interactive env name (f: frame -> 'a) =
  let actives = get_local_var env interactives in
  set_local_var env interactives 
    ((name,((Obj.magic f: frame -> Obj.t),ref None)) :: actives)

let interactive env key_list name (f : frame -> 'a) = 
  let add_interactive env name f key =
    let actives = get_local_var env interactives in
    set_local_var env interactives 
      ((name,((Obj.magic f: frame -> Obj.t),ref (Some key))) :: actives)
  in
  let map =
    match env with
      Buffer buf -> buf.buf_local_map
    | Frame frame -> frame.frm_buffer.buf_local_map
    | Location location -> location.loc_global_map
  in
  add_interactive env name f key_list;
  add_binding map key_list f

let _ =
  add_start_hook (fun location ->
    set_local_var (Location location) interactives [])
