(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(*
  Useful functions.
*)

open Parsers
open Lexers

let count_char_sub str pos len char =
  let pos_end = pos + len in
  let rec occur pos n =
    if pos = pos_end then (n, 0) else
    let next_pos = 
      try
        String.index_from str pos char
      with
        Not_found -> -1
    in
    if next_pos = -1 || next_pos >= pos_end then (n,pos_end - pos) else
    occur (next_pos + 1) (n + 1)
  in
  occur pos 0 

let count_char str char =
  let len = String.length str in
  count_char_sub str 0 len char

let rec list_nth n list =
  match n, list with
    0, (ele :: _) -> ele
  | _, [] -> raise Not_found
  | _, (_:: tail) -> list_nth (n-1) tail


let list_removeq list ele =
  List.fold_left (fun list e ->
                    if e == ele then list
                    else e ::  list) [] list

let remove_assocq ele list =
  List.fold_left (fun list ((e,_) as head) ->
                    if e == ele then list
                    else head ::  list) [] list

let list_remove list ele =
  List.fold_left (fun list e ->
                    if e = ele then list
                    else e ::  list) [] list

let remove_assoc ele list =
  List.fold_left (fun list ((e,_) as head) ->
                    if e = ele then list
                    else head ::  list) [] list

let passwd =
  let ic = open_in "/etc/passwd" in
  let file = 
    try
      parse_passwd lexer_passwd (Lexing.from_channel ic) 
    with
      _ -> close_in ic; []
  in
  file

let wrap_unix f = 
  try
    f ()
  with
    Unix.Unix_error (error,f,arg) ->
      failwith (Printf.sprintf "Unix error %s: %s %s" 
          f (Unix.error_message error) arg)
      

let decomp name =
  let lexbuf = Lexing.from_string name in
  parse_filename lexer_filename lexbuf

(*
let rec longname dirname name =
  if is_relative name then
    longname "" (Filename.concat dirname name)
  else
    let tokens = decomp name in
    let rec iter longname tokens =
      match tokens with
        [] -> ""
      | [
    let len = String.length name in
    let rec iter i n =
    if i+n+1 >= len || name.[i+n] = '/' then
      if i+n+1 >= len then
        String.sub name i n
      else
        (String.sub name i (n-1)
      
  in
  iter 0 0

let longname curdirname name =
  let rec 
  
*)

let file_list dirname =
  try
    let dir = Unix.opendir dirname in
    let rec iter list =
      try
        let file = Unix.readdir dir in
        iter (file :: list)
      with
        _ -> list
    in
    let list = iter [] in
    Unix.closedir dir;
    list
  with
    _ -> []

let string_ncmp s1 s2 n =
  let sz1 = String.length s1 in
  let sz2 = String.length s2 in
  if sz1 < n || sz2 < n then s1 = s2
  else
    let s1' = String.sub s1 0 n in
    let s2' = String.sub s2 0 n in
    s1' = s2'

let completion list start =
  let n = String.length start in
  if n = 0 then list else
  List.fold_left (fun list s ->
    if string_ncmp start s n then
      s :: list
    else
      list
    ) [] list

let common_suffix list start =
  let newlist = completion list start in
  let lenl = List.length newlist in
  let len = String.length start in
  match newlist with
    [] -> "",0
  | ele :: tail ->
      let lenele = String.length ele in
      if tail = [] then
        String.sub ele len (lenele - len),1
      else
        let maxlen = List.fold_left (fun len e -> min len (String.length e))
            lenele tail in
        let suffix = ref "" in
        try
          for i = len to maxlen - 1  do
            let c = ele.[i] in
            List.iter (fun e -> if not (e.[i] == c) then raise Exit) tail;
            suffix := !suffix ^ (String.make 1 c)
          done;
          raise Exit
        with
          _ -> !suffix, lenl

let read_string inc =
  let len = in_channel_length inc in
  let str = String.create len in
  let curs = ref 0 in
  let left = ref len in
  try
    while !left > 0 do
      let read = input inc str !curs !left in
      if read = 0 then raise End_of_file;
      left := !left - read;
      curs := !curs + read;
    done;
    str
  with
    End_of_file ->
      String.sub str 0 !curs

let list_of_hash t =
  let l = ref [] in
  Hashtbl.iter (fun key ele -> l := (key,ele) :: !l) t;
  !l

let find_in_path path name =
  if not (Filename.is_implicit name) then
    if Sys.file_exists name then name else raise Not_found
  else begin
    let rec try_dir = function
        [] -> raise Not_found
      | dir::rem ->
          let fullname = Filename.concat dir name in
          if Sys.file_exists fullname then fullname else try_dir rem
    in try_dir path
  end
      
let string_to_path str =
  let len = String.length str in
  let rec iter start pos =
    if pos >= len then
      [String.sub str start (len - start)]
    else
      if str.[pos] = ' ' || str.[pos] = ':' then
        (String.sub str start (pos - start)) :: (iter (pos+1) (pos+1))
      else
        iter start (pos+1)
  in
  iter 0 0

let hash_add_assoc hash list =
  List.iter (fun (abbrev, repl) -> Hashtbl.add hash abbrev repl) list 


let regexp_from_list list =
  Str2.regexp  
    (match list with
      [] -> ""
    | hd :: tail ->
        List.fold_left (fun str ele ->
          str ^ "\|\(" ^ ele ^ "\)") ("\(" ^ hd ^ "\)") tail)

            (* Simplify a filename before printing it *)

let user =
  try
    Sys.getenv "USER"
  with
    Not_found -> ""

let homedir =
  try
    Sys.getenv "HOME"
  with
    Not_found -> ""

let homedirs = 
  List.fold_left  (fun dirs line ->
      match line with
        login::_::_::_::_::home::_ -> ("~"^login,home) :: dirs
      | _ -> dirs) ["~",homedir] passwd
      
let users = List.map fst homedirs
          
let known_dirs = ref [ homedir,"~"]
  
let check_prefix filename prefix =
  let len = String.length prefix in
  let n = String.length filename - len in
  if n < 0 then raise Not_found;
  for i = 0 to len-1 do
    if filename.[i] <> prefix.[i] then raise Not_found
  done;
  String.sub filename len n
      
let replace_prefix filename prefixes =
  List.fold_left (fun filename (dir,repl) ->
      try
        let fin = check_prefix filename dir in
        repl ^ fin
      with
        Not_found -> filename) filename prefixes
  
let filename_to_string filename =
(* replace known dirs by ~ *)
  replace_prefix filename !known_dirs
  
let string_to_filename str =
  replace_prefix str homedirs
  
