(***********************************************************************)
(*                                                                     *)
(*                           Efuns                                     *)
(*                                                                     *)
(*         Fabrice Le Fessant, projet PARA, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* This code is taken from different module of Ocaml 2.00 distribution.
   It enables the system to know all ocamlc types. *)

(* from utils/config.ml *)
module Config = struct
    let exec_magic_number = "Caml1999X004"
    and cmi_magic_number = "Caml1999I004"
    and cmo_magic_number = "Caml1999O004"
    and cma_magic_number = "Caml1999A004"
    and cmx_magic_number = "Caml1999Y006"
    and cmxa_magic_number = "Caml1999Z006"
    and ast_impl_magic_number = "Caml1999M005"
    and ast_intf_magic_number = "Caml1999N005"
  end
  
  
(* from utils/tbl.ml *)
module Tbl = struct
    
    type ('a, 'b) t =
      Empty
    | Node of ('a, 'b) t * 'a * 'b * ('a, 'b) t * int
    
    let empty = Empty
    
    let height = function
        Empty -> 0
      | Node(_,_,_,_,h) -> h
    
    let create l x d r =
      let hl = height l and hr = height r in
      Node(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))
    
    let bal l x d r =
      let hl = height l and hr = height r in
      if hl > hr + 1 then
        match l with
        | Node (ll, lv, ld, lr, _) when height ll >= height lr ->
            create ll lv ld (create lr x d r)
        | Node (ll, lv, ld, Node (lrl, lrv, lrd, lrr, _), _) ->
            create (create ll lv ld lrl) lrv lrd (create lrr x d r)
        | _ -> assert false
      else if hr > hl + 1 then
          match r with
          | Node (rl, rv, rd, rr, _) when height rr >= height rl ->
              create (create l x d rl) rv rd rr
          | Node (Node (rll, rlv, rld, rlr, _), rv, rd, rr, _) ->
              create (create l x d rll) rlv rld (create rlr rv rd rr)
          | _ -> assert false
        else
          create l x d r
    
    let rec add x data = function
        Empty ->
          Node(Empty, x, data, Empty, 1)
      | Node(l, v, d, r, h) as t ->
          let c = compare x v in
          if c = 0 then
            Node(l, x, data, r, h)
          else if c < 0 then
              bal (add x data l) v d r
            else
              bal l v d (add x data r)
    
    let rec find x = function
        Empty ->
          raise Not_found
      | Node(l, v, d, r, _) ->
          let c = compare x v in
          if c = 0 then d
          else find x (if c < 0 then l else r)
    
    let rec mem x = function
        Empty -> false
      | Node(l, v, d, r, _) ->
          let c = compare x v in
          c = 0 || mem x (if c < 0 then l else r)
    
    let rec merge t1 t2 =
      match (t1, t2) with
        (Empty, t) -> t
      | (t, Empty) -> t
      | (Node(l1, v1, d1, r1, h1), Node(l2, v2, d2, r2, h2)) ->
          bal l1 v1 d1 (bal (merge r1 l2) v2 d2 r2)
    
    let rec remove x = function
        Empty ->
          Empty
      | Node(l, v, d, r, h) as t ->
          let c = compare x v in
          if c = 0 then
            merge l r
          else if c < 0 then
              bal (remove x l) v d r
            else
              bal l v d (remove x r)
    
    let rec iter f = function
        Empty -> ()
      | Node(l, v, d, r, _) ->
          iter f l; f v d; iter f r
  
  end

(* from parsing/longident.ml *)
module Longident = struct
    type t =
      Lident of string
    | Ldot of t * string
    | Lapply of t * t
  end

(* from parsing/asttypes.mli *)
module Asttypes = struct
    
    type constant =
      Const_int of int
    | Const_char of char
    | Const_string of string
    | Const_float of string
    
    type rec_flag = Nonrecursive | Recursive
    
    type direction_flag = Upto | Downto
    
    type private_flag = Private | Public
    
    type mutable_flag = Immutable | Mutable
    
    type virtual_flag = Virtual | Concrete
  end

(* from parsing/location.ml *)
module Location = struct
    type t =
      { loc_start: int; loc_end: int }
  end

(* from parsing/parsetree.mli *)
module Parsetree = struct

(* Abstract syntax tree produced by parsing *)
    
    open Asttypes

(* Type expressions for the core language *)
    
    type core_type =
      { ptyp_desc: core_type_desc;
        ptyp_loc: Location.t }
    
    and core_type_desc = 
      Ptyp_any
    | Ptyp_var of string
    | Ptyp_arrow of core_type * core_type
    | Ptyp_tuple of core_type list
    | Ptyp_constr of Longident.t * core_type list
    | Ptyp_object of core_field_type list
    | Ptyp_class of Longident.t * core_type list
    | Ptyp_alias of core_type * string
    
    and core_field_type =
      { pfield_desc: core_field_desc;
        pfield_loc: Location.t }
    
    and core_field_desc =
      Pfield of string * core_type
    | Pfield_var

(* XXX Type expressions for the class language *)
    
    type 'a class_infos =
      { pci_virt: virtual_flag;
        pci_params: string list * Location.t;
        pci_name: string;
        pci_expr: 'a;
        pci_loc: Location.t }

(* Value expressions for the core language *)
    
    type pattern =
      { ppat_desc: pattern_desc;
        ppat_loc: Location.t }
    
    and pattern_desc =
      Ppat_any
    | Ppat_var of string
    | Ppat_alias of pattern * string
    | Ppat_constant of constant
    | Ppat_tuple of pattern list
    | Ppat_construct of Longident.t * pattern option * bool
    | Ppat_record of (Longident.t * pattern) list
    | Ppat_array of pattern list
    | Ppat_or of pattern * pattern
    | Ppat_constraint of pattern * core_type
    
    type expression =
      { pexp_desc: expression_desc;
        pexp_loc: Location.t }
    
    and expression_desc =
      Pexp_ident of Longident.t
    | Pexp_constant of constant
    | Pexp_let of rec_flag * (pattern * expression) list * expression
    | Pexp_function of (pattern * expression) list
    | Pexp_apply of expression * expression list
    | Pexp_match of expression * (pattern * expression) list
    | Pexp_try of expression * (pattern * expression) list
    | Pexp_tuple of expression list
    | Pexp_construct of Longident.t * expression option * bool
    | Pexp_record of (Longident.t * expression) list * expression option
    | Pexp_field of expression * Longident.t
    | Pexp_setfield of expression * Longident.t * expression
    | Pexp_array of expression list
    | Pexp_ifthenelse of expression * expression * expression option
    | Pexp_sequence of expression * expression
    | Pexp_while of expression * expression
    | Pexp_for of string * expression * expression * direction_flag * expression
    | Pexp_constraint of expression * core_type option * core_type option
    | Pexp_when of expression * expression
    | Pexp_send of expression * string
    | Pexp_new of Longident.t
    | Pexp_setinstvar of string * expression
    | Pexp_override of (string * expression) list
    | Pexp_letmodule of string * module_expr * expression

(* Value descriptions *)
    
    and value_description =
      { pval_type: core_type;
        pval_prim: string list }

(* Type declarations *)
    
    and type_declaration =
      { ptype_params: string list;
        ptype_cstrs: (core_type * core_type * Location.t) list;
        ptype_kind: type_kind;
        ptype_manifest: core_type option;
        ptype_loc: Location.t }
    
    and type_kind =
      Ptype_abstract
    | Ptype_variant of (string * core_type list) list
    | Ptype_record of (string * mutable_flag * core_type) list
    
    and exception_declaration = core_type list

(* Type expressions for the class language *)
    
    and class_type =
      { pcty_desc: class_type_desc;
        pcty_loc: Location.t }
    
    and class_type_desc =
      Pcty_constr of Longident.t * core_type list
    | Pcty_signature of class_signature
    | Pcty_fun of core_type * class_type
    
    and class_signature = core_type * class_type_field list
    
    and class_type_field =
      Pctf_inher of class_type
    | Pctf_val   of (string * mutable_flag * core_type option * Location.t)
    | Pctf_virt  of (string * private_flag * core_type * Location.t)
    | Pctf_meth  of (string * private_flag * core_type * Location.t)
    | Pctf_cstr  of (core_type * core_type * Location.t)
    
    and class_description = class_type class_infos
    
    and class_type_declaration = class_type class_infos

(* Value expressions for the class language *)
    
    and class_expr =
      { pcl_desc: class_expr_desc;
        pcl_loc: Location.t }
    
    and class_expr_desc =
      Pcl_constr of Longident.t * core_type list
    | Pcl_structure of class_structure
    | Pcl_fun of pattern * class_expr
    | Pcl_apply of class_expr * expression list
    | Pcl_let of rec_flag * (pattern * expression) list * class_expr
    | Pcl_constraint of class_expr * class_type
    
    and class_structure = pattern * class_field list
    
    and class_field =
      Pcf_inher of class_expr * string option
    | Pcf_val   of (string * mutable_flag * expression * Location.t)
    | Pcf_virt  of (string * private_flag * core_type * Location.t)
    | Pcf_meth  of (string * private_flag * expression * Location.t)
    | Pcf_cstr  of (core_type * core_type * Location.t)
    | Pcf_let   of rec_flag * (pattern * expression) list * Location.t
    | Pcf_init  of expression
    
    and class_declaration = class_expr class_infos

(* Type expressions for the module language *)
    
    and module_type =
      { pmty_desc: module_type_desc;
        pmty_loc: Location.t }
    
    and module_type_desc =
      Pmty_ident of Longident.t
    | Pmty_signature of signature
    | Pmty_functor of string * module_type * module_type
    | Pmty_with of module_type * (Longident.t * with_constraint) list
    
    and signature = signature_item list
    
    and signature_item =
      { psig_desc: signature_item_desc;
        psig_loc: Location.t }
    
    and signature_item_desc =
      Psig_value of string * value_description
    | Psig_type of (string * type_declaration) list
    | Psig_exception of string * exception_declaration
    | Psig_module of string * module_type
    | Psig_modtype of string * modtype_declaration
    | Psig_open of Longident.t
    | Psig_include of module_type
    | Psig_class of class_description list
    | Psig_class_type of class_type_declaration list
    
    and modtype_declaration =
      Pmodtype_abstract
    | Pmodtype_manifest of module_type
    
    and with_constraint =
      Pwith_type of type_declaration
    | Pwith_module of Longident.t

(* Value expressions for the module language *)
    
    and module_expr =
      { pmod_desc: module_expr_desc;
        pmod_loc: Location.t }
    
    and module_expr_desc =
      Pmod_ident of Longident.t
    | Pmod_structure of structure
    | Pmod_functor of string * module_type * module_expr
    | Pmod_apply of module_expr * module_expr
    | Pmod_constraint of module_expr * module_type
    
    and structure = structure_item list
    
    and structure_item =
      { pstr_desc: structure_item_desc;
        pstr_loc: Location.t }
    
    and structure_item_desc =
      Pstr_eval of expression
    | Pstr_value of rec_flag * (pattern * expression) list
    | Pstr_primitive of string * value_description
    | Pstr_type of (string * type_declaration) list
    | Pstr_exception of string * exception_declaration
    | Pstr_module of string * module_expr
    | Pstr_modtype of string * module_type
    | Pstr_open of Longident.t
    | Pstr_class of class_declaration list
    | Pstr_class_type of class_type_declaration list


(* Toplevel phrases *)
    
    type toplevel_phrase =
      Ptop_def of structure
    | Ptop_dir of string * directive_argument
    
    and directive_argument =
      Pdir_none
    | Pdir_string of string
    | Pdir_int of int
    | Pdir_ident of Longident.t
  
  end

(* from typing/ident.ml *)
module Ident = struct

(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ocamlc.ml,v 1.1 1999/09/30 13:22:46 lefessan Exp $ *)

open Format

type t = { stamp: int; name: string; mutable global: bool }

(* A stamp of 0 denotes a persistent identifier *)

let currentstamp = ref 0

let create s =
  incr currentstamp;
  { name = s; stamp = !currentstamp; global = false }

let create_persistent s =
  { name = s; stamp = 0; global = true }

let name i = i.name

let unique_name i = i.name ^ "_" ^ string_of_int i.stamp

let persistent i = (i.stamp = 0)

let equal i1 i2 = i1.name = i2.name

let same i1 i2 = i1 = i2
  (* Possibly more efficient version (with a real compiler, at least):
       if i1.stamp <> 0
       then i1.stamp = i2.stamp
       else i2.stamp = 0 & i1.name = i2.name *)

let binding_time i = i.stamp

let current_time() = !currentstamp
let set_current_time t = currentstamp := max !currentstamp t

let hide i =
  { stamp = -1; name = i.name; global = i.global }

let make_global i =
  i.global <- true

let global i =
  i.global

let print i =
  print_string i.name;
  match i.stamp with
    0 -> print_string "!"
  | -1 -> print_string "#"
  | n -> print_string "/"; print_int n; if i.global then print_string "g"

type 'a tbl =
    Empty
  | Node of 'a tbl * 'a data * 'a tbl * int

and 'a data =
  { ident: t;
    data: 'a;
    previous: 'a data option }

let empty = Empty

(* Inline expansion of height for better speed
 * let height = function
 *     Empty -> 0
 *   | Node(_,_,_,h) -> h
 *)

let mknode l d r =
  let hl = match l with Empty -> 0 | Node(_,_,_,h) -> h
  and hr = match r with Empty -> 0 | Node(_,_,_,h) -> h in
  Node(l, d, r, (if hl >= hr then hl + 1 else hr + 1))

let balance l d r =
  let hl = match l with Empty -> 0 | Node(_,_,_,h) -> h
  and hr = match r with Empty -> 0 | Node(_,_,_,h) -> h in
  if hl > hr + 1 then
    match l with
    | Node (ll, ld, lr, _)
      when (match ll with Empty -> 0 | Node(_,_,_,h) -> h) >=
           (match lr with Empty -> 0 | Node(_,_,_,h) -> h) ->
        mknode ll ld (mknode lr d r)
    | Node (ll, ld, Node(lrl, lrd, lrr, _), _) ->
        mknode (mknode ll ld lrl) lrd (mknode lrr d r)
    | _ -> assert false
  else if hr > hl + 1 then
    match r with
    | Node (rl, rd, rr, _)
      when (match rr with Empty -> 0 | Node(_,_,_,h) -> h) >=
           (match rl with Empty -> 0 | Node(_,_,_,h) -> h) ->
        mknode (mknode l d rl) rd rr
    | Node (Node (rll, rld, rlr, _), rd, rr, _) ->
        mknode (mknode l d rll) rld (mknode rlr rd rr)
    | _ -> assert false
  else
    mknode l d r

let rec add id data = function
    Empty ->
      Node(Empty, {ident = id; data = data; previous = None}, Empty, 1)
  | Node(l, k, r, h) ->
      let c = compare id.name k.ident.name in
      if c = 0 then
        Node(l, {ident = id; data = data; previous = Some k}, r, h)
      else if c < 0 then
        balance (add id data l) k r
      else
        balance l k (add id data r)

let rec find_stamp s = function
    None ->
      raise Not_found
  | Some k ->
      if k.ident.stamp = s then k.data else find_stamp s k.previous

let rec find_same id = function
    Empty ->
      raise Not_found
  | Node(l, k, r, _) ->
      let c = compare id.name k.ident.name in
      if c = 0 then
        if id.stamp = k.ident.stamp
        then k.data
        else find_stamp id.stamp k.previous
      else
        find_same id (if c < 0 then l else r)

let rec find_name name = function
    Empty ->
      raise Not_found
  | Node(l, k, r, _) ->
      let c = compare name k.ident.name in
      if c = 0 then
        k.data
      else
        find_name name (if c < 0 then l else r)

let rec iter fn = function
    Empty -> ()
  | Node(l, k, r, _) ->
      iter fn l; iter_node fn k; iter fn r
and iter_node fn k =
  fn k.ident k.data;
  match k.previous with None -> () | Some prev_k -> iter_node fn prev_k

let print_tbl print_elt tbl =
  open_box 2;
  print_string "[[";
  iter (fun id data -> 
          open_box 2;
          print id; print_string " ->"; print_space(); print_elt data;
          print_string ";"; close_box(); print_space())
       tbl;
  print_string "]]";
  close_box()

  end

(* from typing/path.ml *)
module Path = struct
    
    type t =
      Pident of Ident.t
    | Pdot of t * string * int
    | Papply of t * t
  
  end

(* taken from typing/primitive.ml *)
module Primitive = struct
  open Format

    type description =
      { prim_name: string;         (* Name of primitive  or C function *)
        prim_arity: int;           (* Number of arguments *)
        prim_alloc: bool;          (* Does it allocates or raise? *)
        prim_native_name: string;  (* Name of C function for the nat. code gen. *)
        prim_native_float: bool }  (* Does the above operate on unboxed floats? *)


let print_quoted s = print_char '"'; print_string s; print_char '"'

let print_description p =
  print_quoted p.prim_name;
  if not p.prim_alloc then
    (print_space(); print_quoted "noalloc");
  if p.prim_native_name <> "" then
    (print_space(); print_quoted p.prim_native_name);
  if p.prim_native_float then
    (print_space(); print_quoted "float")

  end

(* from typing/types.ml *)
module Types = struct

(* Representation of types and declarations *)
    
    open Asttypes

(* Type expressions for the core language *)
    
    type type_expr =
      { mutable desc: type_desc; 
        mutable level: int;
        mutable id: int }
    
    and type_desc =
      Tvar
    | Tarrow of type_expr * type_expr
    | Ttuple of type_expr list
    | Tconstr of Path.t * type_expr list * abbrev_memo ref
    | Tobject of type_expr * (Path.t * type_expr list) option ref
    | Tfield of string * field_kind * type_expr * type_expr
    | Tnil
    | Tlink of type_expr
    
    and abbrev_memo =
      Mnil
    | Mcons of Path.t * type_expr * type_expr * abbrev_memo
    | Mlink of abbrev_memo ref
    
    and field_kind =
      Fvar of field_kind option ref
    | Fpresent
    | Fabsent

(* Maps of methods and instance variables *)


(* Maps of methods and instance variables *)
    
    module OrderedString = struct type t = string let compare = compare end
    module Meths = Map.Make(OrderedString)
    module Vars = Meths


(* Value descriptions *)
    
    type value_description =
      { val_type: type_expr;                (* Type of the value *)
        val_kind: value_kind }
    
    and value_kind =
      Val_reg                             (* Regular value *)
    | Val_prim of Primitive.description   (* Primitive *)
    | Val_ivar of mutable_flag            (* Instance variable (mutable ?) *)
    | Val_self of (Ident.t * type_expr) Meths.t ref *
      (Ident.t * Asttypes.mutable_flag * type_expr) Vars.t ref
                                        (* Self *)
    | Val_anc of (string * Ident.t) list  (* Ancestor *)
    | Val_unbound                         (* Unbound variable *)

(* Constructor descriptions *)
    
    type constructor_description =
      { cstr_res: type_expr;                (* Type of the result *)
        cstr_args: type_expr list;          (* Type of the arguments *)
        cstr_arity: int;                    (* Number of arguments *)
        cstr_tag: constructor_tag;          (* Tag for heap blocks *)
        cstr_consts: int;                   (* Number of constant constructors *)
        cstr_nonconsts: int }               (* Number of non-const constructors *)
    
    and constructor_tag =
      Cstr_constant of int                (* Constant constructor (an int) *)
    | Cstr_block of int                   (* Regular constructor (a block) *)
    | Cstr_exception of Path.t            (* Exception constructor *)

(* Record label descriptions *)
    
    type label_description =
      { lbl_res: type_expr;                 (* Type of the result *)
        lbl_arg: type_expr;                 (* Type of the argument *)
        lbl_mut: mutable_flag;              (* Is this a mutable field? *)
        lbl_pos: int;                       (* Position in block *)
        lbl_all: label_description array;   (* All the labels in this type *)
        lbl_repres: record_representation } (* Representation for this record *)
    
    and record_representation =
      Record_regular                      (* All fields are boxed / tagged *)
    | Record_float                        (* All fields are floats *)

(* Type definitions *)
    
    type type_declaration =
      { type_params: type_expr list;
        type_arity: int;
        type_kind: type_kind;
        type_manifest: type_expr option }
    
    and type_kind =
      Type_abstract
    | Type_variant of (string * type_expr list) list
    | Type_record of (string * mutable_flag * type_expr) list
    
    type exception_declaration = type_expr list

(* Type expressions for the class language *)
    
    module Concr = Set.Make(OrderedString)
    
    type class_type =
      Tcty_constr of Path.t * type_expr list * class_type
    | Tcty_signature of class_signature
    | Tcty_fun of type_expr * class_type
    
    and class_signature =
      { cty_self: type_expr;
        cty_vars: (Asttypes.mutable_flag * type_expr) Vars.t;
        cty_concr: Concr.t }
    
    type class_declaration =
      { cty_params: type_expr list;
        mutable cty_type: class_type;
        cty_path: Path.t;
        cty_new: type_expr option }
    
    type cltype_declaration =
      { clty_params: type_expr list;
        clty_type: class_type;
        clty_path: Path.t }

(* Type expressions for the module language *)
    
    type module_type =
      Tmty_ident of Path.t
    | Tmty_signature of signature
    | Tmty_functor of Ident.t * module_type * module_type
    
    and signature = signature_item list
    
    and signature_item =
      Tsig_value of Ident.t * value_description
    | Tsig_type of Ident.t * type_declaration
    | Tsig_exception of Ident.t * exception_declaration
    | Tsig_module of Ident.t * module_type
    | Tsig_modtype of Ident.t * modtype_declaration
    | Tsig_class of Ident.t * class_declaration
    | Tsig_cltype of Ident.t * cltype_declaration
    
    and modtype_declaration =
      Tmodtype_abstract
    | Tmodtype_manifest of module_type
  
  end

module Btype = struct
open Types

let generic_level = 100000000
let lowest_level = 0
let pivot_level = 2 * lowest_level - 1


let new_id = ref (-1)

let newty2 level desc  =
  incr new_id; { desc = desc; level = level; id = !new_id }
let newgenty desc      = newty2 generic_level desc
let newgenvar ()       = newgenty Tvar
let newmarkedvar level =
  incr new_id; { desc = Tvar; level = pivot_level - level; id = !new_id }
let newmarkedgenvar () =
  incr new_id;
  { desc = Tvar; level = pivot_level - generic_level; id = !new_id }

(**** Representative of a type ****)

let rec field_kind_repr =
  function
    Fvar {contents = Some kind} -> field_kind_repr kind
  | kind                        -> kind

let rec repr =
  function
    {desc = Tlink t'} ->
      (* 
         We do no path compression. Path compression does not seem to
         improve notably efficiency, and it prevents from changing a
         [Tlink] into another type (for instance, for undoing a
         unification).
      *)
      repr t'
  | {desc = Tfield (_, k, _, t')} when field_kind_repr k = Fabsent ->
      repr t'
  | t -> t



end

module Ctype = struct
  open Types
  open Btype
  
(* +++ Il faudra penser a eventuellement expanser l'abreviation *)
let rec opened_object ty =
  match (repr ty).desc with
    Tobject (t, _)     -> opened_object t
  | Tfield(_, _, _, t) -> opened_object t
  | Tvar               -> true
  | _                  -> false

let object_fields ty =
  match (repr ty).desc with
    Tobject (fields, _) -> fields
  | _                   -> assert false

let flatten_fields ty =
  let rec flatten l ty =
    let ty = repr ty in
    match ty.desc with
      Tfield(s, k, ty1, ty2) ->
        flatten ((s, k, ty1)::l) ty2
    | _ ->
        (l, ty)
  in
    let (l, r) = flatten [] ty in
    (Sort.list (fun (n, _, _) (n', _, _) -> n < n') l, r)

let unalias ty =
  let ty = repr ty in
  match ty.desc with
    Tvar ->
      ty
  | _ ->
      newty2 ty.level ty.desc

let rec signature_of_class_type =
  function
    Tcty_constr (_, _, cty) -> signature_of_class_type cty
  | Tcty_signature sign     -> sign
  | Tcty_fun (ty, cty)      -> signature_of_class_type cty

let self_type cty =
  repr (signature_of_class_type cty).cty_self

let rec class_type_arity =
  function
    Tcty_constr (_, _, cty) ->  class_type_arity cty
  | Tcty_signature _        ->  0
  | Tcty_fun (_, cty)       ->  1 + class_type_arity cty

end

(* from typing/env.ml *)
module Env = struct
    
    open Config
    open Types
    
    type error =
      Not_an_interface of string
    | Corrupted_interface of string
    | Illegal_renaming of string * string
    | Inconsistent_import of string * string * string
    
    exception Error of error
    
    type summary =
      Env_empty
    | Env_value of summary * Ident.t * value_description
    | Env_type of summary * Ident.t * type_declaration
    | Env_exception of summary * Ident.t * exception_declaration
    | Env_module of summary * Ident.t * module_type
    | Env_modtype of summary * Ident.t * modtype_declaration
    | Env_class of summary * Ident.t * class_declaration
    | Env_cltype of summary * Ident.t * cltype_declaration
    | Env_open of summary * Path.t
    
    type t = {
        values: (Path.t * value_description) Ident.tbl;
        constrs: constructor_description Ident.tbl;
        labels: label_description Ident.tbl;
        types: (Path.t * type_declaration) Ident.tbl;
        modules: (Path.t * module_type) Ident.tbl;
        modtypes: (Path.t * modtype_declaration) Ident.tbl;
        components: (Path.t * module_components) Ident.tbl;
        classes: (Path.t * class_declaration) Ident.tbl;
        cltypes: (Path.t * cltype_declaration) Ident.tbl;
        summary: summary
      }
    
    and module_components =
      Structure_comps of structure_components
    | Functor_comps of functor_components
    
    and structure_components = {
        mutable comp_values: (string, (value_description * int)) Tbl.t;
        mutable comp_constrs: (string, (constructor_description * int)) Tbl.t;
        mutable comp_labels: (string, (label_description * int)) Tbl.t;
        mutable comp_types: (string, (type_declaration * int)) Tbl.t;
        mutable comp_modules: (string, (module_type * int)) Tbl.t;
        mutable comp_modtypes: (string, (modtype_declaration * int)) Tbl.t;
        mutable comp_components: (string, (module_components * int)) Tbl.t;
        mutable comp_classes: (string, (class_declaration * int)) Tbl.t;
        mutable comp_cltypes: (string, (cltype_declaration * int)) Tbl.t
      }
    
    and functor_components = {
        fcomp_param: Ident.t;
        fcomp_arg: module_type;
        fcomp_res: module_type;
        fcomp_env: t
      }
    
    let empty = {
        values = Ident.empty; constrs = Ident.empty;
        labels = Ident.empty; types = Ident.empty;
        modules = Ident.empty; modtypes = Ident.empty;
        components = Ident.empty; classes = Ident.empty;
        cltypes = Ident.empty;
        summary = Env_empty }

(* Persistent structure descriptions *)
    
    type pers_struct =
      { ps_name: string;
        ps_sig: signature;
        ps_comps: module_components;
        ps_crcs: (string * Digest.t) list }
    
    let persistent_structures =
      (Hashtbl.create 17 : (string, pers_struct) Hashtbl.t)
    
    let read_pers_struct modname filename =
      let ic = open_in_bin filename in
      try
        let buffer = String.create (String.length cmi_magic_number) in
        really_input ic buffer 0 (String.length cmi_magic_number);
        if buffer <> cmi_magic_number then begin
            close_in ic;
            raise(Error(Not_an_interface filename))
          end;
        let (name, sign, comps) = input_value ic in
        let crcs = input_value ic in
        close_in ic;
        let ps = { ps_name = name;
            ps_sig = sign;
            ps_comps = comps;
            ps_crcs = crcs } in
        if ps.ps_name <> modname then
          raise(Error(Illegal_renaming(ps.ps_name, filename)));
        Hashtbl.add persistent_structures modname ps;
        ps
      with End_of_file | Failure _ ->
          close_in ic;
          raise(Error(Corrupted_interface(filename)))
    
    let find_pers_struct load_path name =
      try
        Hashtbl.find persistent_structures name
      with Not_found ->
          read_pers_struct name
            (Misc.find_in_path load_path (String.uncapitalize name ^ ".cmi"))
  
  end

module Typedtree = struct

(* Abstract syntax tree after typing *)
    
    open Asttypes
    open Types

(* Value expressions for the core language *)
    
    type pattern =
      { pat_desc: pattern_desc;
        pat_loc: Location.t;
        pat_type: type_expr;
        pat_env: Env.t }
    
    and pattern_desc =
      Tpat_any
    | Tpat_var of Ident.t
    | Tpat_alias of pattern * Ident.t
    | Tpat_constant of constant
    | Tpat_tuple of pattern list
    | Tpat_construct of constructor_description * pattern list
    | Tpat_record of (label_description * pattern) list
    | Tpat_array of pattern list
    | Tpat_or of pattern * pattern
    
    type expression =
      { exp_desc: expression_desc;
        exp_loc: Location.t;
        exp_type: type_expr;
        exp_env: Env.t }
    
    and expression_desc =
      Texp_ident of Path.t * value_description
    | Texp_constant of constant
    | Texp_let of rec_flag * (pattern * expression) list * expression
    | Texp_function of (pattern * expression) list
    | Texp_apply of expression * expression list
    | Texp_match of expression * (pattern * expression) list
    | Texp_try of expression * (pattern * expression) list
    | Texp_tuple of expression list
    | Texp_construct of constructor_description * expression list
    | Texp_record of (label_description * expression) list * expression option
    | Texp_field of expression * label_description
    | Texp_setfield of expression * label_description * expression
    | Texp_array of expression list
    | Texp_ifthenelse of expression * expression * expression option
    | Texp_sequence of expression * expression
    | Texp_while of expression * expression
    | Texp_for of
      Ident.t * expression * expression * direction_flag * expression
    | Texp_when of expression * expression
    | Texp_send of expression * meth
    | Texp_new of Path.t * class_declaration
    | Texp_instvar of Path.t * Path.t
    | Texp_setinstvar of Path.t * Path.t * expression
    | Texp_override of Path.t * (Path.t * expression) list
    | Texp_letmodule of Ident.t * module_expr * expression
    
    and meth =
      Tmeth_name of string
    | Tmeth_val of Ident.t

(* Value expressions for the class language *)
    
    and class_expr =
      { cl_desc: class_expr_desc;
        cl_loc: Location.t;
        cl_type: class_type }
    
    and class_expr_desc =
      Tclass_ident of Path.t
    | Tclass_structure of class_structure
    | Tclass_fun of pattern * (Ident.t * expression) list * class_expr
    | Tclass_apply of class_expr * expression list
    | Tclass_let of rec_flag *  (pattern * expression) list *
      (Ident.t * expression) list * class_expr
    | Tclass_constraint of class_expr * string list * string list * Concr.t
    (* Visible instance variables, methods and concretes methods *)
    
    and class_structure =
      { cl_field: class_field list;
        cl_meths: Ident.t Meths.t }
    
    and class_field =
      Cf_inher of class_expr * (string * Ident.t) list * (string * Ident.t) list
    (* Inherited instance variables and concrete methods *)
    | Cf_val of string * Ident.t * expression
    | Cf_meth of string * expression
    | Cf_let of rec_flag * (pattern * expression) list *
      (Ident.t * expression) list
    | Cf_init of expression

(* Value expressions for the module language *)
    
    and module_expr =
      { mod_desc: module_expr_desc;
        mod_loc: Location.t;
        mod_type: module_type;
        mod_env: Env.t }
    
    and module_expr_desc =
      Tmod_ident of Path.t
    | Tmod_structure of structure
    | Tmod_functor of Ident.t * module_type * module_expr
    | Tmod_apply of module_expr * module_expr * module_coercion
    | Tmod_constraint of module_expr * module_type * module_coercion
    
    and structure = structure_item list
    
    and structure_item =
      Tstr_eval of expression
    | Tstr_value of rec_flag * (pattern * expression) list
    | Tstr_primitive of Ident.t * value_description
    | Tstr_type of (Ident.t * type_declaration) list
    | Tstr_exception of Ident.t * exception_declaration
    | Tstr_module of Ident.t * module_expr
    | Tstr_modtype of Ident.t * module_type
    | Tstr_open of Path.t
    | Tstr_class of (Ident.t * int * string list * class_expr) list
    | Tstr_cltype of (Ident.t * cltype_declaration) list
    
    and module_coercion =
      Tcoerce_none
    | Tcoerce_structure of (int * module_coercion) list
    | Tcoerce_functor of module_coercion * module_coercion
    | Tcoerce_primitive of Primitive.description

(* Auxiliary functions over the a.s.t. *)
  end
  
