{
  open Lexing
  open Text  
  open Compile
  
  
  type token =
    ANY
  | COMMENT
  | MATHMODE
  | STARTMATH
  | ENDMATH
  | STARTTABBING
  | ENDTABBING
  | KWDTABBING
  | STARTARRAY
  | ENDARRAY
  | KWDARRAY
  | STARTENV
  | ENDENV
  | CONDITION    
  | ITEM
  | PARAMETER
  | COMMAND
  | FONT
  | EOF
  | LBRACE
  | RBRACE
  
  let lexer_start = ref 0
  let position lexbuf =
    let b = lexeme_start lexbuf in
    let e = lexeme_end lexbuf in
    b + !lexer_start, e - b
  
  let lexing text start_point end_point =
    lexer_start := get_position text start_point;
    Text.lexing text start_point end_point

}

rule token = parse
(* comments *)
    '%'+ [^ '\n']* '\n' { position lexbuf, COMMENT }
(* Math mode *)
  | "$" | "$$" { position lexbuf, MATHMODE }
  
  | "\\(" | "\\[" { position lexbuf, STARTMATH }
  
  |  "\\)" | "\\]" { position lexbuf, ENDMATH }

    (* Tabbing *)
  | "\\begin" ' '* "{tabbing}" { position lexbuf, STARTTABBING }
  | "\\end" ' '* "{tabbing}"   { position lexbuf, ENDTABBING }
  | "\\>" | "\\="
  | "\\kill"  { position lexbuf, KWDTABBING }

(* tables and array *)
  | "\\begin" ' '* ("{tabular}" | "{array}") { position lexbuf, STARTARRAY }
  | "\\end" ' '* ("{tabular}" | "{array}")   { position lexbuf, ENDARRAY }
  | "&"                                      { position lexbuf, KWDARRAY }
      
(* environments *)
  |   "\\begin" " "* "{" ['A'-'Z' 'a'-'z']+ '*'?"}"
      { position lexbuf, STARTENV }
  |  "\\end" " " * "{" ['A'-'Z' 'a'-'z']+ '*'? "}"
      { position lexbuf, ENDENV }    
  | "\\item" { position lexbuf, ITEM }
(* Ignore font definitions ... *)
  | "\\font" "\\" ['A'-'Z' 'a'-'z']+ ' '* '='? ' '* ['a'-'z' 'A'-'Z' '0'-'9']+
      { position lexbuf, FONT }
(* conditionals *)
  | "\\newif"
  | "\\else"
  | "\\fi" { position lexbuf, CONDITION }

(* General case for commands *)
  | '#' ['1'-'9'] { position lexbuf, PARAMETER }
  | "\\" (('@' ? ['A'-'Z' 'a'-'z']+ '*'?) | [^ 'A'-'Z' 'a'-'z'])
    { position lexbuf, COMMAND }
  | '{' { position lexbuf, LBRACE }
  | '}' { position lexbuf, RBRACE }
  | eof { position lexbuf, EOF }
  | _   { position lexbuf, ANY }

{
open Text
open Efuns
open Interactive
open Simple
open Abbrevs
open Keymap
open Window

let abbreviations =
  [ "\s", "\section";
    "\ss", "\subsection";
    "\sss", "\subsubsection";
    "\p", "\paragraph";
    "\begit", "\begin{itemize}";
    "\endit", "\end{itemize}";
    "\startarticle", "\\documentclass[twocolumn]{article}

%\\usepackage{isolatin1}
%\\usepackage{francais}
\\usepackage{url}
\\usepackage{epsfig}

\\newcommand{\\comment}[1]{{\\hfuzz 4pt\\par
 \\medskip\\noindent\\hskip-\\fboxsep\\hskip-\\fboxrule\\fbox{\\parbox{\\hsize}{\\it
 #1}}\\par}}

\\newcommand{\\figurebox}[1]{#1}
\\newcommand{\\smallsf}{\\small\\sffamily}
\\newcommand{\\code}{\\texttt}
\\newcommand{\\func}{\\textsf}
\\newcommand{\\msg}{\\textsf}

\\newcommand{\\psfigure}[3]{ % {scale}{filename=label}{caption}
 \\begin{figure}[t]\\begin{center}
 \\figurebox{\\epsfig{file=figs/#2.eps,width=#1\\hsize}}
 \\begin{quote}\\let\\normalsize\\small\\caption{#3\\label{fig:#2}}\\end{quote}
 \\end{center}\\end{figure}}

\\begin{document}

\\title{Change this} 

\\bibliographystyle{plain}

\\author{
        Fabrice Le Fessant\\thanks{
            Ph.D student at \\'Ecole Polytechnique.  Also with INRIA
            projet PARA\\@.
            }
        ~\\\\[2ex]
        INRIA Projet PARA \\\\
        Domaine de Voluceau, Rocquencourt, B.P. 105 \\\\
        78153 Le Chesnay C\\'edex, France \\\\
        \\url{http://pauillac.inria.fr/\~lefessan} \\\\
}

\\maketitle

\\abstract{Here your abstract}

Here you text
\\bibliography{lefessan,bib,gcbib}
\\end{document}
";
  ]

type mode =
  Math_mode
| Arg_mode
| Normal_mode

let tex_color_buffer buf =
  let location = buf.buf_location in
  let red_attr = make_attr (get_color location "red") 1 0 false in
  let yellow_attr = make_attr (get_color location "yellow") 1 0 false in
  let blue_attr = make_attr (get_color location "blue") 1 0 false in
  let gray_attr = make_attr (get_color location "gray") 1 0 false in
  let text = buf.buf_text in
  let start_point = Text.add_point text in
  let curseur = add_point text in
  let end_point = add_point text in 
  set_position text end_point (size text);
  let lexbuf = lexing text start_point end_point in
  let rec iter mode lexbuf =
    let (pos,len), token = token lexbuf in
    let mode =
      match token with
        EOF -> 
          raise Exit
      | COMMENT
        ->
          set_position text curseur pos;
          set_attr text curseur len blue_attr;
          mode
      | MATHMODE ->
          set_position text curseur pos;
          set_attr text curseur len yellow_attr;
          (match mode with
              Math_mode -> Normal_mode
            | _ -> Math_mode)
      | STARTMATH ->
          set_position text curseur pos;
          set_attr text curseur len yellow_attr;
          Math_mode          
      | ENDMATH
        ->
          set_position text curseur pos;
          set_attr text curseur len yellow_attr;
          Normal_mode
      | STARTTABBING
      | ENDTABBING
      | KWDTABBING
        ->
          set_position text curseur pos;
          set_attr text curseur len gray_attr;
          mode
      | STARTARRAY
      | ENDARRAY
      | KWDARRAY
      | STARTENV
      | ENDENV
      | CONDITION    
      | ITEM
      | PARAMETER
      | COMMAND
      | FONT ->
          set_position text curseur pos;
          set_attr text curseur len red_attr;
          mode
      | LBRACE ->
          set_position text curseur pos;
          set_attr text curseur len red_attr;
          Arg_mode
      | RBRACE ->
          set_position text curseur pos;
          set_attr text curseur len red_attr;
          Normal_mode
      | _ -> 
          match mode with
            Math_mode -> 
              set_position text curseur pos;
              set_attr text curseur len yellow_attr;
              mode
          | Arg_mode ->
              set_position text curseur pos;
              set_attr text curseur len gray_attr;
              mode
          | _ -> mode
    in
    iter mode lexbuf
  in
  try
    iter Normal_mode lexbuf
  with
    _ ->
      buf.buf_modified <- buf.buf_modified + 1;
      remove_point text curseur;
      remove_point text end_point;
      remove_point text start_point


let c_c = (ControlMap,Char.code 'c')

let structures = [
    [c_c;NormalMap, Char.code '0'], "\\documentclass[twocolumn^^]{article}\n%^^\\usepackage{isolatin1}\n%^^\\usepackage{francais}\n\\usepackage{url}\n\\usepackage[dvips]{epsfig}\n\n\\begin{document}\n\\title{^^}\n\\bibliographystyle{plain}\n\author{^^}\n\\maketitle\n\\abstract{^^}\n\\section{Introduction}\n^^\n\\bibliography{^^}\n\\end{document}\n";
    [c_c;NormalMap, Char.code '1'], "\\section{^^}\n^^";
    [c_c;NormalMap, Char.code '2'], "\\subsection{^^}\n^^";
    [c_c;NormalMap, Char.code '3'], "\\subsubsection{^^}\n^^";
    [c_c;ControlMap, Char.code '1'], "\\section*{^^}\n^^";
    [c_c;ControlMap, Char.code '2'], "\\subsection*{^^}\n^^";
    [c_c;ControlMap, Char.code '3'], "\\subsubsection*{^^}\n^^";
    [c_c;NormalMap, Char.code 'e'], "\\begin{^^}\n^^\n\\end{^^}\n^^";
    [c_c; NormalMap, Char.code 'i'], "\\begin{itemize}\n\\item ^^\n\\end{itemize}";    
    [c_c; NormalMap, Char.code 'v'], "\\begin{verbatim}\n ^^\n\\end{verbatim}";    
  ]

let tex_error_regexp = Str2.regexp "^l.\([0-9]+\)"
let tex_error_regexp2 = Str2.regexp "^LaTeX Warning:.*line +\([0-9]+\)"  
  
let backward_find_unclosed_paren text point =
  let rec iter n =
    if bmove text point 1 = 0 then raise Not_found;
    let d = get_char text point in
    if d = ')' then iter (n+1) else
    if d = '(' then
      (if n > 0 then iter (n-1))
    else iter n
  in
  iter 0

let filename_after_point text point =
  let endp = dup_point text point in
  let rec iter () = 
    if fmove text endp 1 = 1 &&
      (let d = get_char text endp in
        d <> '\n' && d <> ')') then iter ()
  in
  iter ();
  let filename = Text.sub text point (distance text point endp) in
  remove_point text endp;
  filename

  
let tex_find_error text error_point = 
  let groups = 
    try
      Text.search_forward_groups text tex_error_regexp
        error_point 1 
    with
      Not_found ->
        Text.search_forward_groups text tex_error_regexp2
          error_point 1 
  in
  let point = dup_point text error_point in
  let filename = 
    try
      backward_find_unclosed_paren text point;
      fmove text point 1;
      filename_after_point text point
    with
      Not_found -> ""
  in
  remove_point text point;
  let error =
    { err_msg = Text.get_position text error_point;
      err_filename = filename;
      err_line = int_of_string groups.(0) - 1;
      err_begin = 0;
      err_end = 0
      }
    in
  (* pass two lines *)
  let eol = point_to_eol text error_point in
  fmove text error_point (eol+1);
  let eol = point_to_eol text error_point in
  fmove text error_point (eol+1);
  error

let install buf =
  let map = buf.buf_local_map in
  Keymap.add_prefix map [c_c];
  interactive (Buffer buf) [c_c;ControlMap, Char.code 'l']
    "tex-color-buffer" (fun frame -> tex_color_buffer frame.frm_buffer);
  interactive (Buffer buf) [c_c;ControlMap, Char.code 'c']
    "tex-compile" (compile tex_find_error);
  Keymap.add_binding map [NormalMap, Char.code '{'] (fun frame ->
      expand_sabbrev frame; self_insert_command frame);
  Keymap.add_binding map [NormalMap, Char.code ' '] 
    (fun frame ->
      expand_sabbrev frame;
      electric_insert_space frame);
  Keymap.add_binding map [MetaMap, Char.code 'q'] fill_paragraph;
  tex_color_buffer buf; 
  Accents_mode.install buf;
  List.iter (fun char ->
      Keymap.add_binding map [NormalMap, Char.code char]
        (fun frame ->
          self_insert_command frame;
          highlight_paren frame)
  ) ['}';']';')'];
  (* les \ et ' font parties des mots *)
  buf.buf_syntax_table.(Char.code '\\') <- true;
  buf.buf_syntax_table.(Char.code '\'') <- true;
  buf.buf_syntax_table.(Char.code '-') <- true;
  let abbrevs =
    try
      get_local_var (Buffer buf) abbrev_table
    with
      Failure _ -> 
        let abbrevs = Hashtbl.create 11 in
        set_local_var (Buffer buf) abbrev_table abbrevs;
        abbrevs
  in
  Misc.hash_add_assoc abbrevs abbreviations;
  install_structures buf structures;
  Buffer.set_major_mode buf "TeX";
  ()


let _ =  
  Efuns.add_start_hook (fun location ->
      let alist = get_local_var (Location location) Buffer.modes_alist in
      set_local_var (Location location) Buffer.modes_alist 
        ((".*\.tex",install):: alist)
  )

} 
