(***********************************************************************)
(*                                                                     *)
(*                           xlib for Ocaml                            *)
(*                                                                     *)
(*       Fabrice Le Fessant, projet Para/SOR, INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Xtypes
open Xbuffer
open Display

let addbit mask bit = 
  if mask land (1 lsl bit) == 0 then
    mask lor (1 lsl bit)
  else
    failwith "Mask of list: twice the same arg"

let setWinfowAttributes b pos list =
  let rec iter pos list =
    match list with
      [] -> 0
    | head :: tail -> 
        let mask = iter (pos+4) tail in
        match head with
        | CWBackPixmap pixmap -> 
            setEnum32 b pos pixmap; 
            addbit mask 0
        | CWBackPixel pixel ->
            setEnum32 b pos pixel; 
            addbit mask 1
        | CWBorderPixmap pixmap ->
            setEnum32 b pos pixmap; 
            addbit mask 2
        | CWBorderPixel pixel ->
            setEnum32 b pos pixel; 
            addbit mask 3
        | CWBitGravity bitGravity ->
            setEnum8 b pos bitGravity; 
            addbit mask 4
        | CWWinGravity bitGravity ->
            setEnum8 b pos bitGravity; 
            addbit mask 5
        | CWBackingStore backingStore ->
            setEnum8 b pos backingStore; 
            addbit mask 6
        | CWBackingPlanes planes ->
            setEnum32 b pos planes; 
            addbit mask 7
        | CWBackingPixel pixel ->
            setEnum32 b pos pixel; 
            addbit mask 8
        | CWOverrideRedirect bool ->
            setEnum8 b pos bool; 
            addbit mask 9
        | CWSaveUnder bool ->
            setEnum8 b pos bool; 
            addbit mask 10
        | CWEventMask eventMask ->
            setEnum32 b pos (mask_of_list eventMask); 
            addbit mask 11
        | CWDontPropagate eventMask ->
            setEnum32 b pos (mask_of_list eventMask); 
            addbit mask 12
        | CWColormap colormap ->
            setEnum32 b pos colormap; 
            addbit mask 13
        | CWCursor cursor ->
            setEnum32 b pos cursor; 
            addbit mask 14
  in
  setCard32 b pos (iter (pos+4) (Sort.list (<) list))

let setConfigureWindow b pos list =
  let rec iter pos list =
    match list with
      [] -> 0
    | head :: tail -> 
        let mask = iter (pos+4) tail in
        match head with
          CWX int ->
            setEnum16 b pos int;
            addbit mask 0
        | CWY int ->
            setEnum16 b pos int;
            addbit mask 1
        | CWWidth int ->
            setEnum16 b pos int;
            addbit mask 2
        | CWHeight int ->
            setEnum16 b pos int;
            addbit mask 3
        | CWBorderWidth int ->
            setEnum16 b pos int;
            addbit mask 4
        | CWSibling window ->
            setEnum32 b pos window;
            addbit mask 5
        | CWStackMode stackMode ->
            setEnum8 b pos stackMode;
            addbit mask 6
  in
  setCard16 b pos (iter (pos+4) (Sort.list (<) list))


let setGCattributes b pos list =
  let rec iter pos list =
    match list with
      [] -> 0
    | head :: tail -> 
        let mask = iter (pos+4) tail in
        match head with
        | GCfonction gxFunction ->
            setEnum8 b pos gxFunction;
            addbit mask 0
        | GCplane_mask int  ->
            setEnum32 b pos int;
            addbit mask 1
        | GCforeground pixel   ->
            setEnum32 b pos pixel;
            addbit mask 2
        | GCbackground pixel   ->
            setEnum32 b pos pixel;
            addbit mask 3
        | GCline_width int   ->
            setEnum16 b pos int;
            addbit mask 4
        | GCline_style lineStyle   ->
            setEnum8 b pos lineStyle;
            addbit mask 5
        | GCcap_style capStyle   ->
            setEnum8 b pos capStyle;
            addbit mask 6
        | GCjoin_style joinStyle   ->
            setEnum8 b pos joinStyle;
            addbit mask 7
        | GCfill_style fillStyle   ->
            setEnum8 b pos fillStyle;
            addbit mask 8
        | GCfill_rule fillRule   ->
            setEnum8 b pos fillRule;
            addbit mask 9
        | GCtile pixmap   ->
            setEnum32 b pos pixmap;
            addbit mask 10
        | GCstipple pixmap   ->
            setEnum32 b pos pixmap;
            addbit mask 11
        | GCts_x_origin coord   ->
            setEnum16 b pos coord;
            addbit mask 12
        | GCts_y_origin coord   ->
            setEnum16 b pos coord;
            addbit mask 13
        | GCfont font   ->
            setEnum32 b pos font;
            addbit mask 14
        | GCsubwindow_mode subwindowMode   ->
            setEnum8 b pos subwindowMode;
            addbit mask 15
        | GCgraphics_exposures bool  ->
            setEnum8 b pos bool;
            addbit mask 16
        | GCclip_x_origin coord   ->
            setEnum16 b pos coord;
            addbit mask 17
        | GCclip_y_origin coord   ->
            setEnum16 b pos coord;
            addbit mask 18
        | GCclip_mask pixmap  ->
            setEnum32 b pos pixmap;
            addbit mask 19
        | GCdash_offset int   ->
            setEnum16 b pos int;
            addbit mask 20
        | GCdashes int  ->
            setEnum8 b pos int;
            addbit mask 21
        | GCarc_mode arcMode   ->
            setEnum8 b pos arcMode;
            addbit mask 22
  in
  setCard32 b pos (iter (pos+4) (Sort.list (<) list))

let rec strListLength = function
	 [] -> 0
       | nom::tl ->
	      1+(String.length nom)+(strListLength tl)

let getCharInfo buffer pos =
    {
    char_lbearing = getInt16 buffer pos;
    char_rbearing = getInt16 buffer (pos+2);
    char_width = getInt16 buffer (pos+4);
    char_ascent = getInt16 buffer (pos+6);
    char_descent = getInt16 buffer (pos+8);
    char_attributes = getCard16 buffer (pos+10)
   }

let getFontProp buffer pos =
    (getCard32 buffer pos,getCard32 buffer (pos+4))

let getOneFontInfo b =
   {
  font_min_bounds = getCharInfo b 8;
  font_max_bounds = getCharInfo b 24;
  font_min_char_or_byte2 = getCard16 b 40;
  font_max_char_or_byte2 = getCard16 b 42;
  font_default_char = getCard16 b 44;
  font_draw_direction = getEnum8 b 48;
  font_min_byte1 = getCard8 b 49;
  font_max_byte1 = getCard8 b 50;
  font_all_char_exist = getEnum8 b 51;
  font_ascent = getInt16 b 52;
  font_descent = getInt16 b 54;
  font_properties = 
  (let (p,n) = (ref [],getCard16 b 46)
  in
  for i=0 to n-1 do
    p:=(getFontProp b (60+i*8))::(!p)
  done; !p)              
}

let copyDatas b gc src src_x src_y dst dst_x dst_y dx dy =
    setCard32 b 4 src;
    setCard32 b 8 dst;
    setCard32 b 12 gc;
    setCard16 b 16 src_x;
    setCard16 b 18 src_y;
    setCard16 b 20 dst_x;
    setCard16 b 22 dst_y;
    setCard16 b 24 dx;
    setCard16 b 26 dy

let drawPoly setOne d opcode window gc list =
  let n = List.length list  in
  let b = newWinRequest opcode (3+(n*d)) window
  in
  setCard32 b 8 gc;
  let rec iter pos list =
    match list with 
      [] -> ()
    | head :: tail ->
        setOne b pos head;
        iter (pos+4*d) tail
  in
  iter 12 list;
  b

let setOnePoint  b pos (x,y) =
    setCard16 b pos x;
    setCard16 b (pos+2) y


let setOneSegment = fun b pos
    (x1,y1,x2,y2) ->
    setCard16 b pos x1;
    setCard16 b (pos+2) y1;
    setCard16 b (pos+4) x2;
    setCard16 b (pos+6) y2

   
let setOneArc b pos (x1,y1,x2,y2,start,stop) =
      setCard16 b pos x1;
      setCard16 b (pos+2) y1;
      setCard16 b (pos+4) x2;
      setCard16 b (pos+6) y2;
      setCard16 b (pos+8) start;
      setCard16 b (pos+10) stop
        
   
let textItemLength = function
    Text8 (delta,string,pos,len) -> 2+len
  | Text16 (tab,delta) -> 2+2*(Array.length tab)
  | ShiftFont fid -> 5 


let rec textItemListLength = function
    [] -> 0
  | h::tl -> 
      (textItemLength h)+(textItemListLength tl)

let rec setTextItem size b pos items = 
  let rec iter pos items =
    match items with
      [] -> ()
    | (Text8 (delta,string,spos,len)) :: tail when size = 8 -> 
        setCard8 b pos len;
        setCard8 b (pos+1) delta;
        String.blit string spos b (pos+2) len;
        iter (pos+2+len) tail
    | (Text16 (tab,delta)) :: tail when size = 16 ->
        let n = Array.length tab
        in
        setCard8 b pos n;
        setCard8 b (pos+1) delta;
        setString16 b (pos+2) tab;
        iter (pos+2+2*n) tail
    | (ShiftFont fid) :: tail ->
        setCard8 b pos 255;
        setCard8 b (pos+1) (fid lsr 24);
        setCard8 b (pos+2) ((fid lsr 16) land 255);
        setCard8 b (pos+3) ((fid lsr 8) land 255);
        setCard8 b (pos+4) (fid land 255);
        iter (pos+5) tail
    | _ -> raise (Invalid_argument "setTextItem")
  in
  iter pos items

(* REquest + REply *)

let createWindowReq 
      wid parent x y dx dy depth wa_class visual border_width swa =
  let n = List.length swa in
  let b = newWinRequest X_CreateWindow (8+n) wid
  in
  setCard8 b 1 depth;
  setCard32 b 8 parent;
  setCard16 b 12 x;
  setCard16 b 14 y;
  setCard16 b 16 dx;
  setCard16 b 18 dy;
  setCard16 b 20 border_width;
  setEnum16 b 22 wa_class;
  setCard32 b 24 visual;
  setWinfowAttributes b 28 swa;
  b

let changeWindowAttributesReq window swa =
  let n = List.length swa in
  let b = newWinRequest X_ChangeWindowAttributes (3+n) window
  in
  setWinfowAttributes b 8 swa;
  b

let getWindowAttributesReq = simpleRequest X_GetWindowAttributes

let getWindowAttributesRep wa =
  { gwa_backing_store = getEnum8 wa 1;
    gwa_visual  = getCard32 wa 8;
    gwa_wa_class = getEnum16 wa 12;
    gwa_bit_gravity = getEnum8 wa 14;
    gwa_win_gravity = getEnum8 wa 15;
    gwa_backing_planes = getCard32 wa 16;
    gwa_backing_pixel = getCard32 wa 20;
    gwa_save_under  = getEnum8 wa 24;
    gwa_map_is_installed = getEnum8 wa 25;
    gwa_map_state  = getEnum8 wa 26;
    gwa_override_redirect = getEnum8 wa 27;
    gwa_colormap  = getCard32 wa 28;
    gwa_all_event_mask = list_of_mask (getCard32 wa 32);
    gwa_your_event_mask = list_of_mask (getCard32 wa 36);
    gwa_dont_propagate_mask = list_of_mask (getCard16 wa 40);
  }
    
let destroyWindowReq = simpleRequest X_DestroyWindow

let destroySubwindowsReq = simpleRequest X_DestroySubwindows

let changeSaveSetReq window op =
  let b = newWinRequest X_ChangeSaveSet 2 window
  in
  setEnum8 b 1 op;
  b

let reparentWindowReq window parent x y =
  let b = newWinRequest X_ReparentWindow 4 window
  in
  setCard32 b 8 parent;
  setCard16 b 12 x;
  setCard16 b 14 y;
  b

let mapWindowReq = simpleRequest X_MapWindow 
    
let mapSubwindowsReq = simpleRequest X_MapSubwindows

let unmapWindowReq = simpleRequest X_UnmapWindow

let unmapSubwindowsReq = simpleRequest X_UnmapSubwindows

let configureWindowReq win cw =
  let n = List.length cw in
  let b = newWinRequest X_ConfigureWindow (3+n) win
  in
  setConfigureWindow b 8 cw;
  b

let circulateWindowReq (win : window) (sens : direction) =
    let b = newWinRequest X_CirculateWindow 2 win in
    setEnum8 b 1 sens;
    b

let getGeometryReq win = newWinRequest X_GetGeometry 2 win
let getGeometryRep r =
  {
  gg_root = getCard32 r 8;
  gg_x = getInt16 r 12;
  gg_y = getInt16 r 14;
  gg_width = getCard16 r 16;
  gg_height =getCard16 r 18;
  gg_border_width = getCard16 r 20;
  gg_depth = getCard8 r 1;
} 

let queryTreeReq win = newWinRequest X_QueryTree 2 win
let queryTreeRep r =
  let n = getCard16 r 16 in
  let list = ref [] in
  for i = 1 to n do
    list := (getCard32 r ((7+i)*4))::(!list)
  done;
  { qt_root = getCard32 r 8;
    qt_parent = getCard32 r 12;
    qt_subwindows = !list;
  } 

let internAtomReq string only_if_exists =
  let n = String.length string
  in
  let b = newRequest X_InternAtom (2+(strLen n))
  in
  setEnum8 b 1 only_if_exists;
  setString b 8 string;
  setCard16 b 4 n;
  b

let internAtomRep r = getCard32 r 8

let getAtomNameReq atom =
  let b = newRequest X_GetAtomName 2
  in
  setCard32 b 4 atom;
  b

let getAtomNameRep r = getString r 32 (getCard16 r 8)

let changePropertyReq window mode property property_type format buffer =
  let n = String.length buffer
  in
  if (n mod format)<>0 then
    raise (Invalid_argument "GetProperty: incompatible format")
  else
    let b = newWinRequest X_ChangeProperty (6+(strLen n)) window
    in
    setEnum8 b 1 mode;
    setCard32 b 8 property;
    setCard32 b 12 property_type;
    setCard8 b 16 (format*8);
    setCard32 b 20 (n/format);
    setString b 24 buffer;
    b

let deletePropertyReq win (property:atom) =
  let b = newWinRequest X_DeleteProperty 3 win
  in
  setCard32 b 8 property;
  b
      
let getPropertyReq win mode property property_type offset lenght =
  let b = newWinRequest X_GetProperty 6 win
  in
  setEnum8 b 1 mode;
  setCard32 b 8 property;
  setCard32 b 12 property_type;
  setCard32 b 16 offset;
  setCard32 b 20 lenght;	
  b
    
let getPropertyRep r = 
  let t = getCard32 r 8
  and format = (getCard8 r 1)/8
  and reste = getCard32 r 12
  in
  if t=0 then
    raise Not_found
  else
    { gp_type = t;
      gp_format = format;
      gp_length = getCard32 r 16;
      gp_value = getString r 32 ((getCard32 r 16) *format);
      gp_left = reste;
    }

let listPropertiesReq = simpleRequest X_ListProperties

let listPropertiesRep r =
  let n = getCard16 r 8
  in
  let list = ref []
  in
  for i = 1 to n do
    list := (getCard32 r ((7+i)*4))::(!list)
  done;
  !list

let setSelectionOwnerReq win (selection : atom) (time : time) =
  let b = newWinRequest X_SetSelectionOwner 4 win
  in
  setCard32 b 8 selection;
  setCard32 b 12 time;
  b

let getSelectionOwnerReq = simpleRequest X_GetSelectionOwner

let getSelectionOwnerRep r = getCard32 r 8

let convertSelectionReq win (selection : atom)
    (target : atom) (property : atom) (time : time) =
  let b = newWinRequest X_ConvertSelection 6 win
  in
  setCard32 b 8 selection;
  setCard32 b 12 target;
  setCard32 b 16 property;
  setCard32 b 20 time;
  b

let sendEventReq win (propagate : bool)
    (event_mask : eventMask list) (event : event) =
  let b = newWinRequest X_SendEvent 11 win
  in
  setEnum8 b 1 propagate;
  setCard32 b 8 (mask_of_list event_mask);
  setString b 12 (Conv_event.convertEvent2Core event);
  b

let grabPointerReq win owner_events event_mask pointer_mode
    keyboard_mode confine_to cursor time =
  let b = newWinRequest X_GrabPointer 6 win
  in
  setEnum8 b 1 owner_events;
  setCard32 b 4 win;
  setCard16 b 8 (mask_of_list event_mask);
  setEnum8 b 10 pointer_mode;
  setEnum8 b 11 keyboard_mode;
  setCard32 b 12 confine_to;
  setCard32 b 16 cursor;
  setCard32 b 20 time;
  b

let grabPointerRep r =
  let status = getEnum8 r 1
  in
  if status <>0 then
    raise (GrabError (Obj.magic status))

let ungrabPointerReq = simpleRequest X_UngrabPointer

let grabButtonReq win owner_events event_mask pointer_mode
    keyboard_mode confine_to cursor button modifiers =
  let b = newWinRequest X_GrabButton 6 win
  in
  setEnum8 b 1 owner_events;
  setCard16 b 8 (mask_of_list event_mask);
  setEnum8 b 10 pointer_mode;
  setEnum8 b 11 keyboard_mode;
  setCard32 b 12 confine_to;
  setCard32 b 16 cursor;
  setEnum8 b 20 (int_of_enum button);
  setCard16 b 22 modifiers;
  b

let ungrabButtonReq win button modifiers =
  let b = newWinRequest X_UngrabButton 3 win
  in
  setEnum8 b 1 button;
  setCard16 b 8 modifiers;
  b

let changeActivePointerGrabReq event_mask cursor time =
  let b = newRequest X_ChangeActivePointerGrab 4
  in
  setCard32 b 4 cursor;
  setCard32 b 8 time;
  setCard16 b 12 (mask_of_list event_mask);
  b

let grabKeyboardReq win owner_events pointer_mode keyboard_mode time =
  let b = newWinRequest X_GrabKeyboard 4 win
  in
  setEnum8 b 1 owner_events;
  setCard32 b 8 time;
  setEnum8 b 12 pointer_mode;
  setEnum8 b 13 keyboard_mode;
  b

let grabKeyboardRep = grabPointerRep

let ungrabKeyboardReq = simpleRequest X_UngrabKeyboard

let grabKeyReq win owner_events pointer_mode keyboard_mode key modifiers =
  let b = newWinRequest X_GrabKey 4 win
  in
  setEnum8 b 1 owner_events;
  setCard16 b 8 modifiers;
  setCard8 b 10 key;
  setEnum8 b 11 pointer_mode;
  setEnum8 b 12 keyboard_mode;
  b

let ungrabKeyReq win key modifiers =
  let b = newWinRequest X_UngrabKey 3 win
  in
  setCard8 b 1 key;
  setCard16 b 8 modifiers;
  b

let allowEventsReq mode time =
  let b= newRequest X_AllowEvents 2
  in
  setEnum8 b 1 mode;
  setCard32 b 4 time;
  b

let grabServerReq () = emptyRequest X_GrabServer

let ungrabServerReq () = emptyRequest X_UngrabServer

let queryPointerReq = simpleRequest X_QueryPointer

let queryPointerRep r =
  if (getEnum8 r 1)=0 then
    raise OutOfScreen;
  { qp_root = getCard32 r 8;
    qp_win = getCard32 r 12;
    qp_root_x = getInt16 r 16;
    qp_root_y = getInt16 r 18;
    qp_win_x = getInt16 r 20;
    qp_win_y = getInt16 r 22;
    qp_modifiers = list_of_mask (getCard16 r 24) 
  } 

let getMotionEventsReq win start stop =
  let b = newWinRequest X_GetMotionEvents 3 win
  in
  setCard32 b 8 start;
  setCard32 b 12 stop;
  b

let getMotionEventsRep r =
  let  list = ref []
  in 
  let n = getCard32 r 8
  in
  for i = 1 to n do
    list := { me_win = getCard32 r (24+i*8);
	      me_x = getInt16 r (26+i*8);
	      me_y = getInt16 r (28+i*8);
            } 
       ::(!list)
  done;
  !list

let translateCoordinatesReq src_window dst_window x y =
  let b = newWinRequest X_TranslateCoordinates 4 src_window
  in
  setCard32 b 8 dst_window;
  setCard16 b 12 x;
  setCard16 b 14 y;
  b

let translateCoordinatesRep r =
  if (getEnum8 r 1)=0 then
    raise OutOfScreen;
  { me_win =getCard32 r 8;
    me_x = getInt16 r 12;
    me_y = getInt16 r 14 }

let warpPointerReq src_window src_x src_y src_width src_height
    dst_window x y =
  let b = newWinRequest X_WarpPointer 6 src_window
  in
  setCard32 b 8 dst_window;
  setCard16 b 12 src_x;
  setCard16 b 14 src_y;
  setCard16 b 16 src_width;
  setCard16 b 18 src_height;
  setCard16 b 20 x;
  setCard16 b 22 y;
  b

let setInputFocusReq window revert_to time =
  let b = newWinRequest X_SetInputFocus 3 window
  in
  setEnum8 b 1 revert_to;
  setCard32 b 8 time;
  b

let getInputFocusReq () = emptyRequest X_GetInputFocus

let getInputFocusRep r = { gif_win = getCard32 r 8;
                           gif_mode = getEnum8 r 1 }

let queryKeymapReq () =  emptyRequest X_QueryKeymap

let queryKeymapRep r =
  let (t: int array) = Array.create 32 0
  in
  for i=0 to 31 do
    t.(i) <- getCard8 r (i+8)
  done;
  t

let openFontReq fid name =
  let n = String.length name
  in
  let b = newRequest X_OpenFont (3+(strLen n))
  in
  setString b 12 name;
  setCard32 b 4 fid;
  setCard16 b 8 n;
  b

let closeFontReq = simpleRequest X_CloseFont 

let queryFontReq = simpleRequest X_QueryFont

let queryFontRep b =
  let nchars = getCard32 b 56 in
  let chars = Array.init nchars 
      (fun i -> getCharInfo b (60+(getCard16 b 46)*8+12*i))
  in
  { qf_info = getOneFontInfo b;
    qf_chars = chars }

let queryTextExtentsReq fid string =
  let n = String.length string
  in
  let b = newRequest X_QueryTextExtents (2+(strLen (n*2)))
  in
  setCard8 b 1 1;
  setCard32 b 4 fid;
  for i=0 to (n-1) do
    setEnum16 b (8+2*i) string.[i]
  done;
  b
	
let queryTextExtentsRep r =
  { qte_direction =	getEnum8 r 1;
    qte_f_ascent = getInt16 r 8;
    qte_f_descent = getInt16 r 10;
    qte_ovll_ascent = getInt16 r 12;
    qte_ovll_descent = getInt16 r 14;
    qte_ovll_width = getInt32 r 16;
    qte_ovll_left = getInt32 r 20;
    qte_ovll_right = getInt32 r 24;
  } 

let listFontsReq pattern max_names =
  let n = String.length pattern in
  let b = newRequest X_ListFonts (2+(strLen n))
  in
  setCard16 b 4 max_names;
  setCard16 b 6 n;
  setString b 8 pattern;
  b

let listFontsRep r = get_str_list r 32 (getCard16 r 8)

let listFontsWithInfoReq pattern max_names = 
  failwith "listFontsWithInfoReq: Not implemented"
let listFontsWithInfoRep r = 
  failwith "listFontsWithInfoRep: Not implemented"

let setFontPathReq list =
  let n = strListLength list
  in
  let b = newRequest X_SetFontPath (2+(strLen n))
  in
  setCard16 b 4 (List.length list);
  set_str_list b 8 list;
  b

let getFontPathReq () = emptyRequest X_GetFontPath

let getFontPathRep r = get_str_list r 32 (getCard16 r 8)

let createPixmapReq pid window dx dy depth =
  let b = newRequest X_CreatePixmap 4
  in 
  setCard8 b 1 depth;
  setCard32 b 4 pid;
  setCard32 b 8 window;
  setCard16 b 12 dx;
  setCard16 b 14 dy;
  b

let freePixmapReq = simpleRequest X_FreePixmap

let createGCReq gcid win sgca =
  let n = List.length sgca in
  let b = newRequest X_CreateGC (4+n)
  in
  setCard32 b 4 gcid;
  setCard32 b 8 win;
  setGCattributes b 12 sgca;
  b

let changeGCReq gcid sgca =
  let n = List.length sgca in
  let b = newRequest X_ChangeGC (3+n)
  in
  setCard32 b 4 gcid;
  setGCattributes b 8 sgca;
  b

let copyGCReq src dst mask =
  let b = newRequest X_CopyGC 4
  in
  setCard32 b 4 src;
  setCard32 b 8 dst;
  setCard32 b 12 (mask_of_list mask);
  b

let setDashesReq gc dash_offset dashes =
  let n = String.length dashes in
  let b = newRequest X_SetDashes (3+(strLen n))
  in
  setCard32 b 4 gc;
  setCard16 b 8 dash_offset;
  setCard16 b 10 n;
  setString b 12 dashes;
  b

let setClipRectanglesReq gc ordering clip_x clip_y rects =
  let n = List.length rects in
  let b = newRequest X_SetClipRectangles (3+n*2)
  in
  setEnum8 b 1 ordering;
  setCard32 b 4 gc;
  setCard16 b 8 clip_x;
  setCard16 b 10 clip_y;
  let rec iter pos list =
    match list with
      [] -> ()
    | (x,y,dx,dy) :: tail ->
        setCard16 b pos x;
        setCard16 b (pos+2) y;
        setCard16 b (pos+4) dx;
        setCard16 b (pos+6) dy;
        iter (pos+8) tail
  in
  iter 12 rects;
  b

let freeGCReq = simpleRequest X_FreeGC

let clearAreaReq window x y dx dy exposures =
  let b =  newWinRequest X_ClearArea 4 window
  in
  setEnum8 b 1 exposures;
  setCard16 b 8 x;
  setCard16 b 10 y;
  setCard16 b 12 dx;
  setCard16 b 14 dy;
  b

let copyAreaReq gc src src_x src_y dst dst_x dst_y dx dy =
  let b = newRequest X_CopyArea 7
  in
  copyDatas b gc src src_x src_y dst dst_x dst_y dx dy;
  b

let copyPlaneReq gc src src_x src_y dst dst_x dst_y dx dy planes =
  let b = newRequest X_CopyPlane 8
  in
  setCard32 b 28 planes;
  copyDatas b gc src src_x src_y dst dst_x dst_y dx dy;
  b

let polyPointReq = drawPoly setOnePoint 1 X_PolyPoint

let polyLineReq = drawPoly setOnePoint 1 X_PolyLine

let polySegmentReq = drawPoly setOneSegment 2 X_PolySegment

let polyRectangleReq = drawPoly setOneSegment 2 X_PolyRectangle

let polyArcReq = drawPoly setOneArc 3 X_PolyArc

let fillPolyReq window gc shape coord_mode points =
  let n = List.length points
  in
  let b = newWinRequest X_FillPoly (4+n*4) window
  in
  setCard32 b 8 gc;
  setEnum8 b 12 shape;
  setEnum8 b 14 coord_mode;
  let rec iter pos list =
    match list with
      [] -> ()
    | point :: tail ->
	setOnePoint b pos point;
        iter (pos+4) tail
  in
  iter 16 points;
  b

let polyFillRectangleReq = drawPoly setOneSegment 2 X_PolyFillRectangle

let polyFillArcReq = drawPoly setOneArc 3 X_PolyFillArc

let putImageReq gc window x y dx dy left_pad depth format image =
    let n = String.length image
    in
    let b = newWinRequest X_PutImage (6+(strLen n)) window
    in
    setEnum8 b 1 format;
    setCard32 b 8 gc;
    setCard16 b 12 dx;
    setCard16 b 14 dy;
    setCard16 b 16 x;
    setCard16 b 18 y;
    setCard8  b 20 left_pad;
    setCard8  b 21 depth;
    setString b 24 image;
    b

let getImageReq window x y dx dy plane_mask format =
  let b = newWinRequest X_GetImage 5 window
  in
  setEnum8 b 1 format;
  setCard16 b 8 x;
  setCard16 b 10 y;
  setCard16 b 12 dx;
  setCard16 b 14 dy;
  setCard32 b 16 plane_mask;
  b
	
let getImageRep r =
  { gi_depth = getCard8 r 1;
    gi_visual = getCard32 r 8;
    gi_image = getString r 32 (4*(getCard32 r 4))
  } 

let polyText8Req window gc x y list =
  let n = textItemListLength list
  in 
  let b = newWinRequest X_PolyText8 (4+(strLen n)) window
  in
  setCard32 b 8 gc;
  setCard16 b 12 x;
  setCard16 b 14 y;
  setTextItem 8 b 16 list;
  for i=16+n to 15+(strLen n)*4 do
    setCard8 b i 0
  done;
  b

let polyText16Req window gc x y list =
  let n = textItemListLength list
  in 
  let b = newWinRequest X_PolyText8 (4+(strLen n)) window
  in
  setCard32 b 8 gc;
  setCard16 b 12 x;
  setCard16 b 14 y;
  setTextItem 16 b 16 list;
  for i=16+n to 15+(strLen n)*4 do
    setCard8 b i 0
  done;
  b

let imageText8Req window gc x y string =
  let n = String.length string
  in
  let b = newWinRequest X_ImageText8 (4+(strLen n)) window
  in
  setCard8 b 1 n;
  setCard32 b 8 gc;
  setCard16 b 12 x;
  setCard16 b 14 y;
  setString b 16 string;
  b
    
let imageSubText8Req window gc x y string pos len =
  let b = newWinRequest X_ImageText8 (4+(strLen len)) window
  in
  setCard8 b 1 len;
  setCard32 b 8 gc;
  setCard16 b 12 x;
  setCard16 b 14 y;
  String.blit string pos b 16 len;
  b
    
let imageText16Req window gc x y array =
  let n = Array.length array
  in
  let b = newWinRequest X_ImageText16 (4+(strLen (2*n))) window
  in
  setCard8 b 1 n;
  setCard32 b 8 gc;
  setCard16 b 12 x;
  setCard16 b 14 y;
  for i = 0 to n - 1 do
    setCard16 b (16+i) array.(i)
  done;
  b
	  
let createColormapReq cid  window visual alloc =
  let b = newRequest X_CreateColormap 4
  in
  setEnum8 b 1 alloc;
  setCard32 b 4 cid;
  setCard32 b 8 window;
  setCard32 b 12 visual;
  b

let freeColormapReq = simpleRequest  X_FreeColormap

let copyColormapAndFreeReq new_cid old_cid =
  doubleRequest X_CopyColormapAndFree new_cid old_cid

let installColormapReq = simpleRequest X_InstallColormap

let uninstallColormapReq = simpleRequest X_UninstallColormap

let listInstalledColormapsReq = simpleRequest X_ListInstalledColormaps

let listInstalledColormapsRep r =
  get_card32_list r 32 (getCard16 r 8)

let allocColorReq cmap red green blue =
  let b = newRequest X_AllocColor 4
  in
  setCard32 b 4 cmap;
  setCard16 b 8 red;
  setCard16 b 10 green;
  setCard16 b 12 blue;
  b
	
let allocColorRep r =
  { ac_pixel = getCard32 r 16;
    ac_color = 
    { red = getCard16 r 8;
      green = getCard16 r 10;
      blue = getCard16 r 12;
    } 
  } 

let allocNamedColorReq cmap name =
  let n = String.length name
  in
  let b = newRequest X_AllocNamedColor (3+(strLen n))
  in
  setCard32 b 4 cmap;
  setCard16 b 8 n;
  setString b 12 name;
  b

let allocNamedColorRep r =
  {anc_pixel = getCard32 r 8;
   anc_exact = { red = getCard16 r 12;
                 green = getCard16 r 14;
                 blue = getCard16 r 16;
               };
    anc_visual = { red = getCard16 r 18;
                   green = getCard16 r 20;
                   blue = getCard16 r 22; 
                 } 
  } 

let allocColorCellsReq cmap colors planes contiguous =
  let b = newRequest X_AllocColorCells 3
  in
  setEnum8 b 1 contiguous;
  setCard32 b 4 cmap;
  setCard16 b 8 colors;
  setCard16 b 10 planes;
  b

let allocColorCellsRep r =
  let n = getCard16 r 8
  in
  { acc_pixels = Array.of_list (get_card32_list r 32 n);
    acc_masks = Array.of_list (get_card32_list r (32+4*n) (getCard16 r 10))
  } 

let allocColorPlanesReq cmap colors reds greens blues contiguous =
  let b = newRequest X_AllocColorPlanes 4
  in
  setEnum8 b 1 contiguous;
  setCard32 b 4 cmap;
  setCard16 b 8 colors;
  setCard16 b 10 reds;
  setCard16 b 12 greens;
  setCard16 b 14 blues;
  b
	
let allocColorPlanesRep r = 
  let n = getCard16 r 8
  in
  { acp_pixels = Array.of_list (get_card32_list r 32 n);
    acp_mask = { red = getCard32 r 12;
                 green = getCard32 r 16;
                 blue = getCard32 r 20;
               } 
  } 

let freeColorsReq cmap plane_mask pixels =
  let n = List.length pixels
  in
  let b = newRequest X_FreeColors (3+n)
  in
  setCard32 b 4 cmap;
  setCard32 b 8 plane_mask;
  set_int32_list b 12 pixels;
  b

let storeColorsReq cmap list =
  let n = List.length list
  in
  let b = newRequest X_StoreColors (2+3*n)
  in
  let rec setColorItem pos = function 
      [] -> ()
    | (pixel,red,green,blue,mask)::tl ->
	setCard32 b pos pixel;
	setCard16 b (pos+4) red;
	setCard16 b (pos+6) green;
	setCard16 b (pos+8) blue;
	setCard8 b (pos+10) (mask_of_list mask);
	setColorItem (pos+12) tl
  in
  setCard32 b 4 cmap;
  setColorItem 8 list;
  b

let storeNamedColorReq cmap pixel mask name =
  let n = String.length name
  in
  let b = newRequest X_StoreNamedColor (4+(strLen n))
  in
  setCard8 b 1 (mask_of_list mask);
  setCard32 b 4 cmap;
  setCard32 b 8 pixel;
  setCard16 b 12 n;
  setString b 16 name;
  b
  
let queryColorsReq cmap list =
  let n = List.length list
  in
  let b = newRequest X_QueryColors (2+n)
  in
  setCard32 b 4 cmap;
  set_int32_list b 8 list;
  b
	

let queryColorsRep r =
  let n = getCard16 r 8
  and res = ref []
  in
  for i = 0 to n-1 do
    res := {
       red = getCard16 r (32+8*i);
       green = getCard16 r (34+8*i);
       blue = getCard16 r (36+8*i);
     } 
       :: (!res)
  done;
  Array.of_list  (!res)

let lookupColorReq cmap name =
  let n = String.length name
  in
  let b = newRequest X_LookupColor (3+(strLen n))
  in
  setCard32 b 4 cmap;
  setCard16 b 8 n;
  setString b 12 name;
  b

let lookupColorRep r =
  { lc_exact = { red = getCard16 r 8;
                 green = getCard16 r 10;
                 blue = getCard16 r 12;
               } ;
    lc_visual = { red = getCard16 r 14;
                  green = getCard16 r 16;
                  blue = getCard16 r 18;
                } 
  }
  
let createCursorReq cid src mask fred fgreen fblue bred bgreen bblue x y =
  let b = newRequest X_CreateCursor 8
  in	
  setCard32 b 4 cid;
  setCard32 b 8 src;
  setCard32 b 12 mask;
  setCard16 b 16 fred;
  setCard16 b 18 fgreen;
  setCard16 b 20 fblue;
  setCard16 b 22 bred;
  setCard16 b 24 bgreen;
  setCard16 b 26 bblue;
  setCard16 b 28 x;
  setCard16 b 30 y;
  b

let createGlyphCursorReq cid src mask src_char mask_char 
    fred fgreen fblue bred bgreen bblue =
  let b = newRequest X_CreateGlyphCursor 8
  in	
  setCard32 b 4 cid;
  setCard32 b 8 src;
  setCard32 b 12 mask;
  setCard16 b 16 src_char;
  setCard16 b 18 mask_char;
  setCard16 b 20 fred;
  setCard16 b 22 fgreen;
  setCard16 b 24 fblue;
  setCard16 b 26 bred;
  setCard16 b 28 bgreen;
  setCard16 b 30 bblue;
  b

let freeCursorReq = simpleRequest X_FreeCursor 

let recolorCursorReq cursor fred fgreen fblue bred bgreen bblue =
  let b = newRequest X_RecolorCursor 5
  in
  setCard32 b 4 cursor;
  setCard16 b 8 fred;
  setCard16 b 10 fgreen;
  setCard16 b 12 fblue;
  setCard16 b 14 bred;
  setCard16 b 16 bgreen;
  setCard16 b 18 bblue;
  b

let queryBestSizeReq objet window width height =
  let b = newRequest X_QueryBestSize 3
  in
  setEnum8 b 1 objet;
  setCard8 b 4 window;
  setCard16 b 8 width;
  setCard16 b 10 height;
  b
	
let queryBestSizeRep r = 
  { qbs_width = getCard16 r 8;
    qbs_height =getCard16 r 8;
  } 

let queryExtensionReq name =
  let n = String.length name
  in
  let b = newRequest X_QueryExtension (2+(strLen n))
  in
  setCard16 b 4 n;
  setString b 8 name;
  b
	
let queryExtensionRep r =
  { qe_present = getEnum8 r 8;
    qe_opcode = getCard8 r 9;
    qe_event = getCard8 r 10;
    qe_error = getCard8 r 11;
  } 

let listExtensionsReq () = emptyRequest X_ListExtensions

let listExtensionsRep r = get_str_list r 32 (getCard8 r 1)

let changeKeyboardMappingReq first array =
  let n = Array.length array in
  let m = Array.length array.(0) in
  let b= newRequest X_ChangeKeyboardMapping (2+n*m)
  in
  setCard8 b 1 n;
  setCard8 b 4 first;
  setCard8 b 5 m;
  for i = 0 to n-1 do
    set_card32_array b (8+i*m*4) array.(i)
  done;
  b

let getKeyboardMappingReq first m =
  let b = newRequest X_GetKeyboardMapping 2
  in
  setCard8 b 4 first;
  setCard8 b 5 m;
  b
	
let getKeyboardMappingRep r = 
  let n = getCard8 r 1 in
  let nm = getCard32 r 4 in
  let m = nm / n in
  let keycodes = Array.create m (get_card32_array r 32 n) in
   for i = 1 to m-1 do
     keycodes.(i) <- get_card32_array r (32+4*i*n) n
   done;
  { gkm_keysyms_per_keycode = n;
    gkm_keycodes = keycodes;
  } 

let changeKeyboardControlReq arg = 
  failwith "changeKeyboardControlReq: not implemented"
let getKeyboardControlReq () = 
  failwith "getKeyboardControlReq: not implemented"
let getKeyboardControlRep r = 
  failwith "getKeyboardControlRep: not implemented"

let bellReq percent =
  let b = newRequest X_Bell 1
  in
  setCard8 b 1 percent;
  b

let changePointerControlReq do_acc acc_num acc_den do_thres threshold =
  let b = newRequest X_ChangePointerControl 3
  in
  setCard16 b 4 acc_num;
  setCard16 b 6 acc_den;
  setCard16 b 8 threshold;
  setEnum8 b 10 do_acc;
  setEnum8 b 11 do_thres;
  b

let getPointerControlReq () = emptyRequest X_GetPointerControl

let getPointerControlRep r =
  { gpc_acc_num = getCard16 r 8;
    gpc_acc_den = getCard16 r 10;
    gpc_threshold = getCard16 r 12;
  } 

let setScreenSaverReq timeout interval prefer_blanking allow_exposures =
  let b = newRequest X_SetScreenSaver 3
  in
  setCard16 b 4 timeout;
  setCard16 b 6 interval;
  setEnum8 b 8 prefer_blanking;
  setEnum8 b 9 allow_exposures;
  b

let getScreenSaverReq () = emptyRequest X_GetScreenSaver

let getScreenSaverRep r =
  { gss_timeout = getCard16 r 8;
    gss_interval = getCard16 r 10;
    gss_pref_blanc = getEnum8 r 12;
    gss_allow_exp = getEnum8 r 13;
  } 

let changeHostsReq mode family name =
  let n = String.length name
  in
  let b = newRequest X_ChangeHosts (2+(strLen n))
  in
  setEnum8 b 1 mode;
  setEnum8 b 4 family;
  setCard16 b 6 n;
  setString b 8 name;
  b

let listHostsReq () = emptyRequest X_ListHosts

let listHostsRep r =
  let rec getHosts r pos nbr =
    if nbr = 0 then 
      []
    else
      let n = getCard16 r (pos+2)
      in
      (getEnum8 r pos,
       getString r (pos+4) n
	 )::(getHosts r (pos+4+4*(strLen n)) (nbr-1))
  in
  { lh_control = getEnum8 r 1;
    lh_list = getHosts r 32 (getCard16 r 8);
  }

let setAccessControlReq access = littleRequest X_SetAccessControl access

let setCloseDownModeReq mode = littleRequest X_SetCloseDownMode mode

let killClientReq  = simpleRequest X_KillClient

let rotatePropertiesReq window delta list =
  let n = List.length list
  in
  let b = newWinRequest X_RotateProperties (3+n) window
  in
  setCard16 b 8 n;
  setCard16 b 10 delta;
  set_int32_list b 12 list;
  b

let forceScreenSaverReq state = littleRequest X_ForceScreenSaver state

let setPointerMappingReq map =
  let n = String.length map
  in
  let b = newRequest X_SetPointerMapping (1+(strLen n))
  in
  setCard8 b 1 n;
  setString b 4 map;
  b

let setPointerMappingRep r =
  if (getEnum8 r 1) = 1 then
    raise MappingBusy

let getPointerMappingReq () = emptyRequest X_GetPointerMapping

let getPointerMappingRep r = getString r 32 (getCard8 r 1)

let setModifierMappingReq list =
  let n = List.length list
  in
  let b = newRequest X_SetModifierMapping (1+2*n)
  in
  setCard8 b 1 n;
  let rec setModifiers pos = function
      [] -> ()
    | h::t -> 
	if (Array.length h) <> 8 then
	  raise 
	    (Invalid_argument "setModifierMapping");
	for i = 0 to 7 do
	  setCard8 b (pos+i) h.(i)
	done;
	setModifiers (pos+8) t
  in
  setModifiers 4 list;
  b
	
let setModifierMappingRep r =
  let status =  getEnum8 r 1
  in
  if status = 1 then
    raise MappingBusy
  else
    if status = 2 then
      raise MappingFailed

let getModifierMappingReq () = emptyRequest X_GetModifierMapping

let getModifierMappingRep r =
  let n = getCard8 r 1 in
  Array.init 8 (fun i ->
    let tab = Array.create n 0
    in
    for j=0 to n-1 do
      tab.(j) <- getCard8 r (32+i*n+j)
    done;
    tab)
