(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes

let get_stub_id stub =
  match !(stub.gc_stub) with
    DirectStub gcstub -> 
      DirectId gcstub.gcstub_num
  | _ -> 
      match stub.locator with
        None -> assert false
      | Some locator ->
          IndirectId(locator.lctr_weaknum, locator.lctr_uid)

let send_synchrone automata ord arg =
  Jprims.set_critical_section true;
  try
    match
      
      match automata.local with
        
        LocalAutomata local ->
          let thread_stub = Jprims.self_stub () in
          Jprims.trigger_def_and_wait (automata, ord, (thread_stub, arg))
            
      | FailedObject _ -> Thread.exit ()
            
      | RemoteObject ->
          let dest_space = automata.position in
          if dest_space.space_state = Down then
            Thread.exit ()
          else
            let thread_stub = Jprims.self_stub () in
            Jmarshal.send dest_space 
              (Synchrone (get_stub_id automata,ord, arg, thread_stub));
            Jprims.wait_on_stub ()
              
      | _ -> assert false
       
    with
      Reply v -> Obj.magic v
    | Exception e -> raise e
  with
    e -> 
      Jprims.set_critical_section false;
      raise e


let hard_send_asynchrone automata ord (arg : 'a) =

  match automata.local with
    
    LocalAutomata local ->
      Jprims.trigger_def (automata, ord, arg);

  | FailedObject _ -> ()

  | RemoteObject ->
      Jprims.test_channel "remote" automata;
      let dest_space = automata.position in
      if dest_space.space_state <> Down then
        Jmarshal.send dest_space 
          (Asynchrone (get_stub_id automata,ord, arg))

  | _ -> assert false

let send_asynchrone automata ord (arg : 'a) =
  Jprims.critical_section (function () ->
    hard_send_asynchrone automata ord arg)

(* ==========  DEBUT hack !!!!  ========== *)
open Obj
external unbox_channel:  'a -> automata_stub * int = "unbox_channel"

let _ =
  let automata = (magic  374848: automata_stub) in
  let ord = 298274 in
  let f = repr (send_synchrone automata ord) in
  let g = repr (send_asynchrone automata ord) in
  Jprims.register_send_sync f g

(* ==========   FIN hack !!!!  ========== *)

let hard_reply_to thread_stub v =

  match thread_stub.local with

    LocalThread local -> Jprims.reply_thread_stub thread_stub v

  | FailedObject _ -> ()

  | RemoteObject ->
      let dest_space = thread_stub.position in
      if dest_space.space_state <> Down then
         Jmarshal.send dest_space
                (ReplyThread (get_stub_id thread_stub,Obj.repr v))

  | _ -> assert false
        
let reply_to_and_exit thread_stub v = 
  Jprims.set_critical_section  true;
  try
    hard_reply_to thread_stub v;
    Thread.exit ()
  with
    e ->
      Jprims.set_critical_section false;
        (* on ne peut pas relancer l'exception ! *)
      prerr_string "Warning: exception in reply_to_and_exit: Thread.exit\n";
      Thread.exit ();
      raise e

let reply_to thread_stub v =
  Jprims.critical_section 
    (function () -> hard_reply_to thread_stub v)

(* For now, tail call is only possible between threads and definitions
   of the same location. Such a complete information is available at runtime,
   but not kept actually, and a partial information is computable at compile
   time. We hope compile time information will be enough ... for now
   *)
let safe_tail_synchrone (channel:channel) arg thread_stub =
  let (automata, ord) = unbox_channel channel in
  Jprims.set_critical_section true;
  match automata.local with
    LocalAutomata local ->
      if local.location = Jprims.self_location () then
        begin
          Debug.print0 'S' "tailcall succeeded but tested\n";
          Jprims.exit_trigger_def_and_wait (automata, ord, (thread_stub, arg))
        end
      else
        begin
          Jprims.set_critical_section false;
          reply_to_and_exit thread_stub (Reply ((Obj.magic channel) arg))
        end
  | FailedObject _ -> Thread.exit ()
  | _ ->
      Jprims.set_critical_section false;
      Jprims.exit_trigger_def_and_wait (automata, ord, (thread_stub, arg))

  reply_to_and_exit thread_stub (Reply ((Obj.magic channel) arg))

let unsafe_tail_synchrone (channel:channel) arg thread_stub =
  Debug.print0 'S' "tailcall succeeded\n";
  let (automata, ord) = unbox_channel channel in
  Jprims.set_critical_section true;
  Jprims.exit_trigger_def_and_wait (automata, ord, (thread_stub, arg))

let rec invoke_no_reply object meth_name label args thread_stub =
  Jprims.critical_section
    (function () ->      
      match object.local with
        
        LocalLocation local ->
          Jprims.thread_create_in object
            (fun () ->
              reply_to_and_exit thread_stub
                (try
                  Reply (
                  Mop.invoke_methode 
                    (Mop.lookup (Obj.magic object) label) object 
                    (List.rev args)
                    )    
                with
                  e -> Exception e
                      )
                ) (); ()

      | FailedObject _ -> ()

      | RemoteObject ->
          let dest_space = object.position in
          if dest_space.space_state <> Down then
            Jmarshal.send dest_space
                    (Invoke (get_stub_id object,
                             meth_name,label,args,thread_stub))

      | _ -> assert false)

let create_ext_buckets class_sig =
  Mop.create_ext_buckets class_sig 
    (fun nparams meth_name label ->
      Debug.print2 'o' "Method %s with  %d params\n" meth_name nparams;
      (Obj.magic
         (fun object ->
           let rec iter nparams args =
             Debug.print1 'o' "Iter %d params\n" nparams;
             if nparams = 0 then
               (* ici, on est sur de rien : l'objet peut etre local
                  ou peut avoir migre *)
               let thread_stub = Jprims.self_stub () in
               Jprims.set_critical_section true;
               invoke_no_reply object meth_name label args thread_stub;
               match Jprims.wait_on_stub (Obj.magic object) with
               | Reply v -> Obj.magic v
               | Exception e -> raise e
             else
               Obj.magic (fun arg -> iter (nparams-1) (arg::args))
           in
           iter nparams []
             ): Mop.item)
        )
    
let remote_ext_buckets space class_id class_sig =
  try
    Hashtbl.find space.ext_buckets class_id
  with
    Not_found ->
      let ext_buckets = create_ext_buckets class_sig
      in
      Hashtbl.add space.ext_buckets class_id ext_buckets;
      ext_buckets
        
external sys_exit : exn -> 'a = "join_sys_exit"

let send_exception exn array =
  try
    let array = (Obj.magic array: thread_stub array)
    in
    Jprims.set_critical_section  true;
    Debug.print "send exception";
    Debug.print "send_exception";
    Array.iter (fun thread_stub ->
        hard_reply_to thread_stub (Exception exn);
    ) array;
    Debug.print "now exit";
    Thread.exit ()
  with
    e ->  sys_exit e  
      
let send_synchrone_obj chan obj = send_synchrone obj chan
let send_asynchrone_obj chan obj = send_asynchrone obj chan

